const otherDir = (dir) => (dir === 'left' ? 'right' : 'left');
const mqCmdId = 'data-mathquill-command-id', mqBlockId = 'data-mathquill-block-id';
const noop = () => {
    /* do nothing */
};
const iterator = (generator) => {
    return (fn, ...args) => {
        const yield_ = typeof fn === 'function'
            ? fn
            : (obj) => {
                if (fn in obj)
                    return obj[fn](...args);
            };
        return generator(yield_);
    };
};
// sugar to make defining lots of commands easier.
const bindMixin = (Base, ...args) => class extends Base {
    constructor(..._args) {
        super(...args);
    }
};
const prayWellFormed = (parent, leftward, rightward) => {
    if (!parent)
        throw new Error('a parent must be present');
    // Either leftward is empty and `rightward` is the left end child (possibly empty)
    // or leftward is there and leftward's .right and .parent are properly set up.
    if ((!leftward && parent.ends.left !== rightward) ||
        (leftward && (leftward.right !== rightward || leftward.parent !== parent)))
        throw new Error('leftward is not properly set up');
    // Either rightward is empty and `leftward` is the right end child (possibly empty)
    // or rightward is there and rightward's .left and .parent are properly set up.
    if ((!rightward && parent.ends.right !== leftward) ||
        (rightward && (rightward.left !== leftward || rightward.parent !== parent)))
        throw new Error('rightward is not properly set up');
};
// Registry of LaTeX commands and commands created when typing a single character.
// (Commands are all subclasses of tree/TNode.)
const LatexCmds = {}, CharCmds = {};
const OPP_BRACKS = {
    '(': ')',
    ')': '(',
    '[': ']',
    ']': '[',
    '{': '}',
    '}': '{',
    '\\{': '\\}',
    '\\}': '\\{',
    '&lang;': '&rang;',
    '&rang;': '&lang;',
    '\\langle ': '\\rangle ',
    '\\rangle ': '\\langle ',
    '|': '|',
    '&#8741;': '&#8741;',
    '\\lVert ': '\\rVert ',
    '\\rVert ': '\\lVert '
};
const BRACKET_NAMES = {
    '&lang;': 'angle-bracket',
    '&rang;': 'angle-bracket',
    '|': 'pipe',
    '(': 'parenthesis',
    ')': 'parenthesis',
    '[': 'bracket',
    ']': 'bracket',
    '{': 'brace',
    '}': 'brace'
};
const EMBEDS = {};
// The set of operator names like \arg, \det, etc that are built-into LaTeX,
// see Section 3.15 of the Short Math Guide: http://tug.ctan.org/info/short-math-guide/short-math-guide.pdf
// MathQuill auto-unitalicizes some operator names not in that set, like 'hcf',
// which must be exported as \operatorname{hcf}.
// Note: over/under line/arrow \lim variants like \varlimsup are not supported.
// Note that this no longer includes ln, log, exp, or any of the trig functions.
// Those are now latex commands that are implemented by the MathFunction class.
const BuiltInOpNames = {};
// Standard operators
for (const op of [
    'arg',
    'det',
    'dim',
    'gcd',
    'hom',
    'ker',
    'lg',
    'lim',
    'max',
    'min',
    'sup',
    'limsup',
    'liminf',
    'injlim',
    'projlim',
    'Pr'
]) {
    BuiltInOpNames[op] = 1;
}
const TwoWordOpNames = { limsup: 1, liminf: 1, projlim: 1, injlim: 1 };
const SVG_SYMBOLS = {
    sqrt: {
        width: '',
        html: '<svg preserveAspectRatio="none" viewBox="0 0 32 54">' +
            '<path d="M0 33 L7 27 L12.5 47 L13 47 L30 0 L32 0 L13 54 L11 54 L4.5 31 L0 33"></path>' +
            '</svg>'
    },
    '|': {
        width: '0.4em',
        html: '<svg preserveAspectRatio="none" viewBox="0 0 10 54">' +
            '<path d="M4.4 0 L4.4 54 L5.6 54 L5.6 0"></path>' +
            '</svg>'
    },
    '[': {
        width: '0.55em',
        html: '<svg preserveAspectRatio="none" viewBox="0 0 11 24">' +
            '<path d="M8 0 L3 0 L3 24 L8 24 L8 23 L4 23 L4 1 L8 1"></path>' +
            '</svg>'
    },
    ']': {
        width: '0.55em',
        html: '<svg preserveAspectRatio="none" viewBox="0 0 11 24">' +
            '<path d="M3 0 L8 0 L8 24 L3 24 L3 23 L7 23 L7 1 L3 1"></path>' +
            '</svg>'
    },
    '(': {
        width: '0.55em',
        html: '<svg preserveAspectRatio="none" viewBox="3 0 106 186">' +
            '<path d="M85 0 A61 101 0 0 0 85 186 L75 186 A75 101 0 0 1 75 0"></path>' +
            '</svg>'
    },
    ')': {
        width: '.55em',
        html: '<svg preserveAspectRatio="none" viewBox="3 0 106 186">' +
            '<path d="M24 0 A61 101 0 0 1 24 186 L34 186 A75 101 0 0 0 34 0"></path>' +
            '</svg>'
    },
    '{': {
        width: '0.7em',
        html: '<svg preserveAspectRatio="none" viewBox="10 0 210 350">' +
            '<path d="' +
            'M170 0 L170 6 A47 52 0 0 0 123 60 L123 127 A35 48 0 0 1 88 175 A35 48 0 0 1 123 223 L123 290 A47 52 ' +
            '0 0 0 170 344 L170 350 L160 350 A58 49 0 0 1 102 301 L103 220 A45 40 0 0 0 58 180 L58 170 A45 40 ' +
            '0 0 0 103 130 L103 49 A58 49 0 0 1 161 0' +
            '"></path>' +
            '</svg>'
    },
    '}': {
        width: '0.7em',
        html: '<svg preserveAspectRatio="none" viewBox="10 0 210 350">' +
            '<path d="' +
            'M60 0 L60 6 A47 52 0 0 1 107 60 L107 127 A35 48 0 0 0 142 175 A35 48 0 0 0 107 223 L107 290 A47 52 ' +
            '0 0 1 60 344 L60 350 L70 350 A58 49 0 0 0 128 301 L127 220 A45 40 0 0 1 172 180 L172 170 A45 40 ' +
            '0 0 1 127 130 L127 49 A58 49 0 0 0 70 0' +
            '"></path>' +
            '</svg>'
    },
    '&#8741;': {
        width: '0.7em',
        html: '<svg preserveAspectRatio="none" viewBox="0 0 10 54">' +
            '<path d="M3.2 0 L3.2 54 L4 54 L4 0 M6.8 0 L6.8 54 L6 54 L6 0"></path>' +
            '</svg>'
    },
    '&lang;': {
        width: '0.55em',
        html: '<svg preserveAspectRatio="none" viewBox="0 0 10 54">' +
            '<path d="M6.8 0 L3.2 27 L6.8 54 L7.8 54 L4.2 27 L7.8 0"></path>' +
            '</svg>'
    },
    '&rang;': {
        width: '0.55em',
        html: '<svg preserveAspectRatio="none" viewBox="0 0 10 54">' +
            '<path d="M3.2 0 L6.8 27 L3.2 54 L2.2 54 L5.8 27 L2.2 0"></path>' +
            '</svg>'
    }
};

// Options for the API objects
class Options {
    static config(currentOptions, newOptions) {
        Object.assign(currentOptions, newOptions);
    }
    // Each option has a static variable for the default setting, and an instance variable for the setting of a
    // particular math field.
    // Wether mouse events are active for StaticMath blocks
    static #mouseEvents = true;
    #_mouseEvents;
    get mouseEvents() {
        return this.#_mouseEvents ?? Options.#mouseEvents;
    }
    set mouseEvents(mouseEvents) {
        if (this instanceof Options)
            this.#_mouseEvents = mouseEvents;
        else
            Options.#mouseEvents = mouseEvents;
    }
    // The set of commands that are automatically typeset without typing a preceding backslash.
    static #autoCommands = { _maxLength: 0 };
    #_autoCommands;
    get autoCommands() {
        return this.#_autoCommands ?? Options.#autoCommands;
    }
    set autoCommands(cmds) {
        if (typeof cmds === 'object') {
            if (this instanceof Options) {
                this.#_autoCommands = { _maxLength: 0 };
                Object.assign(this.#_autoCommands, cmds);
            }
            else
                Object.assign(Options.#autoCommands, cmds);
            return;
        }
        if (!/^\s*[a-z]+(?:\s+[a-z]+)*\s*$/i.test(cmds)) {
            throw new Error(`"${cmds}" not a space-delimited list of only letters`);
        }
        const list = cmds.trim().split(/\s+/), dict = { _maxLength: 0 };
        for (const cmd of list) {
            if (cmd.length < 2)
                throw new Error(`autocommand "${cmd}" not minimum length of 2`);
            if (cmd in BuiltInOpNames)
                throw new Error(`"${cmd}" is a built-in operator name`);
            dict[cmd] = 1;
            dict._maxLength = Math.max(dict._maxLength, cmd.length);
        }
        if (this instanceof Options)
            this.#_autoCommands = dict;
        else
            Options.#autoCommands = dict;
    }
    addAutoCommands(cmds) {
        if (!this.#_autoCommands)
            this.autoCommands = Options.#autoCommands;
        if (!this.#_autoCommands)
            throw new Error('autoCommands setter not working');
        const newCmds = cmds instanceof Array ? cmds.map((c) => c.trim()) : [cmds.trim()];
        for (const cmd of newCmds) {
            if (/\s/.test(cmd) || !/^[a-z]*$/i.test(cmd))
                throw new Error(`${cmd} is not a valid autocommand name`);
            if (cmd.length < 2)
                throw new Error(`autocommand "${cmd}" not minimum length of 2`);
            if (cmd in BuiltInOpNames)
                throw new Error(`"${cmd}" is a built-in operator name`);
            this.#_autoCommands[cmd] = 1;
            this.#_autoCommands._maxLength = Math.max(this.#_autoCommands._maxLength, cmd.length);
        }
    }
    removeAutoCommands(cmds) {
        if (!this.#_autoCommands)
            this.autoCommands = Options.#autoCommands;
        if (!this.#_autoCommands)
            throw new Error('autoCommands setter not working');
        const removeCmds = cmds instanceof Array ? cmds.map((c) => c.trim()) : [cmds.trim()];
        for (const cmd of removeCmds) {
            // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
            delete this.#_autoCommands[cmd];
        }
        this.#_autoCommands._maxLength = Object.keys(this.#_autoCommands).reduce((l, cmd) => (cmd === '_maxLength' ? l : cmd.length > l ? cmd.length : l), 0);
    }
    // The set of operator names that MathQuill auto-unitalicizes.
    static #autoOperatorNames = { _maxLength: 0 };
    #_autoOperatorNames;
    get autoOperatorNames() {
        return this.#_autoOperatorNames ?? Options.#autoOperatorNames;
    }
    set autoOperatorNames(cmds) {
        if (typeof cmds === 'object') {
            if (this instanceof Options) {
                this.#_autoOperatorNames = { _maxLength: 0 };
                Object.assign(this.#_autoOperatorNames, cmds);
            }
            else
                Object.assign(Options.#autoOperatorNames, cmds);
            return;
        }
        if (!/^\s*[a-z]+(?:\s+[a-z]+)*\s*$/i.test(cmds)) {
            throw new Error(`"${cmds}" not a space-delimited list of only letters`);
        }
        const list = cmds.trim().split(/\s+/), dict = { _maxLength: 0 };
        for (const cmd of list) {
            if (cmd.length < 2)
                throw new Error(`"${cmd}" not minimum length of 2`);
            dict[cmd] = 1;
            dict._maxLength = Math.max(dict._maxLength, cmd.length);
        }
        if (this instanceof Options)
            this.#_autoOperatorNames = dict;
        else
            Options.#autoOperatorNames = dict;
    }
    addAutoOperatorNames(cmds) {
        if (!this.#_autoOperatorNames)
            this.autoOperatorNames = Options.#autoOperatorNames;
        if (!this.#_autoOperatorNames)
            throw new Error('autoOperatorNames setter not working');
        const newCmds = cmds instanceof Array ? cmds.map((c) => c.trim()) : [cmds.trim()];
        for (const cmd of newCmds) {
            if (/\s/.test(cmd) || !/^[a-z]*$/i.test(cmd))
                throw new Error(`${cmd} is not a valid autocommand name`);
            if (cmd.length < 2)
                throw new Error(`"${cmd}" not minimum length of 2`);
            this.#_autoOperatorNames[cmd] = 1;
            this.#_autoOperatorNames._maxLength = Math.max(this.#_autoOperatorNames._maxLength, cmd.length);
        }
    }
    removeAutoOperatorNames(cmds) {
        if (!this.#_autoOperatorNames)
            this.autoOperatorNames = Options.#autoOperatorNames;
        if (!this.#_autoOperatorNames)
            throw new Error('autoOperatorNames setter not working');
        const removeCmds = cmds instanceof Array ? cmds.map((c) => c.trim()) : [cmds.trim()];
        for (const cmd of removeCmds) {
            // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
            delete this.#_autoOperatorNames[cmd];
        }
        this.#_autoOperatorNames._maxLength = Object.keys(this.#_autoOperatorNames).reduce((l, cmd) => (cmd === '_maxLength' ? l : cmd.length > l ? cmd.length : l), 0);
    }
    // Characters that "break out" of superscripts and subscripts
    static #charsThatBreakOutOfSupSub = '';
    #_charsThatBreakOutOfSupSub;
    get charsThatBreakOutOfSupSub() {
        return this.#_charsThatBreakOutOfSupSub ?? Options.#charsThatBreakOutOfSupSub;
    }
    set charsThatBreakOutOfSupSub(charsThatBreakOutOfSupSub) {
        if (this instanceof Options)
            this.#_charsThatBreakOutOfSupSub = charsThatBreakOutOfSupSub;
        else
            Options.#charsThatBreakOutOfSupSub = charsThatBreakOutOfSupSub;
    }
    // Not fully implemented stateless clipboard
    static #statelessClipboard = false;
    #_statelessClipboard;
    get statelessClipboard() {
        return this.#_statelessClipboard ?? Options.#statelessClipboard;
    }
    set statelessClipboard(statelessClipboard) {
        if (this instanceof Options)
            this.#_statelessClipboard = statelessClipboard;
        else
            Options.#statelessClipboard = statelessClipboard;
    }
    // If true then space will behave like tab escaping from the current block instead of inserting a space.
    static #enableSpaceNavigation = false;
    #_enableSpaceNavigation;
    get enableSpaceNavigation() {
        return this.#_enableSpaceNavigation ?? Options.#enableSpaceNavigation;
    }
    set enableSpaceNavigation(enableSpaceNavigation) {
        if (this instanceof Options)
            this.#_enableSpaceNavigation = enableSpaceNavigation;
        else
            Options.#enableSpaceNavigation = enableSpaceNavigation;
    }
    // Set to 'up' or 'down' so that left and right go up or down (respectively) into commands.
    static #leftRightIntoCmdGoes = undefined;
    #_leftRightIntoCmdGoes;
    get leftRightIntoCmdGoes() {
        return this.#_leftRightIntoCmdGoes ?? Options.#leftRightIntoCmdGoes;
    }
    set leftRightIntoCmdGoes(updown) {
        if (updown && updown !== 'up' && updown !== 'down') {
            throw new Error(`"up" or "down" required for leftRightIntoCmdGoes option, got "${updown}"`);
        }
        if (this instanceof Options)
            this.#_leftRightIntoCmdGoes = updown;
        else
            Options.#leftRightIntoCmdGoes = updown;
    }
    // If true then you can type '[a,b)' and '(a,b]', but if you type '[x}' or '{x)', you'll get '[{x}]' or '{(x)}'
    // instead.
    static #restrictMismatchedBrackets = false;
    #_restrictMismatchedBrackets;
    get restrictMismatchedBrackets() {
        return this.#_restrictMismatchedBrackets ?? Options.#restrictMismatchedBrackets;
    }
    set restrictMismatchedBrackets(restrictMismatchedBrackets) {
        if (this instanceof Options)
            this.#_restrictMismatchedBrackets = restrictMismatchedBrackets;
        else
            Options.#restrictMismatchedBrackets = restrictMismatchedBrackets;
    }
    // If true then when you type '\sum', '\prod', or '\coprod', the lower limit starts out with 'n='.
    static #sumStartsWithNEquals = false;
    #_sumStartsWithNEquals;
    get sumStartsWithNEquals() {
        return this.#_sumStartsWithNEquals ?? Options.#sumStartsWithNEquals;
    }
    set sumStartsWithNEquals(sumStartsWithNEquals) {
        if (this instanceof Options)
            this.#_sumStartsWithNEquals = sumStartsWithNEquals;
        else
            Options.#sumStartsWithNEquals = sumStartsWithNEquals;
    }
    // Disables typing of superscripts and subscripts when there's nothing to the left of the cursor.
    static #supSubsRequireOperand = false;
    #_supSubsRequireOperand;
    get supSubsRequireOperand() {
        return this.#_supSubsRequireOperand ?? Options.#supSubsRequireOperand;
    }
    set supSubsRequireOperand(supSubsRequireOperand) {
        if (this instanceof Options)
            this.#_supSubsRequireOperand = supSubsRequireOperand;
        else
            Options.#supSubsRequireOperand = supSubsRequireOperand;
    }
    // If true then the text output for an nth root will be 'x^(1/n)' instead of 'root(n,x)'.
    static #rootsAreExponents = false;
    #_rootsAreExponents;
    get rootsAreExponents() {
        return this.#_rootsAreExponents ?? Options.#rootsAreExponents;
    }
    set rootsAreExponents(rootsAreExponents) {
        if (this instanceof Options)
            this.#_rootsAreExponents = rootsAreExponents;
        else
            Options.#rootsAreExponents = rootsAreExponents;
    }
    // If true then the text output for the logarithm with base b of x will be 'log(x)/log(b)'.  Otherwise the output
    // will be 'logb(b,x)'.  Note that this option does not affect base 10 output.  That is always "log10(x)".
    static #logsChangeBase = false;
    #_logsChangeBase;
    get logsChangeBase() {
        return this.#_logsChangeBase ?? Options.#logsChangeBase;
    }
    set logsChangeBase(logsChangeBase) {
        if (this instanceof Options)
            this.#_logsChangeBase = logsChangeBase;
        else
            Options.#logsChangeBase = logsChangeBase;
    }
    // Specifies the maximum number of nested MathBlocks allowed.
    static #maxDepth = undefined;
    #_maxDepth;
    get maxDepth() {
        return this.#_maxDepth ?? Options.#maxDepth;
    }
    set maxDepth(maxDepth) {
        if (typeof maxDepth === 'number') {
            if (this instanceof Options)
                this.#_maxDepth = maxDepth;
            else
                Options.#maxDepth = maxDepth;
        }
    }
    // If true then a number typed after a letter will automatically be put into a subscript.
    static #autoSubscriptNumerals = false;
    #_autoSubscriptNumerals;
    get autoSubscriptNumerals() {
        return this.#_autoSubscriptNumerals ?? Options.#autoSubscriptNumerals;
    }
    set autoSubscriptNumerals(autoSubscriptNumerals) {
        if (this instanceof Options)
            this.#_autoSubscriptNumerals = autoSubscriptNumerals;
        else
            Options.#autoSubscriptNumerals = autoSubscriptNumerals;
    }
    // If true then typing a slash gives the division symbol instead of a live fraction.
    static #typingSlashWritesDivisionSymbol = false;
    #_typingSlashWritesDivisionSymbol;
    get typingSlashWritesDivisionSymbol() {
        return this.#_typingSlashWritesDivisionSymbol ?? Options.#typingSlashWritesDivisionSymbol;
    }
    set typingSlashWritesDivisionSymbol(typingSlashWritesDivisionSymbol) {
        if (this instanceof Options)
            this.#_typingSlashWritesDivisionSymbol = typingSlashWritesDivisionSymbol;
        else
            Options.#typingSlashWritesDivisionSymbol = typingSlashWritesDivisionSymbol;
    }
    // If true then typing an asterisk gives the times symbol.
    static #typingAsteriskWritesTimesSymbol = false;
    #_typingAsteriskWritesTimesSymbol;
    get typingAsteriskWritesTimesSymbol() {
        return this.#_typingAsteriskWritesTimesSymbol ?? Options.#typingAsteriskWritesTimesSymbol;
    }
    set typingAsteriskWritesTimesSymbol(typingAsteriskWritesTimesSymbol) {
        if (this instanceof Options)
            this.#_typingAsteriskWritesTimesSymbol = typingAsteriskWritesTimesSymbol;
        else
            Options.#typingAsteriskWritesTimesSymbol = typingAsteriskWritesTimesSymbol;
    }
    substituteTextarea(tabbable) {
        const textarea = document.createElement('textarea');
        textarea.setAttribute('autocapitalize', 'off');
        textarea.setAttribute('autocomplete', 'off');
        textarea.setAttribute('spellcheck', 'false');
        textarea.tabIndex = tabbable ? 0 : -1;
        return textarea;
    }
    overridePaste;
    overrideCut;
    overrideCopy;
    overrideTypedText;
    overrideKeystroke;
    ignoreNextMousedown = () => false;
    blurWithCursor;
    static #tabbable;
    #_tabbable;
    get tabbable() {
        return typeof this.#_tabbable === 'boolean' ? this.#_tabbable : Options.#tabbable;
    }
    set tabbable(tabbable) {
        if (this instanceof Options)
            this.#_tabbable = tabbable;
        else
            Options.#tabbable = tabbable;
    }
}

// Virtual Node class
class VNode {
    contents = [];
    constructor(dom) {
        if (dom instanceof VNode)
            this.contents.push(...dom.contents);
        else if (dom instanceof Node)
            this.contents.push(dom);
        else if (dom instanceof Array)
            this.contents.push(...dom);
        else if (dom) {
            const fragment = document.createRange().createContextualFragment(dom);
            this.contents.push(...fragment.children);
        }
    }
    add(el) {
        this.contents.push(...(el instanceof VNode ? el.contents : el instanceof Array ? el : [el]));
        // Sort the contents into document order.
        this.contents.sort((a, b) => {
            if (a === b)
                return 0;
            const position = a.compareDocumentPosition(b);
            if (position & Node.DOCUMENT_POSITION_FOLLOWING || position & Node.DOCUMENT_POSITION_CONTAINED_BY)
                return -1;
            else if (position & Node.DOCUMENT_POSITION_PRECEDING || position & Node.DOCUMENT_POSITION_CONTAINS)
                return 1;
            return 0;
        });
    }
    // The following methods are guaranteed to return a valid Element or CharacterData object.
    get first() {
        return this.contents[0] instanceof Element || this.contents[0] instanceof CharacterData
            ? this.contents[0]
            : document.createElement('span');
    }
    get last() {
        const last = this.contents[this.contents.length - 1];
        return last instanceof Element || last instanceof CharacterData ? last : document.createElement('span');
    }
    // These methods are guaranteed to return a valid HTMLElement object.
    get firstElement() {
        return this.first instanceof HTMLElement ? this.first : document.createElement('span');
    }
    get lastElement() {
        return this.last instanceof HTMLElement ? this.last : document.createElement('span');
    }
    detach() {
        this.contents.forEach((child) => {
            child.remove();
        });
    }
    remove() {
        this.detach();
        this.contents = [];
    }
    empty() {
        this.contents.forEach((child) => {
            while (child.firstChild)
                child.firstChild.remove();
            child.textContent = '';
        });
    }
    children(selector) {
        return new VNode(this.contents.reduce((ret, el) => {
            ret.push(...Array.from(el.childNodes).filter((child) => {
                return (child.nodeType !== 3 && (!selector || (child instanceof Element && child.matches(selector))));
            }));
            return ret;
        }, []));
    }
    find(selector) {
        return new VNode(this.contents.reduce((ret, el) => {
            ret.push(...el.querySelectorAll(selector));
            return ret;
        }, []));
    }
    addClass(...tokens) {
        this.contents.forEach((child) => {
            if (child instanceof Element)
                child.classList.add(...tokens);
        });
    }
    removeClass(...tokens) {
        this.contents.forEach((child) => {
            if (child instanceof Element)
                child.classList.remove(...tokens);
        });
    }
    toggleClass(token, state) {
        this.contents.forEach((child) => {
            if (child instanceof Element)
                child.classList.toggle(token, state);
        });
    }
    hasClass(token) {
        for (const child of this.contents) {
            if (child instanceof Element && child.classList.contains(token))
                return true;
        }
        return false;
    }
    html(contents) {
        if (typeof contents === 'string') {
            this.contents.forEach((elt) => {
                if (elt instanceof Element)
                    elt.innerHTML = contents;
            });
            return this;
        }
        return this.firstElement.innerHTML;
    }
    text(contents) {
        if (contents) {
            this.contents.forEach((elt) => {
                if (elt instanceof Element)
                    elt.textContent = contents;
            });
            return this;
        }
        return this.contents.reduce((ret, child) => {
            return `${ret}${child.textContent ?? ''}`;
        }, '');
    }
    insDirOf(dir, vNode) {
        if (vNode instanceof VNode && !vNode.contents.length)
            return;
        if (dir === 'left')
            (vNode instanceof VNode ? vNode.first : vNode).before(...this.contents);
        else
            (vNode instanceof VNode ? vNode.last : vNode).after(...this.contents);
    }
    insAtDirEnd(dir, vNode) {
        if (vNode instanceof VNode && !vNode.contents.length)
            return;
        if (dir === 'left')
            (vNode instanceof VNode ? vNode.firstElement : vNode).prepend(...this.contents);
        else
            (vNode instanceof VNode ? vNode.lastElement : vNode).append(...this.contents);
    }
}

// Fragment base classes of edit tree-related objects
// An entity outside the virtual tree with one-way pointers (so it's only a
// "view" of part of the tree, not an actual node/entity in the tree) that
// delimits a doubly-linked list of sibling nodes.
// It's like a fanfic love-child between HTML DOM DocumentFragment and the Range
// classes: like DocumentFragment, its contents must be sibling nodes
// (unlike Range, whose contents are arbitrary contiguous pieces of subtrees),
// but like Range, it has only one-way pointers to its contents, its contents
// have no reference to it and in fact may still be in the visible tree (unlike
// DocumentFragment, whose contents must be detached from the visible tree
// and have their 'parent' pointers set to the DocumentFragment).
class Fragment {
    elements = new VNode();
    ends = {};
    disowned;
    each = iterator((yield_) => {
        let el = this.ends.left;
        if (!el)
            return this;
        for (; el !== this.ends.right?.right; el = el.right) {
            if (!el)
                throw new Error('each unable to continue right in Fragment');
            if (yield_(el) === false)
                break;
        }
        return this;
    });
    constructor(withDir, oppDir, dir = 'left') {
        if (!withDir !== !oppDir)
            throw new Error('no half-empty fragments');
        if (!withDir)
            return;
        if (!(withDir instanceof TNode))
            throw new Error('withDir must be passed to Fragment');
        if (!(oppDir instanceof TNode))
            throw new Error('oppDir must be passed to Fragment');
        if (withDir.parent !== oppDir.parent)
            throw new Error('withDir and oppDir must have the same parent');
        if (dir === 'left') {
            this.ends.left = withDir;
            this.ends.right = oppDir;
        }
        else {
            this.ends.left = oppDir;
            this.ends.right = withDir;
        }
        // To build the html collection for a fragment, accumulate elements into an array and then call elements.add
        // once on the result. elements.add sorts the collection according to document order each time it is called, so
        // building a collection by folding elements.add directly takes more than quadratic time in the number of
        // elements.
        const accum = this.fold([], (accum, el) => {
            accum.push(...el.elements.contents);
            return accum;
        });
        this.elements.add(accum);
    }
    // like Cursor::withDirInsertAt(dir, parent, withDir, oppDir)
    withDirAdopt(dir, parent, withDir, oppDir) {
        return dir === 'left' ? this.adopt(parent, withDir, oppDir) : this.adopt(parent, oppDir, withDir);
    }
    adopt(parent, leftward, rightward) {
        prayWellFormed(parent, leftward, rightward);
        this.disowned = false;
        const leftEnd = this.ends.left;
        if (!leftEnd)
            return this;
        const rightEnd = this.ends.right;
        if (leftward) ;
        else {
            parent.ends.left = leftEnd;
        }
        if (rightward) {
            rightward.left = rightEnd;
        }
        else {
            parent.ends.right = rightEnd;
        }
        if (this.ends.right)
            this.ends.right.right = rightward;
        this.each((el) => {
            el.left = leftward;
            el.parent = parent;
            if (leftward)
                leftward.right = el;
            leftward = el;
            return true;
        });
        return this;
    }
    disown() {
        const leftEnd = this.ends.left;
        // guard for empty and already-disowned fragments
        if (!leftEnd || this.disowned)
            return this;
        this.disowned = true;
        const rightEnd = this.ends.right;
        const parent = leftEnd.parent;
        if (!parent)
            throw new Error('a parent must always present');
        prayWellFormed(parent, leftEnd.left, leftEnd);
        prayWellFormed(parent, rightEnd, rightEnd?.right);
        if (leftEnd.left) {
            leftEnd.left.right = rightEnd?.right;
        }
        else {
            parent.ends.left = rightEnd?.right;
        }
        if (rightEnd?.right) {
            rightEnd.right.left = leftEnd.left;
        }
        else {
            parent.ends.right = leftEnd.left;
        }
        return this;
    }
    remove() {
        this.elements.remove();
        this.each('postOrder', 'dispose');
        return this.disown();
    }
    fold(fold, fn) {
        let ret = fold;
        this.each((el) => {
            ret = fn(ret, el);
            return true;
        });
        return ret;
    }
}

// Selection "singleton" class
// Sort of a singleton, since there should only be one per editable math textbox, but any one HTML document can contain
// many such textboxes, so any one JS environment could actually contain many instances.
class Selection extends Fragment {
    constructor(withDir, oppDir, dir = 'left') {
        super(withDir, oppDir, dir);
        const wrapper = document.createElement('span');
        wrapper.classList.add('mq-selection');
        this.elements.first.before(wrapper);
        wrapper.append(...this.elements.contents);
        this.elements = new VNode(wrapper);
    }
    adopt(parent, leftward, rightward) {
        const children = this.elements.children();
        this.elements.first.replaceWith(...children.contents);
        this.elements = children;
        return super.adopt(parent, leftward, rightward);
    }
    clear() {
        this.elements.first.replaceWith(...this.elements.first.childNodes);
        return this;
    }
    join(methodName, separator = '') {
        return this.fold('', (fold, child) => fold + separator + child[methodName]());
    }
}

// TNode base class of edit tree-related objects
const prayOverridden = (name) => {
    throw new Error(`"${name}" should be overridden or never called on this node`);
};
// MathQuill virtual-DOM tree-node abstract base class
// Only doing tree node manipulation via these adopt/disown methods guarantees well-formedness of the tree.
class TNode {
    static id = 0;
    static byId = new Map();
    static uniqueNodeId = () => ++TNode.id;
    elements = new VNode();
    id;
    parent;
    ends = {};
    left;
    right;
    controller;
    ctrlSeq = '';
    siblingDeleted;
    siblingCreated;
    sub;
    sup;
    isSymbol;
    isSupSubLeft;
    ariaLabel;
    mathspeakName;
    mathspeakTemplate;
    upInto;
    downInto;
    upOutOf;
    downOutOf;
    bubble = iterator((yield_) => {
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        for (let ancestor = this; ancestor; ancestor = ancestor.parent) {
            if (yield_(ancestor) === false)
                break;
        }
        return this;
    });
    postOrder = iterator((yield_) => {
        (function recurse(descendant) {
            descendant.eachChild(recurse);
            yield_(descendant);
            return true;
        })(this);
        return this;
    });
    constructor() {
        this.id = TNode.uniqueNodeId();
        TNode.byId.set(this.id, this);
    }
    dispose() {
        TNode.byId.delete(this.id);
    }
    toString() {
        return `{{ MathQuill TNode #${this.id.toString()} }}`;
    }
    addToElements(el) {
        this.elements.add(el);
    }
    domify(vNode) {
        // Convert html string to DOM contents and add elements to all corresponding TNodes.
        const localVNode = vNode instanceof VNode ? vNode : new VNode(this.html());
        const addToElements = (el) => {
            if (el instanceof HTMLElement) {
                const cmdId = parseInt(el.getAttribute(mqCmdId) ?? '0');
                const blockId = parseInt(el.getAttribute(mqBlockId) ?? '0');
                if (cmdId)
                    TNode.byId.get(cmdId)?.addToElements(el);
                if (blockId)
                    TNode.byId.get(blockId)?.addToElements(el);
            }
            for (let child = el.firstChild; child; child = child.nextSibling) {
                addToElements(child);
            }
        };
        localVNode.contents.forEach((element) => {
            addToElements(element);
        });
        return localVNode;
    }
    createDir(dir, cursor) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        this.domify();
        this.elements.insDirOf(dir, cursor.element);
        if (cursor.parent)
            cursor[dir] = this.adopt(cursor.parent, cursor.left, cursor.right);
        return this;
    }
    createLeftOf(el) {
        this.createDir('left', el);
    }
    selectChildren(leftEnd, rightEnd) {
        return new Selection(leftEnd, rightEnd);
    }
    isEmpty() {
        return !this.ends.left && !this.ends.right;
    }
    isStyleBlock() {
        return false;
    }
    children() {
        return new Fragment(this.ends.left, this.ends.right);
    }
    eachChild(method, order) {
        const children = this.children();
        children.each(method, order);
        return this;
    }
    foldChildren(fold, fn) {
        return this.children().fold(fold, fn);
    }
    withDirAdopt(dir, parent, withDir, oppDir) {
        new Fragment(this, this).withDirAdopt(dir, parent, withDir, oppDir);
        return this;
    }
    adopt(parent, leftward, rightward) {
        new Fragment(this, this).adopt(parent, leftward, rightward);
        return this;
    }
    disown() {
        new Fragment(this, this).disown();
        return this;
    }
    remove() {
        this.elements.remove();
        this.postOrder('dispose');
        return this.disown();
    }
    // Methods that deal with the browser DOM events from interaction with the typist.
    keystroke(key, e, ctrlr) {
        const cursor = ctrlr.cursor;
        switch (key) {
            case 'Ctrl-Shift-Backspace':
            case 'Ctrl-Backspace':
                ctrlr.ctrlDeleteDir('left');
                break;
            case 'Shift-Backspace':
            case 'Backspace':
                ctrlr.backspace();
                break;
            // Esc -> go one block right if it exists, else escape right.
            case 'Escape':
                ctrlr.escapeDir('right', key, e);
                return;
            // Shift-Escape -> go one block left if it exists, else escape left.
            case 'Shift-Escape':
                ctrlr.escapeDir('left', key, e);
                return;
            // End -> move to the end of the current block.
            case 'End':
                if (cursor.parent) {
                    ctrlr.notify('move').cursor.insAtRightEnd(cursor.parent);
                    ctrlr.aria.queue('end of').queue(cursor.parent, true);
                }
                break;
            // Ctrl-End -> move all the way to the end of the root block.
            case 'Ctrl-End':
                ctrlr.notify('move').cursor.insAtRightEnd(ctrlr.root);
                ctrlr.aria.queue('end of').queue(ctrlr.ariaLabel).queue(ctrlr.root).queue(ctrlr.ariaPostLabel);
                break;
            // Shift-End -> select to the end of the current block.
            case 'Shift-End':
                ctrlr.selectToBlockEndInDir('right');
                break;
            // Ctrl-Shift-End -> select to the end of the root block.
            case 'Ctrl-Shift-End':
                ctrlr.selectToRootEndInDir('right');
                break;
            // Home -> move to the start of the root block or the current block.
            case 'Home':
                if (cursor.parent) {
                    ctrlr.notify('move').cursor.insAtLeftEnd(cursor.parent);
                    ctrlr.aria.queue('beginning of').queue(cursor.parent, true);
                }
                break;
            // Ctrl-Home -> move to the start of the current block.
            case 'Ctrl-Home':
                ctrlr.notify('move').cursor.insAtLeftEnd(ctrlr.root);
                ctrlr.aria.queue('beginning of').queue(ctrlr.ariaLabel).queue(ctrlr.root).queue(ctrlr.ariaPostLabel);
                break;
            // Shift-Home -> select to the start of the current block.
            case 'Shift-Home':
                ctrlr.selectToBlockEndInDir('left');
                break;
            // Ctrl-Shift-Home -> move to the start of the root block.
            case 'Ctrl-Shift-Home':
                ctrlr.selectToRootEndInDir('left');
                break;
            case 'Left':
                ctrlr.moveLeft();
                break;
            case 'Shift-Left':
                ctrlr.selectLeft();
                break;
            case 'Ctrl-Left':
                break;
            case 'Right':
                ctrlr.moveRight();
                break;
            case 'Shift-Right':
                ctrlr.selectRight();
                break;
            case 'Ctrl-Right':
                break;
            case 'Up':
                ctrlr.moveUp();
                break;
            case 'Down':
                ctrlr.moveDown();
                break;
            case 'Shift-Up':
                ctrlr.withIncrementalSelection((selectDir) => {
                    if (cursor.left) {
                        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
                        while (cursor.left)
                            selectDir('left');
                    }
                    else {
                        selectDir('left');
                    }
                });
                break;
            case 'Shift-Down':
                ctrlr.withIncrementalSelection((selectDir) => {
                    if (cursor.right) {
                        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
                        while (cursor.right)
                            selectDir('right');
                    }
                    else {
                        selectDir('right');
                    }
                });
                break;
            case 'Ctrl-Up':
            case 'Ctrl-Down':
                break;
            case 'Ctrl-Shift-Delete':
            case 'Ctrl-Delete':
                ctrlr.ctrlDeleteDir('right');
                break;
            case 'Shift-Delete':
            case 'Delete':
                ctrlr.deleteForward();
                break;
            case 'Meta-A':
            case 'Ctrl-A':
                ctrlr.selectAll();
                break;
            // The remaining key strokes are only of benefit to screen reader users.
            // speak parent block that has focus
            case 'Ctrl-Alt-Up':
                if (cursor.parent?.parent && cursor.parent.parent instanceof TNode)
                    ctrlr.aria.queue(cursor.parent.parent);
                else
                    ctrlr.aria.queue('nothing above');
                break;
            // speak current block that has focus
            case 'Ctrl-Alt-Down':
                if (cursor.parent && cursor.parent instanceof TNode)
                    ctrlr.aria.queue(cursor.parent);
                else
                    ctrlr.aria.queue('block is empty');
                break;
            // speak left-adjacent block
            case 'Ctrl-Alt-Left':
                if (cursor.parent?.parent?.ends.left)
                    ctrlr.aria.queue(cursor.parent.parent.ends.left);
                else
                    ctrlr.aria.queue('nothing to the left');
                break;
            // speak right-adjacent block
            case 'Ctrl-Alt-Right':
                if (cursor.parent?.parent?.ends.right)
                    ctrlr.aria.queue(cursor.parent.parent.ends.right);
                else
                    ctrlr.aria.queue('nothing to the right');
                break;
            // speak selection
            case 'Ctrl-Alt-Shift-Down':
                if (cursor.selection)
                    ctrlr.aria.queue(cursor.selection.join('mathspeak', ' ').trim() + ' selected');
                else
                    ctrlr.aria.queue('nothing selected');
                break;
            // speak ARIA post label (evaluation or error)
            case 'Ctrl-Alt-=':
            case 'Ctrl-Alt-Shift-Right':
                if (ctrlr.ariaPostLabel.length)
                    ctrlr.aria.queue(ctrlr.ariaPostLabel);
                else
                    ctrlr.aria.queue('no answer');
                break;
            default:
                return;
        }
        ctrlr.aria.alert();
        e.preventDefault();
        ctrlr.scrollHoriz();
    }
    html() {
        return '';
    }
    text() {
        return '';
    }
    latex() {
        return '';
    }
    focus() {
        /* do nothing */
    }
    blur(_cursor) {
        /* do nothing */
    }
    seek(_left, _cursor) {
        /* do nothing */
    }
    writeLatex(_cursor, _latex) {
        /* do nothing */
    }
    finalizeInsert(_options, _cursor) {
        /* do nothing */
    }
    write(_cursor, _ch) {
        /* do nothing */
    }
    replaces(_fragment) {
        /* do nothing */
    }
    setOptions(_options) {
        return this;
    }
    chToCmd(_ch, _options) {
        return this;
    }
    mathspeak(_options) {
        return '';
    }
    getController() {
        // Navigate up the tree to find the controller.
        return (function getCursor(node) {
            if (node.controller)
                return node.controller;
            if (node.parent)
                return getCursor(node.parent);
        })(this);
    }
    // called by Controller::escapeDir, moveDir
    moveOutOf(_dir, _cursor, _updown) {
        prayOverridden('moveOutOf');
    }
    // called by Controller::moveDir
    moveTowards(_dir, _cursor, _updown) {
        prayOverridden('moveTowards');
    }
    // called by Controller::deleteDir
    deleteOutOf(_dir, _cursor) {
        prayOverridden('deleteOutOf');
    }
    // called by Controller::deleteDir
    deleteTowards(_dir, _cursor) {
        prayOverridden('deleteTowards');
    }
    // called by Controller::selectDir
    unselectInto(_dir, _cursor) {
        prayOverridden('unselectInto');
    }
    // called by Controller::selectDir
    selectOutOf(_dir, _cursor) {
        prayOverridden('selectOutOf');
    }
    // called by Controller::selectDir
    selectTowards(_dir, _cursor) {
        prayOverridden('selectTowards');
    }
}

// An abstraction layer wrapping the textarea in an object with methods to manipulate and listen to events.
// This is an internal abstraction layer intented to handle cross-browser inconsistencies in event handlers.
const saneKeyboardEvents = (() => {
    const KEY_TO_MQ_VALUE = {
        ArrowRight: 'Right',
        ArrowLeft: 'Left',
        ArrowDown: 'Down',
        ArrowUp: 'Up',
        ' ': 'Spacebar'
    };
    const isLowercaseAlphaCharacter = (s) => s.length === 1 && s >= 'a' && s <= 'z';
    // To the extent possible, create a normalized string representation of the key combo.
    const stringify = (evt) => {
        const key = isLowercaseAlphaCharacter(evt.key) ? evt.key.toUpperCase() : (KEY_TO_MQ_VALUE[evt.key] ?? evt.key);
        const modifiers = [];
        if (evt.ctrlKey)
            modifiers.push('Ctrl');
        if (evt.metaKey)
            modifiers.push('Meta');
        if (evt.altKey)
            modifiers.push('Alt');
        if (evt.shiftKey)
            modifiers.push('Shift');
        if (!modifiers.length)
            return key;
        if (key !== 'Alt' && key !== 'Control' && key !== 'Meta' && key !== 'Shift')
            modifiers.push(key);
        return modifiers.join('-');
    };
    // Create a keyboard events shim that calls callbacks at useful times and exports useful public methods.
    return (textarea, controller) => {
        // Virtual keyboards on touch screen devices send 'Unidentified' for almost all keys in the 'keydown' event. As
        // a result the keystroke handler called in that event handler passes 'Unidentified' for the key.  This makes
        // the enableSpaceNavigation option fail on these devices.  So this flag detects the 'Unidentified' key, and
        // calls the keystroke handler again passing 'Spacebar' for the key, when a space is typed.
        let sendInputSpaceKeystroke = false;
        // Public methods
        const select = (text) => {
            textarea.value = text;
            if (text && textarea instanceof HTMLTextAreaElement)
                textarea.select();
        };
        const handleKey = (key, e) => {
            if (controller.options.overrideKeystroke)
                controller.options.overrideKeystroke(key, e);
            else
                controller.keystroke(key, e);
        };
        // Event handlers
        const onKeydown = (e) => {
            if (e.target !== textarea)
                return;
            sendInputSpaceKeystroke = e.key === 'Unidentified';
            handleKey(stringify(e), e);
        };
        const onPaste = (e) => {
            if (e.target !== textarea)
                return;
            // In Linux, middle-click pasting causes onPaste to be called, when the textarea is not necessarily focused.
            // We focus it here to ensure that the pasted text actually ends up in the textarea.
            // It's pretty nifty that by changing focus in this handler, we can change the target of the default action.
            // (Actually, I don't think this is even needed, and is probably not doing anything.)
            textarea.focus();
            // Get the pasted text from the clipboard data in the event.
            const text = e.clipboardData?.getData('text') ?? '';
            textarea.value = '';
            if (text) {
                if (controller.options.overridePaste)
                    controller.options.overridePaste(text);
                else
                    controller.paste(text);
            }
        };
        const onInput = (e) => {
            if (e.inputType === 'insertFromPaste')
                return;
            if (e.inputType === 'insertLineBreak') {
                controller.typedText('\n');
                return;
            }
            const text = e.data ?? '';
            if (text.length === 1) {
                if (controller.options.overrideTypedText) {
                    controller.options.overrideTypedText(text);
                }
                else {
                    if (text === ' ' &&
                        sendInputSpaceKeystroke &&
                        controller.options.enableSpaceNavigation &&
                        controller.cursor.depth() > 1 &&
                        controller.cursor.left?.ctrlSeq !== ',') {
                        handleKey('Spacebar', e);
                        setTimeout(() => (textarea.value = ''));
                    }
                    else {
                        controller.typedText(text);
                        setTimeout(() => (textarea.value = ''));
                    }
                }
            }
            else if (textarea.value.length > 1)
                textarea.select();
        };
        // Attach event handlers
        textarea.addEventListener('keydown', onKeydown);
        textarea.addEventListener('cut', () => {
            if (controller.options.overrideCut)
                controller.options.overrideCut();
            else
                controller.cut();
        });
        textarea.addEventListener('copy', () => {
            if (controller.options.overrideCopy)
                controller.options.overrideCopy();
            else
                controller.copy();
        });
        textarea.addEventListener('paste', onPaste);
        textarea.addEventListener('input', onInput);
        // Export public methods
        return { select };
    };
})();

// Point base class of edit tree-related objects
class Point {
    parent;
    left;
    right;
    ancestors;
    constructor(parent, leftward, rightward) {
        this.parent = parent;
        this.left = leftward;
        this.right = rightward;
    }
    static copy(pt) {
        return new Point(pt.parent, pt.left, pt.right);
    }
}

const RootBlockMixin = (_) => {
    _.moveOutOf = (dir) => _.controller?.handle('moveOutOf', dir);
    _.deleteOutOf = (dir) => _.controller?.handle('deleteOutOf', dir);
    _.selectOutOf = (dir) => _.controller?.handle('selectOutOf', dir);
    _.upOutOf = () => _.controller?.handle('upOutOf');
    _.downOutOf = () => _.controller?.handle('downOutOf');
    _.reflow = () => _.controller?.handle('edit');
};
// Editability methods called by the cursor for editing, cursor movements, and selection of the MathQuill tree.
// These all take in a direction and the cursor.
// The MathCommand and TextBlock classes use this mixin.
const deleteSelectTowardsMixin = (Base) => class extends Base {
    moveTowards(dir, cursor, updown) {
        const nodeAtEnd = (updown && this[`${updown}Into`]) || this.ends[otherDir(dir)];
        if (nodeAtEnd)
            cursor.insAtDirEnd(otherDir(dir), nodeAtEnd);
        if (cursor.parent)
            cursor.controller.aria.queueDirEndOf(otherDir(dir)).queue(cursor.parent, true);
    }
    deleteTowards(dir, cursor) {
        if (this.isEmpty())
            cursor[dir] = this.remove()[dir];
        else
            this.moveTowards(dir, cursor);
    }
    selectTowards(dir, cursor) {
        cursor[otherDir(dir)] = this;
        cursor[dir] = this[dir];
    }
};
const DelimsMixin = (Base) => class extends Base {
    delims;
    content;
    addToElements(el) {
        super.addToElements(el);
        const children = this.elements.children();
        this.delims = [children.firstElement, children.lastElement];
        this.content = children.contents[1];
    }
};

class Parser {
    // The Parser object is a wrapper for a parser function.
    // Externally, you use one to parse a string by calling
    //   const result = SomeParser.parse('Me Me Me! Parse Me!');
    // You should never call the constructor, rather you should
    // construct your Parser from the base parsers and the
    // parser combinator methods.
    _;
    constructor(body) {
        this._ = body;
    }
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-parameters
    parse(stream) {
        return this.skip(Parser.eof)._(
        // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
        `${stream}`, (_stream, result) => result, (stream, message) => {
            throw new Error(`Parse Error: ${message} at ${stream || 'EOF'}`);
        });
    }
    // Primitive combinators
    or(alternative) {
        if (!(alternative instanceof Parser))
            throw new Error('or is passed a parser');
        return new Parser((stream, onSuccess, onFailure) => this._(stream, onSuccess, () => alternative._(stream, onSuccess, onFailure)));
    }
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-parameters
    then(next) {
        return new Parser((stream, onSuccess, onFailure) => this._(stream, (newStream, result) => {
            const nextParser = next instanceof Parser ? next : typeof next === 'function' ? next(result) : undefined;
            if (!(nextParser instanceof Parser))
                throw new Error('a parser is returned');
            return nextParser._(newStream, onSuccess, onFailure);
        }, onFailure));
    }
    // Optimized iterative combinators
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-parameters
    many() {
        return new Parser((stream, onSuccess) => {
            const xs = [];
            while (this._(stream, (newStream, x) => {
                stream = newStream;
                xs.push(x);
                return true;
            }, () => false))
                ;
            return onSuccess(stream, xs);
        });
    }
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-parameters
    times(min, max = min) {
        return new Parser((stream, onSuccess, onFailure) => {
            const xs = [];
            let result = true;
            for (let i = 0; i < max && result; ++i) {
                let failure;
                result = this._(stream, (newStream, x) => {
                    xs.push(x);
                    stream = newStream;
                    return true;
                }, (newStream, msg) => {
                    failure = msg;
                    stream = newStream;
                    return false;
                });
                if (i < min && !result)
                    return onFailure(stream, failure);
            }
            return onSuccess(stream, xs);
        });
    }
    // Higher-level combinators
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-parameters
    result(res) {
        return this.then(Parser.succeed(res));
    }
    atMost(n) {
        return this.times(0, n);
    }
    atLeast(n) {
        return this.times(n).then((start) => this.many().map((end) => start.concat(end)));
    }
    map(fn) {
        return this.then((result) => {
            if (typeof fn === 'function' && /^\s*class\s+/.test(fn.toString())) {
                return Parser.succeed(new fn(result));
            }
            else if (typeof fn === 'function') {
                return Parser.succeed(fn(result));
            }
            return Parser.fail('Invalid map function');
        });
    }
    skip(two) {
        return this.then((result) => two.result(result));
    }
    // Primitive parsers
    static string = (str) => {
        return new Parser((stream, onSuccess, onFailure) => {
            const head = stream.slice(0, str.length);
            if (head === str)
                return onSuccess(stream.slice(str.length), head);
            else
                return onFailure(stream, `expected '${str}'`);
        });
    };
    static regex = (re) => {
        if (re.toString().charAt(1) !== '^')
            throw new Error('regexp parser is anchored');
        return new Parser((stream, onSuccess, onFailure) => {
            const match = re.exec(stream);
            if (match)
                return onSuccess(stream.slice(match[0].length), match[0]);
            else
                return onFailure(stream, `expected ${re.toString()}`);
        });
    };
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-parameters
    static succeed = (result) => new Parser((stream, onSuccess) => onSuccess(stream, result));
    static fail = (msg) => new Parser((stream, _, onFailure) => onFailure(stream, msg));
    static letter = Parser.regex(/^[a-z]/i);
    static letters = Parser.regex(/^[a-z]*/i);
    static digit = Parser.regex(/^[0-9]/);
    static digits = Parser.regex(/^[0-9]*/);
    static whitespace = Parser.regex(/^\s+/);
    static optWhitespace = Parser.regex(/^\s*/);
    static any = new Parser((stream, onSuccess, onFailure) => {
        if (!stream)
            return onFailure(stream, 'expected any character');
        return onSuccess(stream.slice(1), stream.charAt(0));
    });
    static all = new Parser((stream, onSuccess) => onSuccess('', stream));
    static eof = new Parser((stream, onSuccess, onFailure) => {
        if (stream)
            return onFailure(stream, 'expected EOF');
        return onSuccess(stream, stream);
    });
}

// Focus and Blur events
const FocusBlurEvents = (Base) => class extends Base {
    focusHandler;
    blurHandler;
    windowFocusHandler;
    windowBlurHandler;
    windowBlurred = !document.hasFocus();
    blurredWithCursor = false;
    addEditableFocusBlurEvents() {
        const blink = this.cursor.blink;
        this.focusHandler = () => {
            this.blurred = false;
            this.updateMathspeak();
            this.container.classList.add('mq-focused');
            this.cursor.blink = blink;
            if (!this.cursor.parent)
                this.cursor.insAtRightEnd(this.root);
            if (this.cursor.selection) {
                this.cursor.selection.elements.removeClass('mq-blur');
                this.selectionChanged();
            }
            else if (!this.blurredWithCursor && !this.windowBlurred)
                this.selectAll();
            else
                this.cursor.show();
        };
        this.textarea?.addEventListener('focus', this.focusHandler);
        this.blurHandler = (e) => {
            this.blurred = true;
            this.blurredWithCursor = this.options.blurWithCursor?.(e, this.apiClass) ?? false;
            this.container.classList.remove('mq-focused');
            if (this.blurredWithCursor) {
                if (this.cursor.selection)
                    this.cursor.selection.elements.addClass('mq-blur');
                else {
                    // FIXME: This special blink method is a bit of a hack. When focus is regained with a mouse
                    // click, then the mouse down handler caches the cursor blink method which will be the method
                    // defined below. Then the focus handler above is called and resets the cursor blink method to
                    // the default cursor blink method cached here at initialization. But then the mouse up handler
                    // occurs, and sets the cursor blink method back to its cached blink method which is this one.
                    // So if the noop is used here, then that is what ends up being in effect, and that means no
                    // cursor blink. The problem is that the mouse handlers changing the blink method is also a
                    // hack. Instead the cursor show method should accept an optional argument that determines if
                    // the cursor is blinking or not. That will take some work to implement and get right for all of
                    // the places that show is called.
                    this.cursor.blink = () => {
                        if (!this.blurred)
                            blink();
                    };
                    this.cursor.show();
                }
            }
            else {
                if (document.hasFocus()) {
                    this.cursor.hide().parent?.blur(this.cursor);
                    this.cursor.clearSelection().endSelection();
                }
                else {
                    this.cursor.hide().parent?.blur();
                    if (this.cursor.selection)
                        this.cursor.selection.elements.addClass('mq-blur');
                }
            }
            this.updateMathspeak(true);
        };
        this.textarea?.addEventListener('blur', this.blurHandler);
        this.windowFocusHandler = () => {
            if (!this.blurredWithCursor && this.blurred && document.activeElement !== this.textarea)
                this.cursor.clearSelection().endSelection();
            // A timeout is used to delay setting this.windowBlurred to false until after the focus handler has run.
            setTimeout(() => (this.windowBlurred = false));
        };
        window.addEventListener('focus', this.windowFocusHandler);
        this.windowBlurHandler = () => (this.windowBlurred = true);
        window.addEventListener('blur', this.windowBlurHandler);
        this.blurred = true;
        this.cursor.hide().parent?.blur();
    }
    unbindEditableFocusBlurEvents() {
        if (this.focusHandler)
            this.textarea?.removeEventListener('focus', this.focusHandler);
        if (this.blurHandler)
            this.textarea?.removeEventListener('blur', this.blurHandler);
        if (this.windowFocusHandler)
            window.removeEventListener('blur', this.windowFocusHandler);
        if (this.windowBlurHandler)
            window.removeEventListener('blur', this.windowBlurHandler);
        delete this.focusHandler;
        delete this.blurHandler;
        delete this.windowFocusHandler;
        delete this.windowBlurHandler;
    }
    addStaticFocusBlurEvents() {
        this.focusHandler = () => {
            if (!this.cursor.selection)
                this.selectAll();
            this.cursor.selection?.elements.removeClass('mq-blur');
            this.blurred = false;
        };
        this.textarea?.addEventListener('focus', this.focusHandler);
        this.blurHandler = () => {
            if (document.hasFocus()) {
                this.cursor.selection?.clear();
                this.cursor.clearSelection();
            }
            else if (this.cursor.selection)
                this.cursor.selection.elements.addClass('mq-blur');
            this.updateMathspeak(true);
        };
        this.textarea?.addEventListener('blur', this.blurHandler);
        this.windowFocusHandler = () => {
            if (document.activeElement !== this.textarea)
                this.cursor.clearSelection().endSelection();
        };
        window.addEventListener('focus', this.windowFocusHandler);
        this.blurred = true;
        this.cursor.hide().parent?.blur();
    }
};
const BlockFocusBlur = (Base) => class extends Base {
    focus() {
        this.elements.addClass('mq-has-cursor');
        this.elements.removeClass('mq-empty');
    }
    blur() {
        this.elements.removeClass('mq-has-cursor');
        if (this.isEmpty())
            this.elements.addClass('mq-empty');
    }
};

// Abstract classes of math blocks and commands.
// Math tree node base class.
// Some math-tree-specific extensions to TNode.
// Both MathBlock's and MathCommand's descend from it.
class MathElement extends TNode {
    finalizeInsert(options, cursor) {
        this.postOrder('finalizeTree', options);
        this.postOrder('contactWeld', cursor);
        // note: this order is important.
        // empty elements need the empty box provided by blur to
        // be present in order for their dimensions to be measured
        // correctly by 'reflow' handlers.
        this.postOrder('blur');
        this.postOrder('reflow');
        this.right?.siblingCreated?.(options, 'left');
        this.left?.siblingCreated?.(options, 'right');
        this.bubble('reflow');
    }
    // If the maxDepth option is set, make sure
    // deeply nested content is truncated. Just return
    // false if the cursor is already too deep.
    prepareInsertionAt(cursor) {
        const maxDepth = cursor.options.maxDepth;
        if (maxDepth !== undefined) {
            const cursorDepth = cursor.depth();
            if (cursorDepth > maxDepth) {
                return false;
            }
            this.removeNodesDeeperThan(maxDepth - cursorDepth);
        }
        return true;
    }
    // Remove nodes that are more than `cutoff`
    // blocks deep from this node.
    removeNodesDeeperThan(cutoff) {
        let depth = 0;
        const queue = [[this, depth]];
        // Do a breadth-first search of this node's descendants
        // down to cutoff, removing anything deeper.
        while (queue.length) {
            const current = queue.shift();
            current?.[0].children().each((child) => {
                const i = child instanceof MathBlock ? 1 : 0;
                depth = current[1] + i;
                if (depth <= cutoff) {
                    queue.push([child, depth]);
                }
                else {
                    (i ? child.children() : child).remove();
                }
                return true;
            });
        }
    }
}
// Commands and operators, like subscripts, exponents, or fractions.
// Descendant commands are organized into blocks.
class MathCommand extends deleteSelectTowardsMixin(MathElement) {
    blocks;
    contentIndex = 0;
    htmlTemplate;
    textTemplate;
    mathspeakTemplate = [''];
    replacedFragment;
    constructor(ctrlSeq, htmlTemplate, textTemplate) {
        super();
        this.blocks = [];
        this.ctrlSeq = ctrlSeq ?? '';
        this.htmlTemplate = htmlTemplate ?? '';
        this.textTemplate = textTemplate ?? [''];
    }
    // obvious methods
    replaces(replacedFragment) {
        replacedFragment?.disown();
        this.replacedFragment = replacedFragment;
    }
    isEmpty() {
        return this.foldChildren(true, (isEmpty, child) => isEmpty && child.isEmpty());
    }
    parser() {
        return latexMathParser.block.times(this.numBlocks()).map((blocks) => {
            this.blocks = blocks;
            for (const block of blocks) {
                block.adopt(this, this.ends.right);
            }
            return this;
        });
    }
    // createLeftOf(cursor) and the methods it calls
    createLeftOf(cursor) {
        const replacedFragment = this.replacedFragment;
        this.createBlocks();
        super.createLeftOf(cursor);
        if (replacedFragment) {
            replacedFragment.adopt(this.blocks[this.contentIndex]);
            this.blocks[this.contentIndex]?.elements.firstElement.append(...replacedFragment.elements.contents);
            this.placeCursor(cursor);
            this.prepareInsertionAt(cursor);
        }
        this.finalizeInsert(cursor.options, cursor);
        this.placeCursor(cursor);
    }
    createBlocks() {
        const numBlocks = this.numBlocks();
        this.blocks = Array(numBlocks);
        for (let i = 0; i < numBlocks; ++i) {
            this.blocks[i] = new MathBlock();
            this.blocks[i].adopt(this, this.ends.right);
        }
    }
    placeCursor(cursor) {
        if (!this.ends.left)
            return;
        // Insert the cursor at the right end of the first empty child, searching from
        // left to right, or if not empty, then to the right end of the child.
        cursor.insAtRightEnd(this.foldChildren(this.ends.left, (leftward, child) => (leftward.isEmpty() ? leftward : child)));
    }
    selectChildren() {
        return new Selection(this, this);
    }
    unselectInto(dir, cursor) {
        cursor.insAtDirEnd(otherDir(dir), cursor.anticursor?.ancestors?.[this.id]);
    }
    seek(pageX, cursor) {
        const getBounds = (node) => {
            const rect = node.elements.firstElement.getBoundingClientRect();
            return { left: rect.left, right: rect.left + rect.width };
        };
        const cmdBounds = getBounds(this);
        if (pageX < cmdBounds.left) {
            cursor.insLeftOf(this);
            return;
        }
        if (pageX > cmdBounds.right) {
            cursor.insRightOf(this);
            return;
        }
        let leftLeftBound = cmdBounds.left;
        this.eachChild((block) => {
            const blockBounds = getBounds(block);
            if (pageX < blockBounds.left) {
                // closer to this block's left bound, or the bound left of that?
                if (pageX - leftLeftBound < blockBounds.left - pageX) {
                    if (block.left)
                        cursor.insAtRightEnd(block.left);
                    else
                        cursor.insLeftOf(this);
                }
                else
                    cursor.insAtLeftEnd(block);
                return false;
            }
            else if (pageX > blockBounds.right) {
                if (block.right)
                    leftLeftBound = blockBounds.right; // continue to next block
                else {
                    // last (rightmost) block
                    // closer to this block's right bound, or the this's right bound?
                    if (cmdBounds.right - pageX < pageX - blockBounds.right) {
                        cursor.insRightOf(this);
                    }
                    else
                        cursor.insAtRightEnd(block);
                }
            }
            else {
                block.seek(pageX, cursor);
                return false;
            }
            return true;
        });
    }
    // methods involved in creating and cross-linking with HTML DOM nodes
    // They all expect an .htmlTemplate like
    //   '<span>&0</span>'
    // or
    //   '<span><span>&0</span><span>&1</span></span>'
    // See html.test.js for more examples.
    // Requirements:
    // - For each block of the command, there must be exactly one "block content
    //   marker" of the form '&<number>' where <number> is the 0-based index of the
    //   block. (Like the LaTeX \newcommand syntax, but with a 0-based rather than
    //   1-based index, because JavaScript because C because Dijkstra.)
    // - The block content marker must be the sole contents of the containing
    //   element, there can't even be surrounding whitespace, or else we can't
    //   guarantee sticking to within the bounds of the block content marker when
    //   mucking with the HTML DOM.
    // - The HTML not only must be well-formed HTML (of course), but also must
    //   conform to the XHTML requirements on tags, specifically all tags must
    //   either be self-closing (like '<br/>') or come in matching pairs.
    //   Close tags are never optional.
    // Note that &<number> isn't well-formed HTML; if you wanted a literal '&123',
    // your HTML template would have to have '&amp;123'.
    numBlocks() {
        const matches = this.htmlTemplate.match(/&\d+/g);
        return matches ? matches.length : 0;
    }
    html() {
        // Render the entire math subtree rooted at this command, as HTML.
        // Expects the blocks for this object to have been created already (either
        // by .createBlocks or the parser), since it uses the .blocks array of child
        // blocks.
        //
        // See html.test.js for example templates and intended outputs.
        //
        // Given an .htmlTemplate as described above,
        // - insert the mathquill-command-id attribute into all top-level tags,
        //   which will be used to set this.elements in .domify().
        //   This is straightforward:
        //     * tokenize into tags and non-tags
        //     * loop through top-level tokens:
        //         * add #cmdId attribute macro to top-level self-closing tags
        //         * else add #cmdId attribute macro to top-level open tags
        //             * skip the matching top-level close tag and all tag pairs
        //               in between
        // - for each block content marker,
        //     + replace it with the contents of the corresponding block,
        //       rendered as HTML
        //     + insert the mathquill-block-id attribute into the containing tag
        //   This is even easier, a quick regex replace, since block tags cannot
        //   contain anything besides the block content marker.
        //
        // Two notes:
        // - The outermost loop through top-level tokens should never encounter any
        //   top-level close tags, because we should have first encountered a
        //   matching top-level open tag, all inner tags should have appeared in
        //   matching pairs and been skipped, and then we should have skipped the
        //   close tag in question.
        // - All open tags should have matching close tags, which means our inner
        //   loop should always encounter a close tag and drop nesting to 0. If
        //   a close tag is missing, the loop will continue until i >= tokens.length
        //   and token becomes undefined. This will loop infinitely, because it
        //   will then TypeError on .slice().
        const blocks = this.blocks;
        const cmdId = ` ${mqCmdId}=${this.id.toString()}`;
        const tokens = this.htmlTemplate.match(/<[^<>]+>|[^<>]+/g);
        if (tokens.join('') !== this.htmlTemplate)
            throw new Error('no unmatched angle brackets');
        // add cmdId to all top-level tags
        for (let i = 0, token = tokens[0]; token; ++i, token = tokens[i]) {
            // top-level self-closing tags
            if (token.endsWith('/>')) {
                tokens[i] = `${token.slice(0, -2)}${cmdId}/>`;
            }
            // top-level open tags
            else if (token.startsWith('<')) {
                if (token.charAt(1) === '/')
                    throw new Error('unmatched top-level close tag');
                tokens[i] = `${token.slice(0, -1)}${cmdId}>`;
                // skip matching top-level close tag and all tag pairs in between
                let nesting = 1;
                do {
                    i += 1;
                    token = tokens[i];
                    if (!token)
                        throw new Error('missing close tags');
                    // close tags
                    if (token.startsWith('</')) {
                        nesting -= 1;
                    }
                    // non-self-closing open tags
                    else if (token.startsWith('<') && !token.endsWith('/>')) {
                        nesting += 1;
                    }
                } while (nesting > 0);
            }
        }
        return tokens
            .join('')
            .replace(/>&(\d+)/g, (_$0, $1) => ` ${mqBlockId}=${blocks[$1]?.id.toString() ?? ''}>${blocks[$1]?.join('html') ?? ''}`);
    }
    // Methods to export a string representation of the math tree
    latex() {
        return this.foldChildren(this.ctrlSeq, (latex, child) => `${latex}{${child.latex() || ' '}}`);
    }
    text() {
        let i = 0;
        return this.foldChildren(this.textTemplate[i], (text, child) => {
            ++i;
            const child_text = child.text();
            if (text && this.textTemplate[i] === '(' && child_text.startsWith('(') && child_text.endsWith(')'))
                return text + child_text.slice(1, -1) + this.textTemplate[i];
            return text + child_text + (this.textTemplate[i] || '');
        });
    }
    mathspeak() {
        let i = 0;
        return this.foldChildren(`${this.mathspeakTemplate[i] || `Start${this.ctrlSeq.replace(/^\\/, '')}`} `, (speech, block) => {
            ++i;
            return `${speech} ${block.mathspeak()} ${this.mathspeakTemplate[i] || `End${this.ctrlSeq.replace(/^\\/, '')}`} `;
        });
    }
}
// Lightweight command without blocks or children.
class Symbol extends MathCommand {
    constructor(ctrlSeq, html, text, mathspeak) {
        const textTemplate = text ? text : ctrlSeq && ctrlSeq.length > 1 ? ctrlSeq.slice(1) : (ctrlSeq ?? '');
        super(ctrlSeq, html, [textTemplate]);
        this.createBlocks = noop;
        this.isSymbol = true;
        this.mathspeakName = mathspeak || text || ctrlSeq?.replace(/^\\/, '');
    }
    parser() {
        return Parser.succeed(this);
    }
    numBlocks() {
        return 0;
    }
    replaces(replacedFragment) {
        replacedFragment?.remove();
    }
    moveTowards(dir, cursor) {
        if (dir === 'left')
            this.elements.first.before(cursor.element);
        else
            this.elements.last.after(cursor.element);
        cursor[otherDir(dir)] = this;
        cursor[dir] = this[dir];
        cursor.controller.aria.queue(this);
    }
    deleteTowards(dir, cursor) {
        cursor[dir] = this.remove()[dir];
    }
    seek(pageX, cursor) {
        // Insert at whichever side the click was closer to.
        const rect = this.elements.firstElement.getBoundingClientRect();
        if (pageX - rect.left < rect.width / 2)
            cursor.insLeftOf(this);
        else
            cursor.insRightOf(this);
    }
    latex() {
        return this.ctrlSeq;
    }
    text() {
        return this.textTemplate.join('');
    }
    placeCursor() {
        /* do nothing */
    }
    isEmpty() {
        return true;
    }
    mathspeak() {
        return this.mathspeakName || '';
    }
}
class VanillaSymbol extends Symbol {
    constructor(ch, html, text, mathspeak) {
        super(ch, `<span>${html || ch}</span>`, text, mathspeak);
    }
}
class BinaryOperator extends Symbol {
    isUnary = false;
    constructor(ctrlSeq, html, text, mathspeak, useRawHtml = false) {
        super(ctrlSeq, useRawHtml ? html : `<span class="mq-binary-operator">${html ?? ''}</span>`, text, mathspeak);
    }
}
class Inequality extends BinaryOperator {
    data;
    strict;
    constructor(data, strict) {
        const strictness = strict ? 'Strict' : '';
        super(data[`ctrlSeq${strictness}`], data[`html${strictness}`], data[`text${strictness}`], data[`mathspeak${strictness}`]);
        this.data = data;
        this.strict = strict;
    }
    swap(strict) {
        this.strict = strict;
        const strictness = strict ? 'Strict' : '';
        this.ctrlSeq = this.data[`ctrlSeq${strictness}`];
        this.elements.html(this.data[`html${strictness}`]);
        this.textTemplate = [this.data[`text${strictness}`]];
        this.mathspeakName = this.data[`mathspeak${strictness}`];
    }
    deleteTowards(dir, cursor) {
        if (dir === 'left' && !this.strict) {
            this.swap(true);
            this.bubble('reflow');
            return;
        }
        super.deleteTowards(dir, cursor);
    }
}
class FactorialOrNEQ extends Inequality {
    constructor(data, strict) {
        super(data, strict);
        this.isUnary = strict;
    }
    swap(strict) {
        this.isUnary = strict;
        this.contactWeld();
        super.swap(strict);
    }
    contactWeld() {
        if (this.isUnary)
            this.elements.firstElement.className = '';
        else
            this.elements.firstElement.className = 'mq-binary-operator';
        return this;
    }
}
class Equality extends BinaryOperator {
    constructor() {
        super('=', '=', '=', 'equals');
    }
    createLeftOf(cursor) {
        if (cursor.left instanceof Inequality && cursor.left.strict) {
            cursor.left.swap(false);
            cursor.left.bubble('reflow');
            return;
        }
        super.createLeftOf(cursor);
    }
}
class Digit extends VanillaSymbol {
    createLeftOf(cursor) {
        if (cursor.options.autoSubscriptNumerals &&
            cursor.parent !== cursor.parent?.parent?.sub &&
            ((cursor.left instanceof Variable && cursor.left.isItalic) ||
                (cursor.left instanceof SupSub && cursor.left.left instanceof Variable && cursor.left.left.isItalic))) {
            new LatexCmds._().createLeftOf(cursor);
            super.createLeftOf(cursor);
            if (cursor.parent?.parent)
                cursor.insRightOf(cursor.parent.parent);
        }
        else
            super.createLeftOf(cursor);
    }
}
class Variable extends Symbol {
    isItalic = false;
    isPartOfOperator = false;
    constructor(ch, html) {
        super(ch, `<var>${html || ch}</var>`);
    }
    text() {
        let text = this.ctrlSeq;
        if (text.startsWith('\\')) {
            if (text.startsWith('\\operatorname{'))
                text = text.slice(14, text.length);
            else
                text = text.slice(1, text.length);
        }
        else if (text.endsWith(' ') || text.endsWith('}')) {
            text = text.slice(0, -1);
            if (!(this.right instanceof Bracket || this.right instanceof Fraction || this.right instanceof SupSub))
                text += ' ';
        }
        return text;
    }
}
class Letter extends Variable {
    letter;
    constructor(ch, htmlTemplate) {
        super(ch, htmlTemplate);
        this.letter = ch;
        this.siblingDeleted = this.siblingCreated = (opts, dir) => {
            this.finalizeTree(opts, dir);
        };
    }
    createLeftOf(cursor) {
        super.createLeftOf(cursor);
        const autoCmds = cursor.options.autoCommands, maxLength = autoCmds._maxLength;
        if (maxLength > 0) {
            // To find the longest possible autocommand, join the longest sequence of letters.
            let str = '', 
            // eslint-disable-next-line @typescript-eslint/no-this-alias
            l = this, i = 0;
            // FIXME: l.ctrlSeq === l.letter checks if first or last in an operator name
            while (l instanceof Letter && l.ctrlSeq === l.letter && i < maxLength) {
                str = l.letter + str;
                l = l.left;
                ++i;
            }
            // Check for an autocommand, going through substrings from longest to shortest.
            while (str.length) {
                if (autoCmds[str]) {
                    // eslint-disable-next-line @typescript-eslint/no-this-alias
                    for (i = 1, l = this; i < str.length; ++i, l = l?.left)
                        ;
                    new Fragment(l, this).remove();
                    cursor.left = l?.left;
                    new LatexCmds[str](str).createLeftOf(cursor);
                    return;
                }
                str = str.slice(1);
            }
        }
    }
    italicize(bool) {
        this.isItalic = bool;
        this.isPartOfOperator = !bool;
        this.elements.toggleClass('mq-operator-name', !bool);
        return this;
    }
    finalizeTree(opts, dir) {
        // don't auto-un-italicize if the sibling to my right changed (dir === 'right' or
        // undefined) and it's now a Letter, it will un-italicize everyone
        if (dir !== 'left' && this.right instanceof Letter)
            return;
        this.autoUnItalicize(opts);
    }
    autoUnItalicize(opts) {
        const autoOps = opts.autoOperatorNames;
        if (autoOps._maxLength === 0)
            return;
        // want longest possible operator names, so join together entire contiguous
        // sequence of letters
        let str = this.letter, l = this.left, r = this.right;
        for (; l instanceof Letter; l = l.left)
            str = l.letter + str;
        for (; r instanceof Letter; r = r.right)
            str += r.letter;
        // removeClass and delete flags from all letters before figuring out
        // which, if any, are part of an operator name
        new Fragment(l?.right || this.parent?.ends.left, r?.left || this.parent?.ends.right).each((el) => {
            el.italicize(true).elements.removeClass('mq-first', 'mq-last', 'mq-followed-by-supsub');
            el.ctrlSeq = el.letter;
            return true;
        });
        // check for operator names: at each position from left to right, check
        // substrings from longest to shortest
        for (let i = 0, first = l?.right || this.parent?.ends.left; i < str.length; ++i, first = first?.right) {
            for (let len = Math.min(autoOps._maxLength, str.length - i); len > 0; --len) {
                const word = str.slice(i, i + len);
                if (autoOps[word]) {
                    let last;
                    for (let j = 0, letter = first; j < len; j += 1, letter = letter?.right) {
                        letter.italicize(false);
                        last = letter;
                    }
                    const isBuiltIn = BuiltInOpNames[word];
                    first.ctrlSeq = (isBuiltIn ? '\\' : '\\operatorname{') + (first?.ctrlSeq ?? '');
                    last.ctrlSeq += isBuiltIn ? ' ' : '}';
                    if (word in TwoWordOpNames)
                        last?.left?.left?.left?.elements.addClass('mq-last');
                    if (!this.shouldOmitPadding(first?.left))
                        first?.elements.addClass('mq-first');
                    if (!this.shouldOmitPadding(last?.right)) {
                        const rightward = last?.right;
                        if (rightward instanceof SupSub)
                            rightward.siblingCreated?.(opts);
                        else
                            last?.elements.toggleClass('mq-last', !(rightward instanceof Bracket));
                    }
                    i += len - 1;
                    first = last;
                    break;
                }
            }
        }
    }
    shouldOmitPadding(node) {
        // omit padding if no node, or if node already has padding (to avoid double-padding)
        return !node || node instanceof BinaryOperator || node instanceof UpperLowerLimitCommand;
    }
}
function insLeftOfMeUnlessAtEnd(cursor) {
    // cursor.insLeftOf(cmd), unless cursor at the end of block, and every
    // ancestor cmd is at the end of every ancestor block
    const cmd = this.parent;
    if (!cmd)
        return;
    let ancestorCmd = cursor;
    do {
        if (ancestorCmd?.right)
            return cursor.insLeftOf(cmd);
        ancestorCmd = ancestorCmd?.parent?.parent;
    } while (ancestorCmd !== cmd);
    cursor.insRightOf(cmd);
}
// This test is used to determine whether an item may be treated as a whole number
// for shortening the verbalized (mathspeak) forms of some fractions and superscripts.
const intRgx = /^[+-]?[\d]+$/;
// Traverses the passed block's children and returns the concatenation of their ctrlSeq properties.
// Used in shortened mathspeak computations as a block's .text() method can be potentially expensive.
const getCtrlSeqsFromBlock = (block) => {
    if (!block)
        return '';
    let chars = '';
    block.eachChild((child) => {
        chars += child.ctrlSeq;
        return true;
    });
    return chars;
};
class Fraction extends MathCommand {
    constructor() {
        super();
        this.ctrlSeq = '\\frac';
        this.htmlTemplate =
            '<span class="mq-fraction mq-non-leaf">' +
                '<span class="mq-numerator">&0</span>' +
                '<span class="mq-denominator">&1</span>' +
                '<span style="display:inline-block;width:0">&#8203;</span>' +
                '</span>';
        this.textTemplate = ['((', ')/(', '))'];
    }
    text() {
        let leftward = this.left;
        for (; leftward && leftward.ctrlSeq === '\\ '; leftward = leftward.left)
            ;
        const text = (dir) => {
            let needParens = false;
            let numBlocks = 0;
            let haveDigits = false;
            let numPeriods = 0;
            this.ends[dir]?.eachChild((child) => {
                if (child instanceof Digit)
                    haveDigits = true;
                if (child instanceof VanillaSymbol && child.ctrlSeq === '.')
                    ++numPeriods;
                if (!(child instanceof Digit ||
                    (child instanceof VanillaSymbol && child.ctrlSeq === '.' && numPeriods < 2 && !numBlocks) ||
                    (child instanceof BinaryOperator && child.isUnary) ||
                    child instanceof SupSub))
                    ++numBlocks;
                if ((haveDigits && numBlocks) ||
                    numBlocks > 1 ||
                    numPeriods > 1 ||
                    (child instanceof BinaryOperator && !child.isUnary) ||
                    ('text' in LatexCmds && child instanceof LatexCmds.text) ||
                    child instanceof UpperLowerLimitCommand ||
                    child instanceof Fraction ||
                    child.ctrlSeq === '\\ ' ||
                    /^[,;:]$/.test(child.ctrlSeq)) {
                    needParens = true;
                }
                return !needParens;
            });
            const blankDefault = dir === 'left' ? '0' : '1';
            const l = this.ends[dir]?.text() !== ' ' && this.ends[dir]?.text();
            return l ? (needParens ? `(${l})` : l) : blankDefault;
        };
        return (leftward instanceof BinaryOperator && leftward.isUnary) ||
            leftward?.elements.hasClass('mq-operator-name') ||
            (leftward instanceof SupSub && leftward.left?.elements.hasClass('mq-operator-name'))
            ? `(${text('left')}/${text('right')})`
            : ` ${text('left')}/${text('right')} `;
    }
    finalizeTree() {
        this.upInto = this.ends.left;
        if (this.ends.right) {
            this.ends.right.upOutOf = this.ends.left;
            this.ends.right.ariaLabel = 'denominator';
        }
        this.downInto = this.ends.right;
        if (this.ends.left) {
            this.ends.left.downOutOf = this.ends.right;
            this.ends.left.ariaLabel = 'numerator';
        }
        const fracDepth = this.getFracDepth();
        if (fracDepth > 2) {
            this.mathspeakTemplate = [
                `StartDepth${fracDepth.toString()}Fraction,`,
                `Depth${fracDepth.toString()}Over`,
                `, EndDepth${fracDepth.toString()}Fraction`
            ];
        }
        else if (fracDepth > 1) {
            this.mathspeakTemplate = ['StartNestedFraction,', 'NestedOver', ', EndNestedFraction'];
        }
        else {
            this.mathspeakTemplate = ['StartFraction,', 'Over', ', EndFraction'];
        }
    }
    mathspeak(opts) {
        if (opts?.createdLeftOf) {
            const cursor = opts.createdLeftOf;
            return cursor.parent?.mathspeak() ?? '';
        }
        const numText = getCtrlSeqsFromBlock(this.ends.left);
        const denText = getCtrlSeqsFromBlock(this.ends.right);
        // Shorten mathspeak value for whole number fractions whose denominator is less than 10.
        if (!opts?.ignoreShorthand && intRgx.test(numText) && intRgx.test(denText)) {
            const isSingular = numText === '1' || numText === '-1';
            let newDenSpeech = '';
            if (denText === '2')
                newDenSpeech = isSingular ? 'half' : 'halves';
            else if (denText === '3')
                newDenSpeech = isSingular ? 'third' : 'thirds';
            else if (denText === '4')
                newDenSpeech = isSingular ? 'fourth' : 'fourths';
            else if (denText === '5')
                newDenSpeech = isSingular ? 'fifth' : 'fifths';
            else if (denText === '6')
                newDenSpeech = isSingular ? 'sixth' : 'sixths';
            else if (denText === '7')
                newDenSpeech = isSingular ? 'seventh' : 'sevenths';
            else if (denText === '8')
                newDenSpeech = isSingular ? 'eighth' : 'eighths';
            else if (denText === '9')
                newDenSpeech = isSingular ? 'ninth' : 'ninths';
            if (newDenSpeech !== '') {
                // Handle the case of an integer followed by a simplified fraction such as 1\frac{1}{2}.  Such
                // combinations should be spoken aloud as "1 and 1 half."
                let precededByInteger = false;
                for (let sibling = this.left; sibling; sibling = sibling.left) {
                    // Ignore whitespace
                    if (sibling.ctrlSeq === '\\ ')
                        continue;
                    else if (intRgx.test(sibling.ctrlSeq))
                        precededByInteger = true;
                    else
                        break;
                }
                return `${precededByInteger ? 'and ' : ''}${this.ends.left?.mathspeak() ?? ''} ${newDenSpeech}`;
            }
        }
        return super.mathspeak();
    }
    getFracDepth() {
        const level = 0;
        const walkUp = function (item, level) {
            if (item instanceof Fraction)
                ++level;
            if (item.parent)
                return walkUp(item.parent, level);
            else
                return level;
        };
        return walkUp(this, level);
    }
}
const supSubText = (prefix, block) => {
    let needParens = false;
    let numBlocks = 0;
    let haveDigits = false;
    let numPeriods = 0;
    block?.eachChild((child) => {
        if (child instanceof Digit)
            haveDigits = true;
        if (child instanceof VanillaSymbol && child.ctrlSeq === '.')
            ++numPeriods;
        if (!(child instanceof Digit ||
            (child instanceof VanillaSymbol && child.ctrlSeq === '.' && numPeriods < 2 && !numBlocks) ||
            (child instanceof BinaryOperator && child.isUnary)))
            ++numBlocks;
        if ((haveDigits && numBlocks) ||
            numBlocks > 1 ||
            numPeriods > 1 ||
            (child instanceof BinaryOperator && !child.isUnary) ||
            ('text' in LatexCmds && child instanceof LatexCmds.text) ||
            child instanceof UpperLowerLimitCommand ||
            child instanceof Fraction ||
            child.ctrlSeq === '\\ ' ||
            /^[,;:]$/.test(child.ctrlSeq)) {
            needParens = true;
        }
        return !needParens;
    });
    const l = block?.text() !== ' ' && block?.text();
    return l ? prefix + (needParens ? `(${l})` : l) : '';
};
class SupSub extends MathCommand {
    supsub = 'sup';
    constructor(_ctrlSeq, htmlTemplate, textTemplate) {
        super('_{...}^{...}', htmlTemplate, textTemplate);
        this.reflow = () => {
            const block = this.elements; // mq-supsub
            const prev = block.first.previousElementSibling;
            // We can't normalize the superscript without having a previous element (which is the base).
            if (!prev)
                return;
            const sup = block.children('.mq-sup').firstElement; // mq-supsub -> mq-sup
            if (sup) {
                const supStyle = getComputedStyle(sup);
                const supRect = sup.getBoundingClientRect();
                const sup_fontsize = parseInt(supStyle.fontSize);
                const sup_bottom = supRect.top + supRect.height - parseFloat(supStyle.paddingTop) - parseFloat(supStyle.paddingBottom);
                // We want the superscript to overlap the top of the base by 0.7 of its font-size.
                // Then small superscripts like x^2 look ok, but big ones like x^(1/2/3) do too.
                const needed = sup_bottom - prev.getBoundingClientRect().top - 0.7 * sup_fontsize;
                const cur_margin = parseInt(supStyle.marginBottom);
                // Lift the superscript up with margin-bottom.
                sup.style.marginBottom = `${(cur_margin + needed).toString()}px`;
            }
        };
        this.siblingCreated = this.siblingDeleted = (options) => {
            if (this.left instanceof Letter &&
                this.left.isPartOfOperator &&
                this.right &&
                !(this.right instanceof BinaryOperator ||
                    this.right instanceof UpperLowerLimitCommand ||
                    this.right instanceof Bracket))
                this.elements.addClass('mq-after-operator-name');
            else
                this.elements.removeClass('mq-after-operator-name');
            this.maybeFlatten(options);
        };
    }
    hasValidBase(options, leftward, parent) {
        return (!options.supSubsRequireOperand ||
            (leftward &&
                leftward.ctrlSeq !== '\\ ' &&
                !(leftward instanceof BinaryOperator) &&
                !/^[,;:]$/.test(leftward.ctrlSeq)) ||
            (!leftward && parent?.parent instanceof MathFunction && parent == parent.parent.blocks[0]));
    }
    createLeftOf(cursor) {
        if (this.hasValidBase(cursor.options, cursor.left, cursor.parent)) {
            // If this SupSub is being placed on a fraction, then add parentheses around the fraction.
            if (cursor.left instanceof Fraction) {
                const brack = new Bracket('right', '(', ')', '(', ')');
                cursor.selection = cursor.left.selectChildren();
                brack.replaces(cursor.replaceSelection());
                brack.createLeftOf(cursor);
            }
            super.createLeftOf(cursor);
            return;
        }
        if (this.replacedFragment) {
            if (cursor.parent)
                this.replacedFragment.adopt(cursor.parent, cursor.left, cursor.right);
            cursor.left = this.replacedFragment.ends.right;
        }
    }
    contactWeld(cursor) {
        // Look on either side for a SupSub.  If one is found compare this .sub, .sup with its .sub, .sup.  If this has
        // one that it doesn't, then call .addBlock() on it with this block.  If this has one that it also has, then
        // insert this block's children into its block, unless this block has none, in which case insert the cursor into
        // its block (and not this one, since this one will be removed).  If something to the left has been deleted,
        // then this SupSub will be merged with one to the left if it is left next to it after the deletion.
        for (const dir of ['left', 'right']) {
            if (this[dir] instanceof SupSub) {
                let pt;
                for (const supsub of ['sub', 'sup']) {
                    const src = this[supsub], dest = this[dir][supsub];
                    if (!src)
                        continue;
                    if (!dest)
                        this[dir].addBlock(src.disown());
                    else if (!src.isEmpty()) {
                        // Insert src children at -dir end of dest
                        src.elements.children().insAtDirEnd(otherDir(dir), dest.elements);
                        const children = src.children().disown();
                        pt = new Point(dest, children.ends.right, dest.ends.left);
                        if (dir === 'left')
                            children.adopt(dest, dest.ends.right);
                        else
                            children.adopt(dest, undefined, dest.ends.left);
                    }
                    else
                        pt = new Point(dest, undefined, dest.ends.left);
                    this.placeCursor = (cursor) => cursor.insAtDirEnd(otherDir(dir), dest || src);
                }
                this.remove();
                if (cursor) {
                    if (cursor.left === this) {
                        if (dir === 'right' && pt) {
                            if (pt.left)
                                cursor.insRightOf(pt.left);
                            else if (pt.parent)
                                cursor.insAtLeftEnd(pt.parent);
                        }
                        else
                            cursor.insRightOf(this[dir]);
                    }
                    else {
                        if (pt?.right)
                            cursor.insRightOf(pt.right);
                    }
                }
                return;
            }
        }
        if (cursor)
            this.maybeFlatten(cursor.options);
        // Only allow deletion of a sup or sub in a function supsub block when it is empty.
        // This sets that up when the first sup or sub is created.
        if (this.parent?.parent instanceof MathFunction &&
            this.parent == this.parent.parent.blocks[0] &&
            !(this.sup && this.sub)) {
            const fcnBlock = this.parent.parent.blocks[0];
            for (const supsub of ['sub', 'sup']) {
                const src = this[supsub];
                if (!src)
                    continue;
                src.deleteOutOf = (dir, cursor) => {
                    if (src.isEmpty())
                        this.remove();
                    cursor.insAtDirEnd(dir, fcnBlock);
                };
            }
        }
    }
    finalizeTree() {
        if (this.ends.left)
            this.ends.left.isSupSubLeft = true;
    }
    // Check to see if this has an invalid base.  If so bring the children out, and remove this SupSub.
    maybeFlatten(options) {
        const cursor = this.getController()?.cursor;
        const leftward = cursor?.right === this ? cursor.left : this.left;
        if (!this.hasValidBase(options, leftward, this.parent) || leftward instanceof Fraction) {
            for (const supsub of ['sub', 'sup']) {
                const src = this[supsub];
                if (!src)
                    continue;
                if (this.parent)
                    src.children()
                        .disown()
                        .adopt(this.parent, this.left, this)
                        .elements.insDirOf('left', this.elements);
            }
            this.remove();
            if (leftward === cursor?.left) {
                if (cursor?.left?.right)
                    cursor.insLeftOf(cursor.left.right);
                else if (cursor?.parent)
                    cursor.insAtDirEnd('left', cursor.parent);
            }
        }
    }
    moveTowards(dir, cursor, updown) {
        if (cursor.options.autoSubscriptNumerals && !this.sup)
            cursor.insDirOf(dir, this);
        else
            super.moveTowards(dir, cursor, updown);
    }
    deleteTowards(dir, cursor) {
        if (cursor.options.autoSubscriptNumerals && this.sub) {
            const cmd = this.sub.ends[otherDir(dir)];
            if (cmd instanceof Symbol)
                cmd.remove();
            else if (cmd)
                cmd.deleteTowards(dir, cursor.insAtDirEnd(otherDir(dir), this.sub));
            // TODO: factor out a .removeBlock() or something
            if (this.sub.isEmpty()) {
                this.sub.deleteOutOf('left', cursor.insAtLeftEnd(this.sub));
                if (this.sup)
                    cursor.insDirOf(otherDir(dir), this);
                // Note `-dir` because in e.g. x_1^2| want backspacing (leftward)
                // to delete the 1 but to end up rightward of x^2; with non-negated
                // `dir` (try it), the cursor appears to have gone "through" the ^2.
            }
        }
        else
            super.deleteTowards(dir, cursor);
    }
    latex() {
        const latex = (prefix, block) => {
            const l = block?.latex();
            return block ? prefix + (l?.length === 1 ? l : `{${l || ' '}}`) : '';
        };
        return latex('_', this.sub) + latex('^', this.sup);
    }
    text() {
        const mainText = supSubText('_', this.sub) + supSubText('^', this.sup);
        return mainText + (mainText && this.right instanceof Digit ? ' ' : '');
    }
    addBlock(block) {
        if (this.supsub === 'sub') {
            this.sup = this.upInto = block;
            if (this.sub)
                this.sub.upOutOf = block;
            block.adopt(this, this.sub).downOutOf = this.sub;
            const blockEl = document.createElement('span');
            blockEl.classList.add('mq-sup');
            blockEl.append(...block.elements.children().contents);
            blockEl.setAttribute(mqBlockId, block.id.toString());
            this.elements.firstElement.prepend(blockEl);
            block.elements = new VNode(blockEl);
        }
        else {
            this.sub = this.downInto = block;
            if (this.sup)
                this.sup.downOutOf = block;
            block.adopt(this, undefined, this.sup).upOutOf = this.sup;
            const blockEl = document.createElement('span');
            blockEl.classList.add('mq-sub');
            blockEl.append(...block.elements.children().contents);
            blockEl.setAttribute(mqBlockId, block.id.toString());
            this.elements.removeClass('mq-sup-only');
            this.elements.firstElement.append(blockEl);
            block.elements = new VNode(blockEl);
            const span = document.createElement('span');
            span.style.display = 'inline-block';
            span.style.width = '0';
            span.textContent = '\u200B';
            this.elements.firstElement.append(span);
        }
        for (const supsub of ['sub', 'sup']) {
            const oppositeSupsub = supsub === 'sub' ? 'sup' : 'sub';
            const updown = supsub === 'sub' ? 'down' : 'up';
            const thisSupsub = this[supsub];
            if (this.parent?.parent instanceof MathFunction && this.parent === this.parent.parent.blocks[0]) {
                // Only allow deletion of a sup or sub in a function supsub block when it is empty.
                // This sets that up when the second sup or sub is created (or when parsing and both exist).
                thisSupsub.deleteOutOf = (dir, cursor) => {
                    if (thisSupsub.isEmpty()) {
                        cursor[otherDir(dir)] = thisSupsub.ends[dir];
                        this.supsub = oppositeSupsub;
                        if (supsub === 'sup')
                            delete this.sup;
                        else
                            delete this.sub;
                        if (updown === 'up')
                            delete this.upInto;
                        else
                            delete this.downInto;
                        const remainingSupsub = this[oppositeSupsub];
                        remainingSupsub[`${updown}OutOf`] = insLeftOfMeUnlessAtEnd;
                        remainingSupsub.deleteOutOf = (dir, cursor) => {
                            if (remainingSupsub.isEmpty())
                                this.remove();
                            if (this.parent)
                                cursor.insAtDirEnd(dir, this.parent);
                        };
                        if (supsub === 'sub') {
                            this.elements.addClass('mq-sup-only');
                            this.elements.first.lastChild?.remove();
                        }
                        thisSupsub.remove();
                    }
                    if (thisSupsub.parent)
                        cursor.insDirOf(thisSupsub[dir] ? otherDir(dir) : dir, thisSupsub.parent);
                };
            }
            else {
                thisSupsub.deleteOutOf = (dir, cursor) => {
                    if (thisSupsub.parent)
                        cursor.insDirOf(thisSupsub[dir] ? otherDir(dir) : dir, thisSupsub.parent);
                    if (!thisSupsub.isEmpty()) {
                        const end = thisSupsub.ends[dir];
                        if (cursor.parent) {
                            thisSupsub
                                .children()
                                .disown()
                                .withDirAdopt(dir, cursor.parent, cursor[dir], cursor[otherDir(dir)])
                                .elements.insDirOf(otherDir(dir), cursor.element);
                        }
                        cursor[otherDir(dir)] = end;
                    }
                    this.supsub = oppositeSupsub;
                    if (supsub === 'sup')
                        delete this.sup;
                    else
                        delete this.sub;
                    if (updown === 'up')
                        delete this.upInto;
                    else
                        delete this.downInto;
                    if (this[oppositeSupsub])
                        this[oppositeSupsub][`${updown}OutOf`] = insLeftOfMeUnlessAtEnd;
                    delete this[oppositeSupsub].deleteOutOf;
                    if (supsub === 'sub') {
                        this.elements.addClass('mq-sup-only');
                        this.elements.first.lastChild?.remove();
                    }
                    thisSupsub.remove();
                };
            }
        }
    }
}
class UpperLowerLimitCommand extends MathCommand {
    latex() {
        const simplify = (latex) => (latex.length === 1 ? latex : `{${latex || ' '}}`);
        return `${this.ctrlSeq}_${simplify(this.ends.left?.latex() ?? '')}^${simplify(this.ends.right?.latex() ?? '')}`;
    }
    text() {
        const operand = this.ctrlSeq.slice(1, this.ctrlSeq.length - 1);
        return `${operand}(${this.ends.left?.text() ?? ''},${this.ends.right?.text() ?? ''})`;
    }
    mathspeak() {
        return `Start ${this.ariaLabel ?? this.ctrlSeq.replace(/^\\/, '')} from ${this.ends.left?.mathspeak() ?? ''} to ${this.ends.right?.mathspeak() ?? ''}, End ${this.ariaLabel ?? this.ctrlSeq.replace(/^\\/, '')}, `;
    }
    parser() {
        const blocks = (this.blocks = [new MathBlock(), new MathBlock()]);
        for (const block of blocks) {
            block.adopt(this, this.ends.right);
        }
        return Parser.optWhitespace
            .then(Parser.string('_').or(Parser.string('^')))
            .then((supOrSub) => {
            const child = blocks[supOrSub === '_' ? 0 : 1];
            return latexMathParser.block.then((block) => {
                block.children().adopt(child, child.ends.right);
                return Parser.succeed(this);
            });
        })
            .many()
            .result(this);
    }
    finalizeTree() {
        this.downInto = this.ends.left;
        this.upInto = this.ends.right;
        if (this.ends.left) {
            this.ends.left.upOutOf = this.ends.right;
            this.ends.left.ariaLabel = 'lower bound';
        }
        if (this.ends.right) {
            this.ends.right.downOutOf = this.ends.left;
            this.ends.right.ariaLabel = 'upper bound';
        }
    }
}
const BracketMixin = (Base) => class extends DelimsMixin(Base) {
    side = 'left';
    sides = { left: { ch: '(', ctrlSeq: '(' }, right: { ch: ')', ctrlSeq: ')' } };
    inserted = false;
    matchBrack(opts, expectedSide, node) {
        // Return node iff it's a matching 1-sided bracket of expected side (if any).
        // A function is only allowed to match an ending parentheses.
        return (((node instanceof Bracket &&
            (!(this instanceof MathFunction) || (!!node.side && node.sides[node.side].ch) === ')')) ||
            (node instanceof MathFunction && !!this.side && this.sides[this.side].ch === ')')) &&
            !!node.side &&
            node.side !== (expectedSide === 'left' ? 'right' : expectedSide === 'right' ? 'left' : 0) &&
            (!opts.restrictMismatchedBrackets ||
                (this.side && OPP_BRACKS[this.sides[this.side].ch] === node.sides[node.side].ch) ||
                { '(': ']', '[': ')' }[this.sides.left.ch] === node.sides.right.ch) &&
            node);
    }
    closeOpposing(brack) {
        delete brack.side;
        if (!this.side)
            return;
        // Copy this objects info to brack as this may be a different type of bracket (like (a, b]).
        brack.sides[this.side] = this.sides[this.side];
        const delim = brack.delims?.[this.side === 'left' ? 0 : 1];
        if (delim) {
            delim.classList.remove('mq-ghost');
            this.replaceBracket(delim, this.side);
        }
    }
    createLeftOf(cursor) {
        let brack, side;
        if (!this.replacedFragment) {
            // If a selection is not being wrapped in this bracket, check to see if this
            // bracket is next to or inside an opposing one-sided bracket.
            const opts = cursor.options;
            if (this.sides.left.ch === '|') {
                // Check both sides if this is an absolute value bracket.
                brack =
                    this.matchBrack(opts, 'right', cursor.right) ||
                        this.matchBrack(opts, 'left', cursor.left) ||
                        this.matchBrack(opts, undefined, cursor.parent?.parent);
            }
            else {
                const otherSide = otherDir(this.side);
                brack =
                    this.matchBrack(opts, otherSide, cursor[otherSide]) ||
                        this.matchBrack(opts, otherSide, cursor.parent?.parent);
            }
        }
        if (brack) {
            // brack may be an absolute value with .side not yet set
            side = this.side = otherDir(brack.side);
            // Move the stuff between this bracket and the ghost of the other bracket outside.
            if (brack === cursor.parent?.parent && cursor[side] && brack.parent) {
                new Fragment(cursor[side], cursor.parent.ends[side], otherDir(side))
                    .disown()
                    .withDirAdopt(otherDir(side), brack.parent, brack, brack[side])
                    .elements.insDirOf(side, brack.elements);
            }
            if (this instanceof MathFunction && side === 'left') {
                // If a math function is typed between a ghosted left parenthesis and a solid right parenthesis,
                // then adopt its contents and replace the parentheses.
                const rightward = brack === cursor.parent?.parent && cursor.right !== brack.ends.right?.right
                    ? new Fragment(cursor.right, brack.ends.right?.ends.right, 'left').disown()
                    : undefined;
                cursor.insRightOf(brack);
                delete this.side;
                if (rightward)
                    this.replaces(rightward);
                super.createLeftOf(cursor);
                brack.remove();
                this.bubble('reflow');
                return;
            }
            else {
                this.closeOpposing(brack);
                brack.bubble('reflow');
            }
        }
        else {
            // If a math function is typed to the left of a parentheses and a selection is not being wrapped,
            // then adopt then contents of the parentheses and remove them.
            if (this instanceof MathFunction &&
                !this.replacedFragment &&
                cursor.right instanceof Bracket &&
                cursor.right.sides.left.ch === '(' &&
                cursor.right.sides.right.ch === ')') {
                const paren = cursor.right;
                this.side = paren.side;
                this.replaces(new Fragment(paren.ends.left?.ends.left, paren.ends.right?.ends.right, 'left'));
                super.createLeftOf(cursor);
                paren.remove();
                this.bubble('reflow');
                return;
            }
            // eslint-disable-next-line @typescript-eslint/no-this-alias
            brack = this;
            side = brack.side;
            // If wrapping a selection, don't be one-sided.
            if (brack.replacedFragment)
                delete brack.side;
            else if (cursor[otherDir(side)]) {
                // Auto-expand so the ghost is at the far end.
                brack.replaces(new Fragment(cursor[otherDir(side)], cursor.parent?.ends[otherDir(side)], side));
                if (side === 'left')
                    delete cursor.right;
                else
                    delete cursor.left;
            }
            super.createLeftOf(cursor);
        }
        if (side === 'left' && brack.ends.left)
            cursor.insAtLeftEnd(brack.ends.left);
        else
            cursor.insRightOf(brack);
    }
    unwrap() {
        const node = this.parent
            ? this.blocks[this.contentIndex].children().disown().adopt(this.parent, this, this.right)
            : undefined;
        if (node)
            this.elements.last.after(...node.elements.contents);
        this.remove();
    }
    deleteSide(side, outward, cursor) {
        const parent = this.parent, sib = this[side], farEnd = parent?.ends[side];
        if (side === this.side) {
            // If deleting a non-ghost of a one-sided bracket, then unwrap the bracket.
            this.unwrap();
            if (sib)
                cursor.insDirOf(otherDir(side), sib);
            else if (parent)
                cursor.insAtDirEnd(side, parent);
            return;
        }
        const wasSolid = !this.side;
        this.side = otherDir(side);
        // If deleting a like, outer close-brace of [(1+2)+3} where the inner open-paren
        // is a ghost then become [1+2)+3.
        if (this.matchBrack(cursor.options, side, this.blocks[this.contentIndex].ends[this.side])) {
            this.closeOpposing(this.blocks[this.contentIndex].ends[this.side]);
            const origEnd = this.blocks[this.contentIndex].ends[side];
            this.unwrap();
            origEnd?.siblingCreated?.(cursor.options, side);
            if (sib)
                cursor.insDirOf(otherDir(side), sib);
            else if (parent)
                cursor.insAtDirEnd(side, parent);
        }
        else {
            // If deleting a like, inner close-brace of ([1+2}+3) where the outer open-paren is a ghost,
            // then become [1+3+3).
            if (this.matchBrack(cursor.options, side, this.parent?.parent)) {
                (this.parent?.parent).closeOpposing(this);
                (this.parent?.parent).unwrap();
            }
            else if (outward && wasSolid) {
                // If deleting outward from a solid pair, unwrap.
                this.unwrap();
                if (sib)
                    cursor.insDirOf(otherDir(side), sib);
                else if (parent)
                    cursor.insAtDirEnd(side, parent);
                return;
            }
            else {
                // If deleting just one of a pair of brackets, become one-sided.
                this.sides[side] = {
                    ch: OPP_BRACKS[this.sides[this.side].ch],
                    ctrlSeq: OPP_BRACKS[this.sides[this.side].ctrlSeq]
                };
                this.delims?.forEach((delim, index) => {
                    delim.classList.remove('mq-ghost');
                    if (index === (side === 'left' ? 0 : 1)) {
                        delim.classList.add('mq-ghost');
                        this.replaceBracket(delim, side);
                    }
                });
            }
            if (sib) {
                // Auto-expand so the ghost is at the far end.
                const origEnd = this.blocks[this.contentIndex].ends[side];
                this.blocks[this.contentIndex].elements.removeClass('mq-empty');
                new Fragment(sib, farEnd, otherDir(side))
                    .disown()
                    .withDirAdopt(otherDir(side), this.blocks[this.contentIndex], origEnd)
                    .elements.insAtDirEnd(side, this.blocks[this.contentIndex].elements);
                origEnd?.siblingCreated?.(cursor.options, side);
                cursor.insDirOf(otherDir(side), sib);
            }
            else {
                // Otherwise the cursor goes just outside or just inside the parentheses.
                if (outward)
                    cursor.insDirOf(side, this);
                else
                    cursor.insAtDirEnd(side, this.blocks[this.contentIndex]);
            }
        }
    }
    replaceBracket(brackFrag, side) {
        if (!(this instanceof Bracket) && !(this instanceof MathFunction))
            throw new Error('can only replace bracket for a Bracket or MathFunction');
        const symbol = this.getSymbol(side);
        brackFrag.innerHTML = symbol.html;
        brackFrag.style.width = symbol.width;
        if (side === 'left') {
            const next = brackFrag.nextElementSibling;
            if (next instanceof HTMLElement)
                next.style.marginLeft = symbol.width;
        }
        else {
            const prev = brackFrag.previousElementSibling;
            if (prev instanceof HTMLElement)
                prev.style.marginRight = symbol.width;
        }
    }
    getSymbol(side) {
        return (SVG_SYMBOLS[this.sides[side || 'right'].ch] || {
            width: '0',
            html: ''
        });
    }
    deleteTowards(dir, cursor) {
        this.deleteSide(otherDir(dir), false, cursor);
    }
    finalizeTree() {
        if (!this.inserted) {
            this.blocks[this.contentIndex].deleteOutOf = (dir, cursor) => {
                this.deleteSide(dir, true, cursor);
            };
            this.inserted = true;
            return;
        }
        this.delims?.[this.side === 'left' ? 1 : 0].classList.remove('mq-ghost');
        delete this.side;
    }
};
// Round/Square/Curly/Angle Brackets (aka Parens/Brackets/Braces)
// First typed as one-sided bracket with matching "ghost" bracket at
// far end of current block, until you type an opposing one.
class Bracket extends BracketMixin(MathCommand) {
    constructor(side, open, close, ctrlSeq, end) {
        super(`\\left${ctrlSeq}`, undefined, [open, close]);
        this.side = side;
        this.sides = {
            left: { ch: open, ctrlSeq: ctrlSeq },
            right: { ch: close, ctrlSeq: end }
        };
        this.placeCursor = noop;
        // If something is typed between the ghost and the far end of its block, then solidify the ghost.
        this.siblingCreated = (_opts, dir) => {
            if (dir === otherDir(this.side))
                this.finalizeTree();
        };
    }
    numBlocks() {
        return 1;
    }
    html() {
        const leftSymbol = this.getSymbol('left');
        const rightSymbol = this.getSymbol('right');
        // Wait until now to set the html template so that .side may be set by createLeftOf or the parser.
        this.htmlTemplate =
            '<span class="mq-bracket-container mq-non-leaf">' +
                `<span class="mq-bracket-l mq-scaled mq-paren${this.side === 'right' ? ' mq-ghost' : ''}" style="width:${leftSymbol.width}">` +
                leftSymbol.html +
                '</span>' +
                `<span class="mq-bracket-middle mq-non-leaf" style="margin-left:${leftSymbol.width};margin-right:${rightSymbol.width}">&0</span>` +
                `<span class="mq-bracket-r mq-scaled mq-paren${this.side === 'left' ? ' mq-ghost' : ''}" style="width:${rightSymbol.width}">` +
                rightSymbol.html +
                '</span>' +
                '</span>';
        return super.html();
    }
    latex() {
        return `\\left${this.sides.left.ctrlSeq}${this.ends.left?.latex() ?? ''}\\right${this.sides.right.ctrlSeq}`;
    }
    text() {
        return `${this.sides.left.ch}${this.ends.left?.text() ?? ''}${this.sides.right.ch}`;
    }
    mathspeak(opts) {
        const open = this.sides.left.ch, close = this.sides.right.ch;
        if (open === '|' && close === '|') {
            this.mathspeakTemplate = ['StartAbsoluteValue,', ', EndAbsoluteValue'];
            this.ariaLabel = 'absolute value';
        }
        else if (opts?.createdLeftOf && this.side) {
            return `${this.side} ${BRACKET_NAMES[this.side === 'left' ? this.textTemplate[0] : this.textTemplate[1]]}`;
        }
        else {
            this.mathspeakTemplate = ['left ' + BRACKET_NAMES[open] + ',', ', right ' + BRACKET_NAMES[close]];
            this.ariaLabel = BRACKET_NAMES[open] + ' block';
        }
        return super.mathspeak();
    }
}
class MathFunction extends BracketMixin(MathCommand) {
    contentIndex = 1;
    constructor(ctrlSeq) {
        // Change asin into arcsin (and the same for the other "a" trig variants).
        ctrlSeq = ctrlSeq.replace(/\\a(sin|cos|tan|sec|csc|cot)/, '\\arc$1');
        super(ctrlSeq, undefined, [`${ctrlSeq.slice(1)}(`, ')']);
        // If something is typed after the ghost, then solidify the end parenthesis.
        // Also if something is typed before or after the function,
        // then determine if padding is needed before the function name.
        this.siblingCreated = (_opts, dir) => {
            if (dir === 'right')
                this.finalizeTree();
            else
                this.updateFirst();
        };
        this.siblingDeleted = () => {
            this.updateFirst();
        };
        this.setAriaLabel();
    }
    setAriaLabel() {
        const baseName = this.ctrlSeq.slice(1).replace(/^arc/, '').replace(/h$/, '');
        this.ariaLabel =
            (this.ctrlSeq.endsWith('h') ? 'hyperbolic ' : '') +
                (this.ctrlSeq.startsWith('\\arc') ? 'arc' : '') +
                {
                    sin: 'sine',
                    cos: 'cosine',
                    tan: 'tangent',
                    sec: 'secant',
                    csc: 'cosecant',
                    cot: 'cotangent',
                    exp: 'natural exponential',
                    ln: 'natural logarithm',
                    log: 'logarithm'
                }[baseName];
        if (this.ends.right)
            this.ends.right.ariaLabel = `${this.ariaLabel ?? ''} parameter`;
    }
    // Add or remove padding depending on what is before the function name.
    updateFirst() {
        if (this.left &&
            !(this.left instanceof BinaryOperator) &&
            (!(this.left instanceof Variable) || !this.left.isPartOfOperator))
            this.elements.first.classList.add('mq-first');
        else
            this.elements.first.classList.remove('mq-first');
    }
    addToElements(el) {
        this.elements.add(el);
        const children = new VNode(this.elements.children().last).children();
        this.delims = [children.contents[0], children.contents[2]];
        this.content = children.contents[1];
    }
    numBlocks() {
        return 2;
    }
    html() {
        this.htmlTemplate =
            '<span class="mq-non-leaf mq-math-function">' +
                `<span class="mq-non-leaf mq-function-name">${this.ctrlSeq.slice(1)}</span>` +
                '<span class="mq-non-leaf mq-function-supsub">&0</span>' +
                '<span class="mq-function-params mq-bracket-container mq-non-leaf">' +
                `<span class="mq-bracket-l mq-scaled mq-paren" style="width:${SVG_SYMBOLS['('].width}">` +
                SVG_SYMBOLS['('].html +
                '</span>' +
                `<span class="mq-bracket-middle mq-non-leaf" style="margin-left:${SVG_SYMBOLS['('].width};margin-right:${SVG_SYMBOLS[')'].width}">&1</span>` +
                `<span class="mq-bracket-r mq-scaled mq-paren${this.side === 'left' ? ' mq-ghost' : ''}" style="width:${SVG_SYMBOLS[')'].width}">` +
                SVG_SYMBOLS[')'].html +
                '</span>' +
                '</span>' +
                '</span>';
        // Only allow supsub's to be inserted into the block before the parentheses.
        // If anything else is typed, move to the content block, so it will be inserted there.
        this.blocks[0].writeHandler = (cursor, ch) => {
            if (ch === '_' || ch === '^')
                return false;
            // If at the left end of the supsub block and a character was typed that extends this function name to
            // another one, then extend the function name.
            if (!cursor.left &&
                LatexCmds[`${this.ctrlSeq.slice(1)}${ch}`]?.prototype instanceof
                    MathFunction) {
                this.ctrlSeq = `${this.ctrlSeq}${ch}`;
                this.setAriaLabel();
                this.elements.children().first.textContent = (this.elements.children().first.textContent ?? '') + ch;
                this.bubble('reflow');
                return true;
            }
            this.enterContentBlock('left', cursor);
            // If a starting parenthesis is typed, don't actually create it.  Just move the cursor to the content
            // block, and let the hardcoded parenthesis appear.  So it appears as if typing the parenthesis does add
            // a parenthesis.
            if (ch === '(')
                return true;
            return false;
        };
        return super.html();
    }
    enterContentBlock(dir, cursor) {
        if (dir === 'left')
            cursor.insAtLeftEnd(this.blocks[1]);
        else
            cursor.insAtRightEnd(this.blocks[1]);
    }
    deleteSide(side, outward, cursor) {
        if (side === 'left' && !(this.blocks[0].isEmpty() && this.blocks[1].isEmpty())) {
            let brack, supsub;
            if (!this.blocks[1].isEmpty()) {
                cursor.insLeftOf(this);
                brack = new Bracket('left', '(', ')', '(', ')');
                brack.replaces(this.blocks[1].children());
                brack.createLeftOf(cursor);
                // Make the parentheses one sided if this function was.
                if (this.side) {
                    brack.delims?.[1].classList.add('mq-ghost');
                    brack.side = 'left';
                }
            }
            // If there is a SupSub in the supsub block, then it must be replaced with a new instance.  This is because
            // the original instance has the special MathFunction deleteOutOf handlers.  Those will not work for a usual
            // SupSub.
            if (!this.blocks[0].isEmpty() && this.blocks[0].ends.left instanceof SupSub) {
                cursor.insLeftOf(brack ?? this);
                // A temporary base is created first and then removed later, because if supSubsRequireOperand is true
                // the following causes errors when the contact weld is called later otherwise.
                const tmpBase = new Variable('x');
                tmpBase.createLeftOf(cursor);
                const origSupSub = this.blocks[0].ends.left;
                for (const supOrSub of ['sub', 'sup']) {
                    const src = origSupSub[supOrSub];
                    if (src) {
                        if (supsub)
                            cursor.insRightOf(supsub);
                        const newSupOrSub = new LatexCmds[supOrSub === 'sub' ? 'subscript' : 'superscript']();
                        newSupOrSub.replaces(src.children());
                        newSupOrSub.createLeftOf(cursor);
                        if (!supsub)
                            supsub = newSupOrSub;
                    }
                }
                tmpBase.remove();
            }
            cursor.controller.aria.queue(this.ariaLabel ?? this.ctrlSeq.slice(1));
            this.remove();
            if (supsub)
                cursor.insLeftOf(supsub);
            else if (brack)
                cursor.insLeftOf(brack);
        }
        else {
            if (side === 'left')
                cursor.controller.aria.queue(this);
            else
                cursor.controller.aria.queue('right parenthesis');
            super.deleteSide(side, outward, cursor);
        }
    }
    finalizeTree() {
        const inserted = this.inserted;
        super.finalizeTree();
        if (!inserted) {
            this.blocks[0].deleteOutOf = (dir, cursor) => {
                // If at the left end of the supsub block and removing the last character of this function name is still
                // a valid function name, then shorten the function name.
                if (!cursor.left &&
                    LatexCmds[this.ctrlSeq.slice(1, -1)]?.prototype instanceof
                        MathFunction) {
                    cursor.controller.aria.queue(this.ctrlSeq.slice(-1));
                    this.ctrlSeq = this.ctrlSeq.slice(0, -1);
                    this.setAriaLabel();
                    this.elements.children().first.textContent = this.ctrlSeq.slice(1);
                    this.bubble('reflow');
                    return;
                }
                if (dir === 'left')
                    this.deleteSide(dir, true, cursor);
                // If deleting right out of the supsub block, move the cursor to the content block.
                else
                    this.enterContentBlock('left', cursor);
            };
            this.blocks[1].deleteOutOf = (dir, cursor) => {
                // If deleting left out of the content block, move the cursor to the supsub block.
                if (dir === 'left')
                    cursor.insAtRightEnd(this.blocks[0]);
                else
                    this.deleteSide(dir, true, cursor);
            };
        }
        if (this.ends.right)
            this.ends.right.ariaLabel = `${this.ariaLabel ?? ''} parameter`;
        this.updateFirst();
    }
    latex() {
        return `${this.ctrlSeq}${this.blocks[0]?.latex() ?? ''}\\left(${this.blocks[1]?.latex() ?? ''}\\right)`;
    }
    text() {
        return `${this.left instanceof Letter ? ' ' : ''}${this.ctrlSeq.slice(1)}${this.blocks[0]?.text() ?? ''}(${this.blocks[1]?.text() ?? ''})`;
    }
    mathspeak() {
        return `${this.ariaLabel ?? ''} ${this.blocks[0].mathspeak()} left parenthesis, ${this.blocks[1].mathspeak()}, right parenthesis`;
    }
    parser() {
        // Create the other end solid.
        delete this.side;
        this.blocks = [new MathBlock(), new MathBlock()];
        for (const block of this.blocks) {
            block.adopt(this, this.ends.right);
        }
        return Parser.optWhitespace
            .then(Parser.regex(/^(?=[_^])/))
            .then(latexMathParser.block.map((block) => {
            block.children().adopt(this.blocks[0], this.blocks[0].ends.right);
        }))
            .many()
            .or(Parser.succeed(this))
            .then(Parser.optWhitespace)
            .then(Parser.regex(/^(?=\))|^(?=\\right\))/).or(Parser.string('\\left(')
            .then(latexMathParser.block
            .many()
            .map((blocks) => {
            for (const block of blocks) {
                block.children().adopt(this.blocks[1], this.blocks[1].ends.right);
            }
        })
            .then(Parser.optWhitespace)
            .skip(Parser.string('\\right)')))
            .or(Parser.string('(')
            .then(latexMathParser.block
            .many()
            .map((blocks) => {
            if (blocks[blocks.length - 1].text() === ')')
                blocks.splice(-1);
            for (const block of blocks) {
                block.children().adopt(this.blocks[1], this.blocks[1].ends.right);
            }
        })
            .then(Parser.optWhitespace)
            .skip(Parser.string(')').or(Parser.succeed(this))))
            .or(Parser.regex(/^\d+\.?\d*|^\d*\.?\d+/)
            .map((number) => {
            for (const d of number) {
                const digit = d === '.' ? new VanillaSymbol('.') : new Digit(d);
                digit.adopt(this.blocks[1], this.blocks[1].ends.right);
            }
        })
            .or(latexMathParser.block.map((block) => {
            block.children().adopt(this.blocks[1], this.blocks[1].ends.right);
        }))))
            .or(Parser.succeed(this))))
            .result(this);
    }
}
const latexMathParser = (() => {
    const commandToBlock = (cmd) => {
        // can also take in a Fragment
        const block = new MathBlock();
        cmd.adopt(block);
        return block;
    };
    const joinBlocks = (blocks) => {
        const firstBlock = blocks[0] || new MathBlock();
        for (const block of blocks.slice(1)) {
            block.children().adopt(firstBlock, firstBlock.ends.right);
        }
        return firstBlock;
    };
    // Parsers yielding either MathCommands, or Fragments of MathCommands
    //   (either way, something that can be adopted by a MathBlock)
    const variable = Parser.letter.map((c) => new Letter(c));
    const digit = Parser.regex(/^\d/).map((c) => new Digit(c));
    const symbol = Parser.regex(/^[^${}\\_^]/).map((c) => new VanillaSymbol(c));
    const controlSequence = Parser.regex(/^[^\\a-eg-zA-Z]/) // hotfix #164; match MathBlock::write
        .or(Parser.string('\\').then(Parser.regex(/^[a-z]+/i)
        .or(Parser.regex(/^\s+/).result(' '))
        .or(Parser.any)))
        .then((ctrlSeq) => {
        const cmdKlass = LatexCmds[ctrlSeq];
        if (cmdKlass) {
            return new cmdKlass(ctrlSeq).parser();
        }
        else {
            return Parser.fail(`unknown command: \\${ctrlSeq}`);
        }
    });
    const command = controlSequence.or(variable).or(digit).or(symbol);
    // Parsers yielding MathBlocks
    const mathGroup = Parser.string('{')
        .then(() => mathSequence)
        .skip(Parser.string('}'));
    const mathBlock = Parser.optWhitespace.then(mathGroup.or(command.map(commandToBlock)));
    const mathSequence = mathBlock.many().map(joinBlocks).skip(Parser.optWhitespace);
    const optMathBlock = Parser.string('[')
        .then(mathBlock
        .then((block) => {
        return block.join('latex') !== ']' ? Parser.succeed(block) : Parser.fail();
    })
        .many()
        .map(joinBlocks)
        .skip(Parser.optWhitespace))
        .skip(Parser.string(']'));
    const latexMath = mathSequence;
    latexMath.block = mathBlock;
    latexMath.optBlock = optMathBlock;
    return latexMath;
})();

// The MathBlock and the RootMathCommand (used by the RootTextBlock) use this.
const writeMethodMixin = (Base) => class extends Base {
    writeHandler;
    chToCmd(ch, options) {
        const cons = CharCmds[ch] || LatexCmds[ch];
        // exclude f because it gets a dedicated command with more spacing
        if (/^[a-eg-zA-Z]$/.exec(ch))
            return new Letter(ch);
        else if (/^\d$/.test(ch))
            return new Digit(ch);
        else if (options && options.typingSlashWritesDivisionSymbol && ch === '/')
            return new LatexCmds['\u00f7'](ch);
        else if (options && options.typingAsteriskWritesTimesSymbol && ch === '*')
            return new LatexCmds['\u00d7'](ch);
        else if (cons)
            return new cons(ch);
        else
            return new VanillaSymbol(ch);
    }
    write(cursor, ch) {
        if (this.writeHandler?.(cursor, ch))
            return;
        if (this.isSupSubLeft) {
            if (cursor.options.autoSubscriptNumerals && this === this.parent?.sub) {
                if (ch === '_')
                    return;
                const cmd = this.chToCmd(ch, cursor.options);
                if (cmd.isSymbol)
                    cursor.deleteSelection();
                else
                    cursor.clearSelection().insRightOf(this.parent);
                cmd.createLeftOf(cursor.show());
                cursor.controller.aria.queue('Baseline').alert(cmd.mathspeak({ createdLeftOf: cursor }));
                return;
            }
            if (cursor.left &&
                !cursor.right &&
                !cursor.selection &&
                cursor.options.charsThatBreakOutOfSupSub.includes(ch) &&
                this.parent) {
                cursor.insRightOf(this.parent);
                cursor.controller.aria.queue('Baseline');
            }
        }
        const cmd = this.chToCmd(ch, cursor.options);
        if (cursor.selection)
            cmd.replaces(cursor.replaceSelection());
        if (!cursor.isTooDeep()) {
            cmd.createLeftOf(cursor.show());
            // There is a special case for the slash so that fractions are voiced while typing.
            if (ch === '/')
                cursor.controller.aria.alert('over');
            else
                cursor.controller.aria.alert(cmd.mathspeak({ createdLeftOf: cursor }));
        }
    }
};
// Children and parent of MathCommand's. Basically partitions all the
// symbols and operators that descend (in the Math DOM tree) from
// ancestor operators.
class MathBlock extends BlockFocusBlur(writeMethodMixin(MathElement)) {
    ariaLabel = 'block';
    join(methodName) {
        return this.foldChildren('', (fold, child) => fold + child[methodName]());
    }
    html() {
        return this.join('html');
    }
    latex() {
        return this.join('latex');
    }
    text() {
        return this.ends.left && this.ends.left === this.ends.right ? this.ends.left.text() : this.join('text');
    }
    mathspeak() {
        let tempOp = '';
        const autoOps = this.controller ? this.controller.options.autoOperatorNames : { _maxLength: 0 };
        return (this.foldChildren([], (speechArray, cmd) => {
            if (cmd instanceof Letter && cmd.isPartOfOperator) {
                tempOp += cmd.mathspeak();
            }
            else {
                if (tempOp !== '') {
                    if (autoOps._maxLength > 0) {
                        const x = autoOps[tempOp.toLowerCase()];
                        if (typeof x === 'string')
                            tempOp = x;
                    }
                    speechArray.push(tempOp + ' ');
                    tempOp = '';
                }
                let mathspeakText = cmd.mathspeak();
                const cmdText = cmd.ctrlSeq;
                if (isNaN(cmdText) && // TODO - revisit this to improve the isNumber() check
                    cmdText !== '.' &&
                    (!cmd.parent?.parent || !(cmd.parent.parent instanceof LatexCmds.mathrm))) {
                    mathspeakText = ' ' + mathspeakText + ' ';
                }
                speechArray.push(mathspeakText);
            }
            return speechArray;
        })
            .join('')
            .replace(/ +(?= )/g, '')
            // For Apple devices in particular, split out digits after a decimal point so they aren't read aloud as
            // whole words.  Not doing so makes 123.456 potentially spoken as "one hundred twenty three point four
            // hundred fifty six." Instead, add spaces so it is spoken as "one hundred twenty three point four five
            // six."
            .replace(/(\.)([0-9]+)/g, (_match, p1, p2) => p1 + p2.split('').join(' ').trim()));
    }
    keystroke(key, e, ctrlr) {
        if (ctrlr.options.enableSpaceNavigation &&
            ctrlr.cursor.depth() > 1 &&
            (key === 'Spacebar' || key === 'Shift-Spacebar') &&
            ctrlr.cursor.left?.ctrlSeq !== ',') {
            e.preventDefault();
            ctrlr.escapeDir(key === 'Shift-Spacebar' ? 'left' : 'right', key, e);
            return;
        }
        super.keystroke(key, e, ctrlr);
    }
    // editability methods: called by the cursor for editing, cursor movements,
    // and selection of the MathQuill tree, these all take in a direction and
    // the cursor
    moveOutOf(dir, cursor, updown) {
        const updownInto = updown && this.parent?.[`${updown}Into`];
        if (!updownInto && this[dir]) {
            cursor.insAtDirEnd(otherDir(dir), this[dir]);
            if (cursor.parent)
                cursor.controller.aria.queueDirEndOf(otherDir(dir)).queue(cursor.parent, true);
        }
        else if (this.parent) {
            cursor.insDirOf(dir, this.parent);
            cursor.controller.aria.queueDirOf(dir).queue(this.parent);
        }
    }
    selectOutOf(dir, cursor) {
        if (this.parent)
            cursor.insDirOf(dir, this.parent);
    }
    deleteOutOf(_dir, cursor) {
        cursor.unwrapGramp();
    }
    seek(pageX, cursor) {
        let node = this.ends.right;
        const rect = node?.elements.firstElement.getBoundingClientRect() ?? undefined;
        if (!node || (rect?.left ?? 0) + (rect?.width ?? 0) < pageX) {
            cursor.insAtRightEnd(this);
            return;
        }
        if (pageX < (this.ends.left?.elements.firstElement.getBoundingClientRect().left ?? 0)) {
            cursor.insAtLeftEnd(this);
            return;
        }
        while (pageX < (node?.elements.firstElement.getBoundingClientRect().left ?? 0))
            node = node?.left;
        node?.seek(pageX, cursor);
    }
    writeLatex(cursor, latex) {
        const all = Parser.all;
        const eof = Parser.eof;
        const block = latexMathParser.skip(eof).or(all.result(false)).parse(latex);
        if (block && !block.isEmpty() && block.prepareInsertionAt(cursor)) {
            if (cursor.parent)
                block.children().adopt(cursor.parent, cursor.left, cursor.right);
            const elements = block.domify();
            cursor.element.before(...elements.contents);
            cursor.left = block.ends.right;
            block.finalizeInsert(cursor.options, cursor);
            block.ends.right?.right?.siblingCreated?.(cursor.options, 'left');
            block.ends.left?.left?.siblingCreated?.(cursor.options, 'right');
            cursor.parent?.bubble('reflow');
        }
    }
    focus() {
        super.focus();
        return this;
    }
    blur() {
        super.blur();
        return this;
    }
}
class RootMathBlock extends MathBlock {
    constructor() {
        super();
        RootBlockMixin(this);
    }
}
class RootMathCommand extends writeMethodMixin(MathCommand) {
    cursor;
    constructor(cursor) {
        super('$');
        this.cursor = cursor;
        this.htmlTemplate = '<span class="mq-math-mode">&0</span>';
    }
    createBlocks() {
        super.createBlocks();
        const leftEnd = this.ends.left;
        leftEnd.write = (cursor, ch) => {
            if (ch !== '$')
                this.write(cursor, ch);
            else if (leftEnd.isEmpty()) {
                if (leftEnd.parent)
                    cursor.insRightOf(leftEnd.parent);
                leftEnd.parent?.deleteTowards('left', cursor);
                new VanillaSymbol('\\$', '$').createLeftOf(cursor.show());
            }
            else if (!cursor.right && leftEnd.parent)
                cursor.insRightOf(leftEnd.parent);
            else if (!cursor.left && leftEnd.parent)
                cursor.insLeftOf(leftEnd.parent);
            else
                this.write(cursor, ch);
        };
    }
    latex() {
        return `$${this.ends.left?.latex() ?? ''}$`;
    }
}

// Cursor "singleton" class
// The main thing that manipulates the Math DOM. Makes sure to manipulate the HTML DOM to match.
// Sort of a singleton, since there should only be one per editable math textbox, but any one HTML document can contain
// many such textboxes, so any one JS environment could actually contain many instances.
// A fake cursor in the fake textbox that the math is rendered in.
class Cursor extends Point {
    controller;
    options;
    element = document.createElement('span');
    upDownCache = {};
    intervalId;
    selection;
    anticursor;
    selectionChanged;
    // Closured for setInterval
    blink = () => this.element.classList.toggle('mq-blink');
    constructor(initParent, options, controller) {
        super(initParent);
        this.controller = controller;
        this.options = options;
        this.element.classList.add('mq-cursor');
        this.element.textContent = '\u200B';
    }
    show() {
        this.element.classList.remove('mq-blink');
        this.element.style.display = '';
        if (this.intervalId) {
            // The cursor is already shown, just restart the interval.
            clearInterval(this.intervalId);
        }
        else {
            // The cursor was hidden and removed, so insert this.element back into the DOM.
            if (this.right) {
                if (this.selection && this.selection.ends.left?.left === this.left)
                    this.selection.elements.first.before(this.element);
                else
                    this.right.elements.first.before(this.element);
            }
            else
                this.parent?.elements.firstElement.append(this.element);
            this.parent?.focus();
        }
        this.intervalId = setInterval(this.blink, 500);
        return this;
    }
    hide() {
        if (this.intervalId)
            clearInterval(this.intervalId);
        delete this.intervalId;
        this.element.style.display = 'none';
        this.element.remove();
        return this;
    }
    withDirInsertAt(dir, parent, withDir, oppDir) {
        const oldParent = this.parent;
        this.parent = parent;
        this[dir] = withDir;
        this[otherDir(dir)] = oppDir;
        // By contract blur is called after all has been said and done and the cursor has actually been moved.
        if (oldParent !== parent && oldParent?.blur)
            oldParent.blur(this);
    }
    insDirOf(dir, el) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        if (dir === 'left')
            el.elements.first.before(this.element);
        else
            el.elements.last.after(this.element);
        if (el.parent)
            this.withDirInsertAt(dir, el.parent, el[dir], el);
        this.parent?.elements.addClass('mq-has-cursor');
        return this;
    }
    insLeftOf(el) {
        return this.insDirOf('left', el);
    }
    insRightOf(el) {
        return this.insDirOf('right', el);
    }
    insAtDirEnd(dir, el) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        if (dir === 'left')
            el.elements.firstElement.prepend(this.element);
        else
            el.elements.lastElement.append(this.element);
        this.withDirInsertAt(dir, el, undefined, el.ends[dir]);
        el.focus();
        return this;
    }
    insAtLeftEnd(el) {
        return this.insAtDirEnd('left', el);
    }
    insAtRightEnd(el) {
        return this.insAtDirEnd('right', el);
    }
    // Jump up or down from one block TNode to another:
    // - cache the current Point in the TNode we are jumping from
    // - check if there is a Point in it cached for the node we are jumping to
    //   + if so put the cursor there,
    //   + if not seek a position in the node that is horizontally closest to the cursor's current position
    jumpUpDown(from, to) {
        this.upDownCache[from.id] = Point.copy(this);
        const cached = this.upDownCache[to.id];
        if (cached) {
            if (cached.right)
                this.insLeftOf(cached.right);
            else if (cached.parent)
                this.insAtRightEnd(cached.parent);
        }
        else {
            to.seek(this.offset().left, this);
        }
        this.controller.aria.queue(to, true);
    }
    offset() {
        return this.element.getBoundingClientRect();
    }
    unwrapGramp() {
        const gramp = this.parent?.parent;
        const greatgramp = gramp?.parent;
        const rightward = gramp?.right;
        let leftward = gramp?.left;
        gramp?.disown().eachChild((uncle) => {
            if (uncle.isEmpty())
                return true;
            if (greatgramp)
                uncle
                    .children()
                    .adopt(greatgramp, leftward, rightward)
                    .each((cousin) => {
                    gramp.elements.first.before(...cousin.elements.contents);
                    return true;
                });
            leftward = uncle.ends.right;
            return true;
        });
        if (!this.right) {
            // Find something rightward to insert left of.
            if (this.left)
                this.right = this.left.right;
            else {
                while (!this.right) {
                    this.parent = this.parent?.right;
                    if (this.parent)
                        this.right = this.parent.ends.left;
                    else {
                        this.right = gramp?.right;
                        this.parent = greatgramp;
                        break;
                    }
                }
            }
        }
        if (this.right)
            this.insLeftOf(this.right);
        else if (greatgramp)
            this.insAtRightEnd(greatgramp);
        gramp?.elements.remove();
        if (gramp?.left?.siblingDeleted)
            gramp.left.siblingDeleted(this.options, 'right');
        if (gramp?.right?.siblingDeleted)
            gramp.right.siblingDeleted(this.options, 'left');
    }
    startSelection() {
        this.anticursor = Point.copy(this);
        // Create a map from each ancestor of the anticursor to its child that is also an ancestor.
        // In other words, the anticursor's ancestor chain in reverse order.
        this.anticursor.ancestors = {};
        for (let ancestor = this.anticursor; ancestor.parent; ancestor = ancestor.parent) {
            this.anticursor.ancestors[ancestor.parent.id] = ancestor;
        }
    }
    endSelection() {
        delete this.anticursor;
    }
    select() {
        if (this.left === this.anticursor?.left && this.parent === this.anticursor?.parent)
            return false;
        if (!this.anticursor?.ancestors)
            throw new Error('selection not well formed');
        // Find the lowest common ancestor (`lca`), and the ancestor of the cursor
        // whose parent is the LCA (which will be an end of the selection fragment).
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        let ancestor = this, lca = null;
        for (; ancestor.parent; ancestor = ancestor.parent) {
            if (ancestor.parent.id in this.anticursor.ancestors) {
                lca = ancestor.parent;
                break;
            }
        }
        if (!lca)
            throw new Error('cursor and anticursor must be in the same tree');
        // The cursor and the anticursor should be in the same tree, because the
        // mousemove handler attached to the document, unlike the one attached to
        // the root HTML DOM element, doesn't try to get the math tree node of the
        // mousemove target, and Cursor::seek() based solely on coordinates stays
        // within the tree of `this` cursor's root.
        // The other end of the selection fragment, the ancestor of the anticursor
        // whose parent is the LCA.
        const antiAncestor = this.anticursor.ancestors[lca.id];
        // Now we have two either Nodes or Points, guaranteed to have a common
        // parent and guaranteed that if both are Points, they are not the same,
        // and we have to figure out which is the left end and which the right end
        // of the selection.
        let leftEnd, rightEnd, dir = 'right';
        // This is an extremely subtle algorithm.
        // As a special case, `ancestor` could be a Point and `antiAncestor` a TNode
        // immediately to `ancestor`'s left.
        // In all other cases,
        // - both Nodes
        // - `ancestor` a Point and `antiAncestor` a TNode
        // - `ancestor` a TNode and `antiAncestor` a Point
        // `antiAncestor.right === rightward.right` for some `rightward` that is
        // `ancestor` or to its right, if and only if `antiAncestor` is to
        // the right of `ancestor`.
        if (ancestor.left !== antiAncestor) {
            for (let rightward = ancestor; rightward; rightward = rightward.right) {
                if (rightward.right === antiAncestor.right) {
                    dir = 'left';
                    leftEnd = ancestor;
                    rightEnd = antiAncestor;
                    break;
                }
            }
        }
        if (dir === 'right') {
            leftEnd = antiAncestor;
            rightEnd = ancestor;
        }
        // only want to select Nodes up to Points, can't select Points themselves
        if (leftEnd instanceof Point)
            leftEnd = leftEnd.right;
        if (rightEnd instanceof Point)
            rightEnd = rightEnd.left;
        this.hide().selection = lca.selectChildren(leftEnd, rightEnd);
        const selectionEndDir = this.selection?.ends[dir];
        if (selectionEndDir)
            this.insDirOf(dir, selectionEndDir);
        this.selectionChanged?.();
        return true;
    }
    clearSelection() {
        if (this.selection) {
            this.selection.clear();
            delete this.selection;
            this.selectionChanged?.();
        }
        return this;
    }
    deleteSelection() {
        if (!this.selection)
            return;
        this.left = this.selection.ends.left?.left;
        this.right = this.selection.ends.right?.right;
        this.selection.remove();
        this.selectionChanged?.();
        delete this.selection;
    }
    replaceSelection() {
        const seln = this.selection;
        if (seln) {
            this.left = seln.ends.left?.left;
            this.right = seln.ends.right?.right;
            delete this.selection;
        }
        return seln;
    }
    depth() {
        let node = this.parent;
        let depth = 0;
        while (node) {
            depth += node instanceof MathBlock ? 1 : 0;
            node = node.parent;
        }
        return depth;
    }
    isTooDeep(offset) {
        if (this.options.maxDepth !== undefined) {
            return this.depth() + (offset || 0) > this.options.maxDepth;
        }
    }
}

// Horizontal panning for editable fields that overflow their width
const HorizontalScroll = (Base) => class extends Base {
    scrollHoriz() {
        const rootRect = this.root.elements.firstElement.getBoundingClientRect();
        let scrollBy = 0;
        if (!this.cursor.selection) {
            const x = this.cursor.element.getBoundingClientRect().left;
            if (x > rootRect.right - 20)
                scrollBy = x - (rootRect.right - 20);
            else if (x < rootRect.left + 20)
                scrollBy = x - (rootRect.left + 20);
            else
                return;
        }
        else {
            const rect = this.cursor.selection.elements.firstElement.getBoundingClientRect();
            const overLeft = rect.left - (rootRect.left + 20);
            const overRight = rect.right - (rootRect.right - 20);
            if (this.cursor.selection.ends.left === this.cursor.right) {
                if (overLeft < 0)
                    scrollBy = overLeft;
                else if (overRight > 0) {
                    if (rect.left - overRight < rootRect.left + 20)
                        scrollBy = overLeft;
                    else
                        scrollBy = overRight;
                }
                else
                    return;
            }
            else {
                if (overRight > 0)
                    scrollBy = overRight;
                else if (overLeft < 0) {
                    if (rect.right - overLeft > rootRect.right - 20)
                        scrollBy = overRight;
                    else
                        scrollBy = overLeft;
                }
                else
                    return;
            }
        }
        setTimeout(() => (this.root.elements.firstElement.scrollLeft += scrollBy), 100);
    }
};

// Latex Controller Extension
const LatexControllerExtension = (Base) => class extends Base {
    exportLatex() {
        return this.root.latex().replace(/(\\[a-z]+) (?![a-z])/gi, '$1');
    }
    writeLatex(latex) {
        const cursor = this.notify('edit').cursor;
        cursor.parent?.writeLatex(cursor, latex);
        return this;
    }
    renderLatexMath(latex) {
        const block = latexMathParser
            .skip(Parser.eof)
            .or(Parser.all.result(false))
            .parse(latex);
        this.root.eachChild('postOrder', 'dispose');
        delete this.root.ends.left;
        delete this.root.ends.right;
        if (block instanceof MathBlock && block.prepareInsertionAt(this.cursor)) {
            block.children().adopt(this.root);
            const html = block.join('html');
            this.root.elements.html(html);
            this.root.domify(this.root.elements.children());
            this.root.finalizeInsert(this.cursor.options, this.cursor);
        }
        else {
            this.root.elements.empty();
        }
        delete this.cursor.selection;
        this.updateMathspeak();
        this.cursor.insAtRightEnd(this.root);
    }
    renderLatexText(latex) {
        this.root.elements
            .children()
            .contents.slice(1)
            .forEach((el) => {
            el.remove();
        });
        this.root.eachChild('postOrder', 'dispose');
        delete this.root.ends.left;
        delete this.root.ends.right;
        delete this.cursor.selection;
        this.cursor.show().insAtRightEnd(this.root);
        // Parser RootMathCommand
        const mathMode = Parser.string('$')
            .then(latexMathParser)
            // because TeX is insane, math mode doesn't necessarily
            // have to end.  So we allow for the case that math mode
            // continues to the end of the stream.
            .skip(Parser.string('$').or(Parser.eof))
            .map((block) => {
            // HACK FIXME: This shouldn't have to have access to cursor
            const rootMathCommand = new RootMathCommand(this.cursor);
            rootMathCommand.createBlocks();
            const rootMathBlock = rootMathCommand.ends.left;
            block.children().adopt(rootMathBlock);
            return rootMathCommand;
        });
        const escapedDollar = Parser.string('\\$').result('$');
        const textChar = escapedDollar.or(Parser.regex(/^[^$]/)).map(VanillaSymbol);
        const latexText = mathMode.or(textChar).many();
        const commands = latexText.skip(Parser.eof).or(Parser.all.result(false)).parse(latex);
        if (commands) {
            for (const command of commands) {
                command.adopt(this.root, this.root.ends.right);
            }
            this.root.elements.lastElement.append(...this.root.domify().contents);
            this.root.finalizeInsert(this.cursor.options, this.cursor);
        }
    }
};

// Elements for abstract classes of text blocks
// Blocks of plain text, with one or two TextPiece's as children.
// Represents flat strings of typically serif-font Roman characters, as
// opposed to hierchical, nested, tree-structured math.
// Wraps a single HTMLSpanElement.
class TextBlock extends BlockFocusBlur(deleteSelectTowardsMixin(TNode)) {
    replacedText;
    anticursorPosition = 0;
    mathspeakTemplate = ['StartText', 'EndText'];
    constructor() {
        super();
        this.ctrlSeq = '\\text';
        this.ariaLabel = 'Text';
    }
    replaces(replacedText) {
        if (replacedText instanceof Fragment)
            this.replacedText = replacedText.remove().elements.text();
        else if (typeof replacedText === 'string')
            this.replacedText = replacedText;
    }
    addToElements(el) {
        super.addToElements(el);
        this.ends.left?.addToElements(this.elements.first.firstChild);
    }
    createLeftOf(cursor) {
        super.createLeftOf(cursor);
        cursor.insAtRightEnd(this);
        if (this.replacedText) {
            for (const char of this.replacedText)
                this.write(cursor, char);
        }
        this.right?.siblingCreated?.(cursor.options, 'left');
        this.left?.siblingCreated?.(cursor.options, 'right');
        this.bubble('reflow');
    }
    parser() {
        // TODO: correctly parse text mode
        return Parser.optWhitespace
            .then(Parser.string('{'))
            .then(Parser.regex(/^(\\}|[^}])*/))
            .skip(Parser.string('}'))
            .map((text) => {
            if (text.length === 0)
                return new Fragment();
            new TextPiece(text.replace(/\\{/g, '{').replace(/\\}/g, '}')).adopt(this);
            return this;
        });
    }
    textContents() {
        return this.foldChildren('', (text, child) => text + child.text());
    }
    text() {
        return this.textContents();
    }
    latex() {
        const contents = this.textContents();
        if (contents.length === 0)
            return '';
        return '\\text{' + contents.replace(/[{}]/g, '\\$&') + '}';
    }
    html() {
        return `<span class="mq-text-mode" ${mqCmdId}=${this.id.toString()}>${this.textContents()}</span>`;
    }
    mathspeak(opts) {
        if (opts?.ignoreShorthand) {
            return `${this.mathspeakTemplate[0]}, ${this.textContents()}, ${this.mathspeakTemplate[1]}`;
        }
        else {
            return this.textContents();
        }
    }
    // editability methods: called by the cursor for editing, cursor movements,
    // and selection of the MathQuill tree, these all take in a direction and
    // the cursor
    moveTowards(dir, cursor) {
        cursor.insAtDirEnd(otherDir(dir), this);
        if (cursor.parent)
            cursor.controller.aria.queueDirEndOf(otherDir(dir)).queue(cursor.parent, true);
    }
    moveOutOf(dir, cursor) {
        cursor.insDirOf(dir, this);
        cursor.controller.aria.queueDirOf(dir).queue(this);
    }
    unselectInto(dir, cursor) {
        cursor.insAtDirEnd(otherDir(dir), this);
        // Split the text at the stored anticursor position, and reconstruct the anticursor.
        const newTextPc = cursor[dir].splitRight(this.anticursorPosition);
        if (dir === 'left')
            cursor.left = newTextPc;
        cursor.anticursor = new Point(this, newTextPc.left, newTextPc);
        cursor.anticursor.ancestors = {};
        for (let ancestor = cursor.anticursor; ancestor.parent; ancestor = ancestor.parent) {
            cursor.anticursor.ancestors[ancestor.parent.id] = ancestor;
        }
    }
    selectOutOf(dir, cursor) {
        this.anticursorPosition =
            dir === 'left'
                ? (cursor.selection?.elements.first.textContent?.length ?? 1) - 1
                : this.textContents().length - (cursor.selection?.elements.first.textContent?.length ?? 1) + 1;
        cursor.insDirOf(dir, this);
    }
    deleteOutOf(_dir, cursor) {
        // backspace and delete at ends of block don't unwrap
        if (this.isEmpty())
            cursor.insRightOf(this);
    }
    write(cursor, ch) {
        cursor.show().deleteSelection();
        if (ch !== '$') {
            this.postOrder('reflow');
            if (!cursor.left)
                new TextPiece(ch).createLeftOf(cursor);
            else
                cursor.left.appendText(ch);
            this.bubble('reflow');
        }
        else if (this.isEmpty()) {
            cursor.insRightOf(this);
            new VanillaSymbol('\\$', '$').createLeftOf(cursor);
        }
        else if (!cursor.right)
            cursor.insRightOf(this);
        else if (!cursor.left)
            cursor.insLeftOf(this);
        else {
            // split apart
            const leftBlock = new TextBlock();
            const leftPc = this.ends.left;
            leftPc?.disown().elements.detach();
            leftPc?.adopt(leftBlock);
            cursor.insLeftOf(this);
            super.createLeftOf.call(leftBlock, cursor); // micro-optimization, not for correctness
        }
        this.bubble('reflow');
        cursor.controller.aria.alert(ch);
    }
    writeLatex(cursor, latex) {
        if (!cursor.left)
            new TextPiece(latex).createLeftOf(cursor);
        else
            cursor.left.appendText(latex);
        this.bubble('reflow');
    }
    seek(pageX, cursor) {
        cursor.hide();
        const textPc = this.fuseChildren();
        if (!textPc)
            return;
        // Insert cursor at approx position in DOMTextNode
        const cursorStyle = getComputedStyle(this.elements.firstElement);
        const avgChWidth = (this.elements.firstElement.getBoundingClientRect().width -
            parseFloat(cursorStyle.paddingLeft) -
            parseFloat(cursorStyle.paddingRight)) /
            this.text().length;
        const approxPosition = Math.round((pageX - this.elements.firstElement.getBoundingClientRect().left) / avgChWidth);
        if (approxPosition <= 0)
            cursor.insAtLeftEnd(this);
        else if (approxPosition >= textPc.text().length)
            cursor.insAtRightEnd(this);
        else
            cursor.insLeftOf(textPc.splitRight(approxPosition));
        // Move towards mousedown (pageX)
        let displ = pageX - cursor.show().offset().left; // displacement
        const dir = displ && displ < 0 ? 'left' : 'right';
        const numericDir = dir === 'left' ? -1 : 1;
        let prevDispl = numericDir;
        // displ * prevDispl > 0 iff displacement direction === previous direction
        while (cursor[dir] && displ * prevDispl > 0) {
            cursor[dir].moveTowards(dir, cursor);
            prevDispl = displ;
            displ = pageX - cursor.offset().left;
        }
        if (numericDir * displ < -numericDir * prevDispl)
            cursor[otherDir(dir)]?.moveTowards(otherDir(dir), cursor);
        if (!cursor.anticursor) {
            // About to start mouse-selecting, the anticursor is going to be placed here.
            this.anticursorPosition = cursor.left?.text().length ?? 0;
        }
        else if (cursor.anticursor.parent === this) {
            // Mouse selecting within this TextBlock, re-insert the anticursor.
            const cursorPosition = cursor.left?.text().length ?? 0;
            if (this.anticursorPosition === cursorPosition) {
                cursor.startSelection();
            }
            else {
                let newTextPc;
                if (this.anticursorPosition < cursorPosition) {
                    newTextPc = cursor.left.splitRight(this.anticursorPosition);
                    cursor.left = newTextPc;
                }
                else {
                    newTextPc = cursor.right.splitRight(this.anticursorPosition - cursorPosition);
                }
                cursor.anticursor = new Point(this, newTextPc.left, newTextPc);
                cursor.anticursor.ancestors = {};
                for (let ancestor = cursor.anticursor; ancestor.parent; ancestor = ancestor.parent) {
                    cursor.anticursor.ancestors[ancestor.parent.id] = ancestor;
                }
            }
        }
    }
    blur(cursor) {
        super.blur();
        if (!cursor)
            return;
        if (this.textContents() === '') {
            this.remove();
            if (cursor.parent === this) {
                cursor.parent = this.parent;
                cursor.left = this.left;
                cursor.right = this.right;
            }
            else if (cursor.left === this)
                cursor.left = this.left;
            else if (cursor.right === this)
                cursor.right = this.right;
        }
        else {
            // If the text block contains the selection, then that needs to be removed before fuseChildren is called.
            if (this.elements.find('.mq-selection').contents.length)
                cursor.clearSelection();
            this.fuseChildren();
        }
        this.getController()?.handle('textBlockExit');
    }
    fuseChildren() {
        this.elements.first.normalize();
        const textPcDom = this.elements.first.firstChild;
        if (!textPcDom)
            return;
        // nodeType === 3 is a text node.
        // See https://developer.mozilla.org/en-US/docs/Web/API/Node/nodeType.
        if (textPcDom.nodeType !== 3)
            throw new Error('only node in TextBlock span must be a Text node');
        const textPc = new TextPiece(textPcDom.data);
        textPc.addToElements(textPcDom);
        this.children().disown();
        return textPc.adopt(this);
    }
    focus() {
        super.focus();
        this.getController()?.handle('textBlockEnter');
    }
}
// A piece of plain text, with a TextBlock as a parent and no children. This wraps a single DOMTextNode.
// For convenience, it has a textStr property that is a string that mirrors the text contents of the DOMTextNode.
// Text contents must always be nonempty.
class TextPiece extends TNode {
    textStr;
    dom;
    constructor(text) {
        super();
        this.textStr = text;
    }
    text() {
        return this.textStr;
    }
    addToElements(dom) {
        this.dom = dom;
        this.elements = new VNode(dom);
    }
    domify() {
        this.addToElements(document.createTextNode(this.textStr));
        return this.elements;
    }
    appendText(text) {
        this.textStr += text;
        this.dom?.appendData(text);
    }
    prependText(text) {
        this.textStr = text + this.textStr;
        this.dom?.insertData(0, text);
    }
    insTextAtDirEnd(text, dir) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        if (dir === 'right')
            this.appendText(text);
        else
            this.prependText(text);
    }
    splitRight(i) {
        const newPc = new TextPiece(this.textStr.slice(i));
        if (this.parent)
            newPc.adopt(this.parent, this, this.right);
        if (this.dom)
            newPc.addToElements(this.dom.splitText(i));
        this.textStr = this.textStr.slice(0, i);
        return newPc;
    }
    endChar(dir, text) {
        return text.charAt(dir === 'left' ? 0 : -1 + text.length);
    }
    moveTowards(dir, cursor) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        const ch = this.endChar(otherDir(dir), this.textStr);
        const from = this[otherDir(dir)];
        if (from)
            from.insTextAtDirEnd(ch, dir);
        else
            new TextPiece(ch).createDir(otherDir(dir), cursor);
        this.deleteTowards(dir, cursor);
    }
    latex() {
        return this.textStr;
    }
    deleteTowards(dir, cursor) {
        if (this.textStr.length > 1) {
            let deletedChar;
            if (dir === 'right') {
                this.dom?.deleteData(0, 1);
                deletedChar = this.textStr[0];
                this.textStr = this.textStr.slice(1);
            }
            else {
                // Note that the order of these 2 lines is important.
                // (the second line mutates this.textStr.length)
                this.dom?.deleteData(-1 + this.textStr.length, 1);
                deletedChar = this.textStr[this.textStr.length - 1];
                this.textStr = this.textStr.slice(0, -1);
            }
            cursor.controller.aria.queue(deletedChar);
        }
        else {
            this.remove();
            this.elements.remove();
            cursor[dir] = this[dir];
            cursor.controller.aria.queue(this.textStr);
        }
    }
    selectTowards(dir, cursor) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        const anticursor = cursor.anticursor;
        const ch = this.endChar(otherDir(dir), this.textStr);
        if (anticursor?.[dir] === this) {
            const newPc = new TextPiece(ch).createDir(dir, cursor);
            anticursor[dir] = newPc;
            cursor.insDirOf(dir, newPc);
        }
        else {
            const from = this[otherDir(dir)];
            if (from)
                from.insTextAtDirEnd(ch, dir);
            else {
                const newPc = new TextPiece(ch).createDir(otherDir(dir), cursor);
                if (cursor.selection)
                    newPc.elements.insDirOf(otherDir(dir), cursor.selection.elements);
            }
            if (this.textStr.length === 1 && anticursor?.[otherDir(dir)] === this) {
                anticursor[otherDir(dir)] = this[otherDir(dir)];
            }
        }
        this.deleteTowards(dir, cursor);
    }
    mathspeak() {
        return this.textStr;
    }
}
LatexCmds.text =
    LatexCmds.textnormal =
        LatexCmds.textrm =
            LatexCmds.textup =
                CharCmds['"'] =
                    LatexCmds.textmd =
                        TextBlock;
const makeTextBlock = (latex, tagName, attrs, ariaLabel) => class extends TextBlock {
    htmlTemplate;
    constructor() {
        super();
        this.ctrlSeq = latex;
        this.htmlTemplate = `<${tagName} ${attrs}>&0</${tagName}>`;
        this.ariaLabel = ariaLabel;
    }
};
LatexCmds.em =
    LatexCmds.italic =
        LatexCmds.italics =
            LatexCmds.emph =
                LatexCmds.textit =
                    LatexCmds.textsl =
                        makeTextBlock('\\textit', 'i', 'class="mq-text-mode"', 'Italic');
LatexCmds.strong = LatexCmds.bold = LatexCmds.textbf = makeTextBlock('\\textbf', 'b', 'class="mq-text-mode"', 'Bold');
LatexCmds.sf = LatexCmds.textsf = makeTextBlock('\\textsf', 'span', 'class="mq-sans-serif mq-text-mode"', 'Sans serif font');
LatexCmds.tt = LatexCmds.texttt = makeTextBlock('\\texttt', 'span', 'class="mq-monospace mq-text-mode"', 'Mono space font');
LatexCmds.textsc = makeTextBlock('\\textsc', 'span', 'style="font-variant:small-caps" class="mq-text-mode"', 'Variable font');
LatexCmds.uppercase = makeTextBlock('\\uppercase', 'span', 'style="text-transform:uppercase" class="mq-text-mode"', 'Uppercase');
LatexCmds.lowercase = makeTextBlock('\\lowercase', 'span', 'style="text-transform:lowercase" class="mq-text-mode"', 'Lowercase');

// Deals with mouse events for clicking and drag-to-select.
const MouseEventController = (Base) => class extends Base {
    mouseDownHandler;
    delegateMouseEvents() {
        const ultimateRootEl = this.root.elements.firstElement;
        // Drag-to-select event handling
        this.mouseDownHandler = (e) => {
            const rootEl = e.target?.closest('.mq-root-block');
            const root = TNode.byId.get(parseInt((rootEl?.getAttribute(mqBlockId) || ultimateRootEl.getAttribute(mqBlockId)) ?? '0'));
            if (!root?.controller) {
                throw new Error('controller undefined... what?');
            }
            const ctrlr = root.controller, cursor = ctrlr.cursor, blink = cursor.blink;
            const textarea = ctrlr.textarea;
            e.preventDefault();
            if (cursor.options.ignoreNextMousedown(e))
                return;
            else
                cursor.options.ignoreNextMousedown = () => false;
            // End any previous selection.
            cursor.endSelection();
            // Cache the ownerDocument as it is not available in the mouseup handler if the mouse button is released
            // while the cursor is not in the window.
            const ownerDocument = e.target.ownerDocument;
            let target;
            const mousemove = (e) => (target = e.target);
            const docmousemove = (e) => {
                if (!cursor.anticursor)
                    cursor.startSelection();
                ctrlr.seek(target, e.pageX).cursor.select();
                target = undefined;
                if (cursor.selection)
                    ctrlr.aria
                        .clear()
                        .queue(cursor.selection.join('mathspeak') + ' selected')
                        .alert();
            };
            // Outside rootEl, the MathQuill node corresponding to the target (if any)
            // won't be inside this root.  So don't mislead Controller::seek with it.
            const mouseup = () => {
                cursor.blink = blink;
                if (!cursor.selection) {
                    if (ctrlr.editable) {
                        cursor.show();
                        if (cursor.parent)
                            ctrlr.aria.queue(cursor.parent).alert();
                    }
                }
                // Delete the mouse handlers now that the drag has ended.
                rootEl?.removeEventListener('mousemove', mousemove);
                ownerDocument.removeEventListener('mousemove', docmousemove);
                ownerDocument.removeEventListener('mouseup', mouseup);
            };
            if (e.detail === 3) {
                // If this is a triple click, then select all and return.
                ctrlr.selectAll();
                ctrlr.aria.alert();
                mouseup();
                return;
            }
            else if (e.detail === 2) {
                // If this is a double click, then select the block that is to the right of the cursor, and return.
                // Note that the interpretation of what a block is in this situation is not a true MathQuill block.
                // Rather an attempt is made to select word like blocks.
                ctrlr.seek(e.target, e.pageX);
                if (!cursor.right && cursor.left?.parent === root)
                    ctrlr.moveLeft();
                ctrlr.withIncrementalSelection((selectDir) => {
                    if (cursor.right instanceof Letter) {
                        // If a "Letter" is to the right of the cursor, then try to select all adjacent "Letter"s
                        // that are of the same basic ilk.  That means all "Letter"s that are part of an operator
                        // name, or all "Letter"s that are not part of an operator name.
                        const currentNode = cursor.right;
                        while (cursor.left &&
                            cursor.left instanceof Letter &&
                            cursor.left.isPartOfOperator === currentNode.isPartOfOperator)
                            ctrlr.moveLeft();
                        cursor.startSelection();
                        while (
                        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
                        cursor.right &&
                            cursor.right instanceof Letter &&
                            cursor.right.isPartOfOperator === currentNode.isPartOfOperator)
                            selectDir('right');
                    }
                    else if (cursor.right instanceof Digit) {
                        // If a "Digit" is to the right of the cursor, then select all adjacent "Digit"s.
                        while (cursor.left && cursor.left instanceof Digit)
                            ctrlr.moveLeft();
                        cursor.startSelection();
                        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
                        while (cursor.right && cursor.right instanceof Digit)
                            selectDir('right');
                    }
                    else {
                        cursor.startSelection();
                        selectDir('right');
                    }
                    // If the cursor is in a text block, then select the whole text block.
                    if (cursor.left?.parent instanceof TextBlock) {
                        cursor.left.parent.moveOutOf('left', cursor);
                        selectDir('right');
                    }
                });
                ctrlr.aria.alert();
                mouseup();
                return;
            }
            if (ctrlr.blurred)
                textarea?.focus();
            cursor.blink = noop;
            ctrlr.seek(e.target, e.pageX).cursor.startSelection();
            rootEl?.addEventListener('mousemove', mousemove);
            ownerDocument.addEventListener('mousemove', docmousemove);
            ownerDocument.addEventListener('mouseup', mouseup);
            // Listen on document not just body to not only hear about mousemove and
            // mouseup on page outside field, but even outside page (except iframes).
        };
        this.container.addEventListener('mousedown', this.mouseDownHandler);
    }
    seek(target, pageX) {
        const cursor = this.notify('select').cursor;
        let nodeId = 0;
        if (target) {
            nodeId = parseInt((target.getAttribute(mqBlockId) || target.getAttribute(mqCmdId)) ?? '0');
            if (!nodeId) {
                const targetParent = target.parentElement;
                nodeId = parseInt((targetParent?.getAttribute(mqBlockId) || targetParent?.getAttribute(mqCmdId)) ?? '0');
            }
        }
        const node = nodeId ? TNode.byId.get(nodeId) : this.root;
        if (!node)
            throw new Error('nodeId is not the id of a TNode that exists');
        // Don't clear the selection until after getting node from target, in case
        // target was selection span.  Otherwise target will have no parent and will
        // seek from root, which is less accurate (e.g. fraction).
        cursor.clearSelection().show();
        node.seek(pageX, cursor);
        // Before .selectFrom when mouse-selecting, so
        // always hits no-selection case in scrollHoriz and scrolls slower
        this.scrollHoriz();
        return this;
    }
};

// Extension to the Controller that allows exporting of math in a human-readable text format.
const ExportText = (Base) => class extends Base {
    exportText() {
        return this.root.foldChildren('', (text, child) => text + child.text());
    }
};

// Manage the MathQuill instance's textarea (as owned by the Controller)
const generateUUID = () => {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, (c) => {
        const r = (Math.random() * 16) | 0;
        const v = c === 'x' ? r : (r & 0x3) | 0x8;
        return v.toString(16);
    });
};
const TextAreaController = (Base) => class extends Base {
    textareaSelectionTimeout;
    selectFn;
    createTextarea() {
        this.textareaSpan = document.createElement('span');
        this.textareaSpan.classList.add('mq-textarea');
        const tabbable = typeof this.options.tabbable === 'boolean' ? this.options.tabbable : this.KIND_OF_MQ !== 'StaticMath';
        const textarea = this.options.substituteTextarea(tabbable);
        if (!textarea.nodeType)
            throw new Error('substituteTextarea() must return a DOM element');
        // aria-hide noninteractive textarea element for static math
        if (!this.options.tabbable && this.KIND_OF_MQ === 'StaticMath')
            textarea.setAttribute('aria-hidden', 'true');
        this.textareaSpan.append(textarea);
        this.textarea = textarea;
        if (!this.mathspeakSpan) {
            this.mathspeakId = generateUUID();
            this.mathspeakSpan = document.createElement('span');
            this.mathspeakSpan.classList.add('mq-mathspeak');
            this.mathspeakSpan.id = this.mathspeakId;
            this.textareaSpan.prepend(this.mathspeakSpan);
        }
        if (this.mathspeakId)
            textarea.setAttribute('aria-labelledby', this.mathspeakId);
        if (tabbable)
            this.mathspeakSpan.setAttribute('aria-hidden', 'true');
        this.cursor.selectionChanged = () => {
            this.selectionChanged();
        };
    }
    selectionChanged() {
        // throttle calls to setTextareaSelection(), because setting textarea.value
        // and/or calling textarea.select() can have anomalously bad performance:
        // https://github.com/mathquill/mathquill/issues/43#issuecomment-1399080
        if (!this.textareaSelectionTimeout) {
            this.textareaSelectionTimeout = setTimeout(() => {
                this.setTextareaSelection();
            });
        }
    }
    setTextareaSelection() {
        delete this.textareaSelectionTimeout;
        let latex = '';
        if (this.cursor.selection) {
            latex = this.cursor.selection.join('latex');
            if (this.options.statelessClipboard) {
                // FIXME: like paste, only this works for math fields; should ask parent
                latex = `$${latex}$`;
            }
        }
        this.selectFn?.(latex);
    }
    staticMathTextareaEvents() {
        this.textarea?.addEventListener('copy', () => {
            this.setTextareaSelection();
        });
        this.selectFn = (text) => {
            if (this.textarea)
                this.textarea.value = text;
            if (text)
                this.textarea?.select();
        };
        this.container.prepend(this.textareaSpan);
        this.addStaticFocusBlurEvents();
    }
    editablesTextareaEvents() {
        if (this.textarea) {
            const { select } = saneKeyboardEvents(this.textarea, this);
            this.selectFn = select;
        }
        this.container.prepend(this.textareaSpan);
        this.addEditableFocusBlurEvents();
        this.updateMathspeak();
    }
    unbindEditablesEvents() {
        this.selectFn = (text) => {
            if (this.textarea)
                this.textarea.value = text;
            if (text)
                this.textarea?.select();
        };
        this.textareaSpan?.remove();
        this.unbindEditableFocusBlurEvents();
        this.blurred = true;
    }
    keystroke(key, evt) {
        this.cursor.parent?.keystroke(key, evt, this);
    }
    typedText(ch) {
        if (ch === '\n') {
            this.handle('enter');
            return;
        }
        const cursor = this.notify().cursor;
        cursor.parent?.write(cursor, ch);
        this.scrollHoriz();
    }
    cut() {
        if (this.cursor.selection) {
            setTimeout(() => {
                this.notify('edit'); // deletes selection if present
                this.cursor.parent?.bubble('reflow');
            });
        }
    }
    copy() {
        this.setTextareaSelection();
    }
    paste(text) {
        // TODO: document `statelessClipboard` config option in README, after
        // making it work like it should, that is, in both text and math mode
        // (currently only works in math fields, so worse than pointless, it
        //  only gets in the way by \text{}-ifying pasted stuff and $-ifying
        //  cut/copied LaTeX)
        if (this.options.statelessClipboard) {
            if (text.startsWith('$') && text.endsWith('$')) {
                text = text.slice(1, -1);
            }
            else {
                text = `\\text{${text}}`;
            }
        }
        // FIXME: this always inserts math or a TextBlock, even in a RootTextBlock
        this.writeLatex(text).cursor.show();
    }
    setupStaticField() {
        this.updateMathspeak();
        this.cursor.hide().parent?.blur(this.cursor);
    }
    updateMathspeak(emptyContent = false) {
        // If the controller's ARIA label doesn't end with a punctuation mark, add a colon by default to better
        // separate it from mathspeak.
        const ariaLabel = this.getAriaLabel();
        const labelWithSuffix = /[A-Za-z0-9]$/.test(ariaLabel) ? ariaLabel + ':' : ariaLabel;
        const mathspeak = this.root.mathspeak().trim();
        this.aria.clear(emptyContent);
        // For static math, provide mathspeak in a visually hidden span to allow screen readers and other AT to
        // traverse the content.  For editable math, assign the mathspeak to the textarea's ARIA label (AT can use
        // text navigation to interrogate the content).  Be certain to include the mathspeak for only one of these,
        // though, as we don't want to include outdated labels if a field's editable state changes.  By design, also
        // take careful note that the ariaPostLabel is meant to exist only for editable math (e.g. to serve as an
        // evaluation or error message) so it is not included for static math mathspeak calculations.  The
        // mathspeakSpan should exist only for static math, so we use its presence to decide which approach to take.
        if (this.mathspeakSpan)
            this.mathspeakSpan.textContent = (labelWithSuffix + ' ' + mathspeak).trim();
    }
};

// Add the capability for MathQuill to generate ARIA alerts.
class Aria {
    controller;
    span = document.createElement('span');
    msg = '';
    items = [];
    constructor(controller) {
        this.controller = controller;
        this.span.classList.add('mq-aria-alert');
        this.span.setAttribute('aria-live', 'assertive');
        this.span.setAttribute('aria-atomic', 'true');
    }
    attach() {
        const container = this.controller.container;
        if (this.span.parentNode !== container)
            container.prepend(this.span);
    }
    queue(item, shouldDescribe = false) {
        let output = '';
        if (item instanceof TNode) {
            // Some constructs include verbal shorthand (such as simple fractions and exponents).  Since ARIA alerts
            // relate to moving through interactive content, we don't want to use that shorthand if it exists since
            // doing so may be ambiguous or confusing.
            const itemMathspeak = item.mathspeak({ ignoreShorthand: true });
            if (shouldDescribe) {
                // Used to ensure item is described when cursor reaches block boundaries.
                if (item.parent?.ariaLabel && item.ariaLabel === 'block') {
                    output = `${item.parent.ariaLabel} ${itemMathspeak}`;
                }
                else if (item.ariaLabel) {
                    output = `${item.ariaLabel} ${itemMathspeak}`;
                }
            }
            if (output === '')
                output = itemMathspeak;
        }
        else {
            output = item || '';
        }
        this.items.push(output);
        return this;
    }
    queueDirOf(dir) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        return this.queue(dir === 'left' ? 'before' : 'after');
    }
    queueDirEndOf(dir) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        return this.queue(dir === 'left' ? 'beginning of' : 'end of');
    }
    alert(t) {
        this.attach();
        if (t)
            this.queue(t);
        if (this.items.length) {
            // To cut down on potential verbiage from multiple Mathquills firing near-simultaneous ARIA alerts, update
            // the text of this instance if its container also has keyboard focus.  If it does not, leave the DOM
            // unchanged but flush the queue regardless.
            // Note: The msg variable is updated regardless of focus for unit tests.
            this.msg = this.items
                .join(' ')
                .replace(/ +(?= )/g, '')
                .trim();
            if (this.controller.containerHasFocus())
                this.span.textContent = this.msg;
        }
        return this.clear();
    }
    // Clear out the internal alert message queue.  If emptyContent is set, also clear the text content for the alert
    // element (typically when the focused field has been blurred) so that stale alert text is not hanging around.
    clear(emptyContent = false) {
        this.items.length = 0;
        if (emptyContent)
            this.span.textContent = '';
        return this;
    }
}

// Controller for a MathQuill instance, on which services are registered.
class ControllerBase {
    id;
    root;
    container;
    options;
    cursor;
    apiClass;
    KIND_OF_MQ = '';
    editable = false;
    blurred;
    textareaSpan;
    textarea;
    mathspeakSpan;
    mathspeakId;
    aria;
    ariaLabel;
    ariaPostLabel;
    _ariaAlertTimeout;
    constructor(root, container, options) {
        this.id = root.id;
        this.root = root;
        this.container = container;
        this.options = options;
        this.cursor = new Cursor(root, options, this);
        this.aria = new Aria(this);
        this.ariaLabel = 'Math Input';
        this.ariaPostLabel = '';
    }
    handle(name, dir) {
        const handlers = this.options.handlers;
        if (handlers?.[name]) {
            if (dir === 'left' || dir === 'right')
                handlers[name](dir, this.apiClass);
            else
                handlers[name](this.apiClass);
        }
    }
    notify(e) {
        if (e === 'move' || e === 'upDown')
            this.cursor.show().clearSelection();
        if (e !== 'upDown')
            this.cursor.upDownCache = {};
        if (e === 'edit')
            this.cursor.show().deleteSelection();
        if (e !== 'select')
            this.cursor.endSelection();
        return this;
    }
    // The textarea mixin overrides this.
    selectionChanged() {
        /* do nothing */
    }
    containerHasFocus() {
        return document.activeElement && this.container.contains(document.activeElement);
    }
    setAriaLabel(ariaLabel) {
        const oldAriaLabel = this.getAriaLabel();
        if (typeof ariaLabel === 'string' && ariaLabel !== '')
            this.ariaLabel = ariaLabel;
        else if (this.editable)
            this.ariaLabel = 'Math Input';
        else
            this.ariaLabel = '';
        // If this field does not have focus, update its computed mathspeak value.  Check for focus because updating the
        // aria-label attribute of a focused element will cause most screen readers to announce the new value.  If the
        // field does have focus at the time, it will be updated once a blur event occurs.
        if (this.ariaLabel !== oldAriaLabel && !this.containerHasFocus())
            this.updateMathspeak();
        return this;
    }
    getAriaLabel() {
        if (this.ariaLabel !== 'Math Input')
            return this.ariaLabel;
        else if (this.editable)
            return 'Math Input';
        else
            return '';
    }
    setAriaPostLabel(ariaPostLabel, timeout) {
        if (typeof ariaPostLabel === 'string' && ariaPostLabel !== '') {
            if (ariaPostLabel !== this.ariaPostLabel && typeof timeout === 'number') {
                if (this._ariaAlertTimeout)
                    clearTimeout(this._ariaAlertTimeout);
                this._ariaAlertTimeout = setTimeout(() => {
                    if (this.containerHasFocus()) {
                        // Voice the new label, but do not update mathspeak content to prevent double-speech.
                        this.aria.alert(this.root.mathspeak().trim() + ' ' + ariaPostLabel.trim());
                    }
                    else {
                        // This mathquill does not have focus, so update its mathspeak.
                        this.updateMathspeak();
                    }
                }, timeout);
            }
            this.ariaPostLabel = ariaPostLabel;
        }
        else {
            if (this._ariaAlertTimeout)
                clearTimeout(this._ariaAlertTimeout);
            this.ariaPostLabel = '';
        }
        return this;
    }
    getAriaPostLabel() {
        return this.ariaPostLabel || '';
    }
    exportMathSpeak() {
        return this.root.mathspeak();
    }
    updateMathspeak(_emptyContent = false) {
        // This is defined here so that it can be called above without jumping through a lot of hoops to pacify
        // typescript.  The method defined in services/textarea.ts will override this, and it is what will actually be
        // called above.
    }
}
class Controller extends ExportText(TextAreaController(LatexControllerExtension(FocusBlurEvents(MouseEventController(HorizontalScroll(ControllerBase)))))) {
    constructor(root, container, options) {
        super(root, container, options);
        root.controller = this;
    }
    escapeDir(dir, _key, e) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        const cursor = this.cursor;
        // only prevent default of Tab if not in the root editable
        if (cursor.parent !== this.root)
            e.preventDefault();
        // want to be a noop if in the root editable (in fact, Tab has an unrelated
        // default browser action if so)
        if (cursor.parent === this.root)
            return;
        cursor.parent?.moveOutOf(dir, cursor);
        this.aria.alert();
        return this.notify('move');
    }
    moveDir(dir) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        const cursor = this.cursor, updown = cursor.options.leftRightIntoCmdGoes;
        if (cursor.selection) {
            const end = cursor.selection.ends[dir];
            if (end)
                cursor.insDirOf(dir, end);
        }
        else if (cursor[dir])
            cursor[dir].moveTowards(dir, cursor, updown);
        else
            cursor.parent?.moveOutOf(dir, cursor, updown);
        return this.notify('move');
    }
    moveLeft() {
        return this.moveDir('left');
    }
    moveRight() {
        return this.moveDir('right');
    }
    // moveUp and moveDown have almost identical algorithms:
    // - first check left and right, if so insAtLeft/RightEnd of them
    // - else check the parent's 'upOutOf'/'downOutOf' property:
    //   + if it's a function, call it with the cursor as the sole argument and
    //     use the return value as if it were the value of the property
    //   + if it's a TNode, jump up or down into it:
    //     - if there is a cached Point in the block, insert there
    //     - else, seekHoriz within the block to the current x-coordinate (to be
    //       as close to directly above/below the current position as possible)
    //   + unless it's exactly `true`, stop bubbling
    moveUpDown(dir) {
        const cursor = this.notify('upDown').cursor;
        const dirInto = `${dir}Into`, dirOutOf = `${dir}OutOf`;
        if (cursor.right?.[dirInto])
            cursor.insAtLeftEnd(cursor.right[dirInto]);
        else if (cursor.left?.[dirInto])
            cursor.insAtRightEnd(cursor.left[dirInto]);
        else {
            cursor.parent?.bubble((ancestor) => {
                if (ancestor[dirOutOf]) {
                    if (typeof ancestor[dirOutOf] === 'function')
                        ancestor[dirOutOf](cursor);
                    if (ancestor[dirOutOf] instanceof TNode)
                        cursor.jumpUpDown(ancestor, ancestor[dirOutOf]);
                    if (ancestor[dirOutOf] !== true)
                        return false;
                }
                return true;
            });
        }
        return this;
    }
    moveUp() {
        return this.moveUpDown('up');
    }
    moveDown() {
        return this.moveUpDown('down');
    }
    deleteDir(dir) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        const cursor = this.cursor;
        // FIXME: This should be done in the methods of the objects that need this.
        if (cursor[dir]) {
            if (cursor[dir] instanceof Bracket) {
                if (cursor[dir].parent) {
                    this.aria.queue(cursor[dir].parent
                        .chToCmd(cursor[dir].sides[otherDir(dir)].ch, cursor.options)
                        .mathspeak({ createdLeftOf: cursor }));
                }
                // Speak the current element if it has no blocks, but don't for text block commands as the
                // deleteTowards method in the TextCommand class is responsible for speaking the new character under the
                // cursor.
            }
            else if (cursor[dir] instanceof MathCommand &&
                !cursor[dir].blocks.length &&
                cursor[dir].parent?.ctrlSeq !== '\\text') {
                this.aria.queue(cursor[dir]);
            }
        }
        else if (cursor.parent?.parent && cursor.parent.parent instanceof TNode) {
            if (cursor.parent.parent instanceof Bracket) {
                if (cursor.parent.parent.parent) {
                    this.aria.queue(cursor.parent.parent.parent
                        .chToCmd(cursor.parent.parent.sides[dir].ch, cursor.options)
                        .mathspeak({ createdLeftOf: cursor }));
                }
            }
            else if (cursor.parent.parent instanceof MathCommand &&
                cursor.parent.parent.blocks.length &&
                cursor.parent.parent.mathspeakTemplate.length) {
                if (cursor.parent.parent.upInto && cursor.parent.parent.downInto) {
                    // likely a fraction, and we just backspaced over the slash
                    this.aria.queue(cursor.parent.parent.mathspeakTemplate[1]);
                }
                else {
                    this.aria.queue(dir === 'left'
                        ? cursor.parent.parent.mathspeakTemplate[0]
                        : (cursor.parent.parent.mathspeakTemplate.at(-1) ?? ''));
                }
            }
            else {
                this.aria.queue(cursor.parent.parent);
            }
        }
        const hadSelection = cursor.selection;
        this.notify('edit'); // Shows the cursor and deletes a selection if present.
        if (!hadSelection) {
            if (cursor[dir])
                cursor[dir].deleteTowards(dir, cursor);
            else
                cursor.parent?.deleteOutOf(dir, cursor);
        }
        // Call the contactWeld for a SupSub so that it can deal with having its base deleted.
        cursor.right?.postOrder('contactWeld', cursor);
        cursor.left?.siblingDeleted?.(cursor.options, 'right');
        cursor.right?.siblingDeleted?.(cursor.options, 'left');
        cursor.parent?.bubble('reflow');
        return this;
    }
    ctrlDeleteDir(dir) {
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        const cursor = this.cursor;
        if (!cursor[dir] || cursor.selection)
            return this.deleteDir(dir);
        this.notify('edit');
        let fragmentRemoved;
        if (dir === 'left')
            fragmentRemoved = new Fragment(cursor.parent?.ends.left, cursor.left).remove();
        else
            fragmentRemoved = new Fragment(cursor.right, cursor.parent?.ends.right).remove();
        cursor.controller.aria.queue(fragmentRemoved);
        if (cursor.parent)
            cursor.insAtDirEnd(dir, cursor.parent);
        // Call the contactWeld for a SupSub so that it can deal with having its base deleted.
        cursor.right?.postOrder('contactWeld', cursor);
        cursor.left?.siblingDeleted?.(cursor.options, 'right');
        cursor.right?.siblingDeleted?.(cursor.options, 'left');
        cursor.parent?.bubble('reflow');
        return this;
    }
    backspace() {
        return this.deleteDir('left');
    }
    deleteForward() {
        return this.deleteDir('right');
    }
    incrementalSelectionOpen = false;
    // startIncrementalSelection, selectDirIncremental, and finishIncrementalSelection should only be called by
    // withIncrementalSelection because they must be called in sequence.
    // Start a selection.
    startIncrementalSelection() {
        if (this.incrementalSelectionOpen)
            throw new Error('multiple selections cannot be simultaneously open');
        this.incrementalSelectionOpen = true;
        this.notify('select');
        if (!this.cursor.anticursor)
            this.cursor.startSelection();
    }
    // Update the selection model stored in the cursor without modifying the selection DOM.
    selectDirIncremental(dir) {
        if (!this.incrementalSelectionOpen)
            throw new Error('a selection is not open');
        if (dir !== 'left' && dir !== 'right')
            throw new Error('a direction was not passed');
        const cursor = this.cursor, seln = cursor.selection;
        const node = cursor[dir];
        if (node) {
            // if node we're selecting towards is inside selection (hence retracting)
            // and is on the *far side* of the selection (hence is only node selected)
            // and the anticursor is *inside* that node, not just on the other side
            if (seln && seln.ends[dir] === node && cursor.anticursor?.[otherDir(dir)] !== node) {
                node.unselectInto(dir, cursor);
            }
            else
                node.selectTowards(dir, cursor);
        }
        else
            cursor.parent?.selectOutOf(dir, cursor);
    }
    // Update selection DOM to match cursor model.
    finishIncrementalSelection() {
        if (!this.incrementalSelectionOpen)
            throw new Error('a selection is not open');
        this.cursor.clearSelection();
        if (!this.cursor.select())
            this.cursor.show();
        if (this.cursor.selection) {
            // Clear first.  A selection can fire several times, and if not cleared it would result in repeated speech.
            this.aria.clear().queue(this.cursor.selection.join('mathspeak', ' ').trim() + ' selected');
        }
        this.incrementalSelectionOpen = false;
    }
    // Used to build a selection incrementally in a loop. Calls the passed callback with a selectDir function that may
    // be called many times, and defers updating the view until the incremental selection is complete.
    //
    // Wraps up calling
    //
    //     this.startIncrementalSelection()
    //     this.selectDirIncremental(dir) // possibly many times
    //     this.finishIncrementalSelection()
    //
    // with extra error handling and invariant enforcement.
    withIncrementalSelection(cb) {
        try {
            this.startIncrementalSelection();
            try {
                cb((dir) => {
                    this.selectDirIncremental(dir);
                });
            }
            finally {
                // Since a selection has been started, attempt to finish it even if the callback throws an error.
                this.finishIncrementalSelection();
            }
        }
        finally {
            // Mark the selection as closed even if finishIncrementalSelection throws an error. Makes a possible error
            // in finishIncrementalSelection more recoverable.
            this.incrementalSelectionOpen = false;
        }
    }
    selectDir(dir) {
        this.withIncrementalSelection((selectDir) => {
            selectDir(dir);
        });
    }
    selectLeft() {
        this.selectDir('left');
    }
    selectRight() {
        this.selectDir('right');
    }
    selectAll() {
        this.notify('move').cursor.insAtRightEnd(this.root);
        while (this.cursor.left)
            this.selectLeft();
        this.withIncrementalSelection((selectDir) => {
            while (this.cursor.left)
                selectDir('left');
        });
    }
    selectToBlockEndInDir(dir) {
        this.withIncrementalSelection((selectDir) => {
            while (this.cursor[dir])
                selectDir(dir);
        });
    }
    selectToRootEndInDir(dir) {
        const cursor = this.cursor;
        this.withIncrementalSelection((selectDir) => {
            while (cursor[dir] || cursor.parent !== this.root) {
                selectDir(dir);
            }
        });
    }
}

class AbstractMathQuill {
    __controller;
    __options;
    id;
    revert;
    static RootBlock;
    constructor(ctrlr) {
        this.__controller = ctrlr;
        this.__controller.apiClass = this;
        this.__options = ctrlr.options;
        this.id = ctrlr.id;
    }
    __mathquillify(...classNames) {
        const root = this.__controller.root, el = this.__controller.container;
        this.__controller.createTextarea();
        el.classList.add(...classNames);
        const contents = Array.from(el.childNodes).map((child) => el.removeChild(child));
        const rootEl = document.createElement('span');
        rootEl.classList.add('mq-root-block');
        rootEl.setAttribute(mqBlockId, root.id.toString());
        rootEl.setAttribute('aria-hidden', 'true');
        root.elements.add(rootEl);
        el.append(rootEl);
        this.latex(contents.reduce((ret, child) => (child.nodeType === 8 ? ret : `${ret}${child.textContent ?? ''}`), ''));
        this.revert = () => {
            while (el.firstChild)
                el.firstChild.remove();
            el.classList.remove('mq-editable-field', 'mq-math-mode', 'mq-text-mode');
            if (this.__controller.mouseDownHandler)
                el.removeEventListener('mousedown', this.__controller.mouseDownHandler);
            el.append(...contents);
            return el;
        };
        return this;
    }
    get options() {
        return this.__options;
    }
    config(opts) {
        Options.config(this.__options, opts);
        return this;
    }
    el() {
        return this.__controller.container;
    }
    text() {
        return this.__controller.exportText();
    }
    latex(latex) {
        if (typeof latex !== 'undefined') {
            this.__controller.renderLatexMath(latex);
            if (this.__controller.blurred)
                this.__controller.cursor.hide().parent?.blur();
            return this;
        }
        return this.__controller.exportLatex();
    }
    html() {
        return this.__controller.root.elements
            .html()
            .replace(new RegExp(` (?:${mqBlockId}|${mqCmdId})="?\\d+"?`, 'g'), '')
            .replace(/<span class="?mq-cursor( mq-blink)?"?>.?<\/span>/i, '')
            .replace(/ mq-has-cursor|mq-has-cursor ?/, '')
            .replace(/ class=(""|(?= |>))/g, '');
    }
    reflow() {
        this.__controller.root.postOrder('reflow');
        return this;
    }
    focus() {
        if (document.activeElement === this.__controller.textarea)
            this.__controller.textarea.dispatchEvent(new FocusEvent('focus'));
        else
            this.__controller.textarea?.focus();
        return this;
    }
    blur() {
        if (document.activeElement !== this.__controller.textarea)
            this.__controller.textarea?.dispatchEvent(new FocusEvent('blur'));
        else
            this.__controller.textarea.blur();
        return this;
    }
    setAriaLabel(ariaLabel) {
        this.__controller.setAriaLabel(ariaLabel);
        return this;
    }
    getAriaLabel() {
        return this.__controller.getAriaLabel();
    }
    mathspeak() {
        return this.__controller.exportMathSpeak();
    }
}
class EditableField extends AbstractMathQuill {
    __mathquillify(...classNames) {
        super.__mathquillify(...classNames);
        this.__controller.editable = true;
        this.__controller.delegateMouseEvents();
        this.__controller.editablesTextareaEvents();
        return this;
    }
    write(latex) {
        this.__controller.writeLatex(latex);
        this.__controller.scrollHoriz();
        if (this.__controller.blurred)
            this.__controller.cursor.hide().parent?.blur();
        return this;
    }
    empty() {
        const root = this.__controller.root, cursor = this.__controller.cursor;
        root.eachChild('postOrder', 'dispose');
        delete root.ends.left;
        delete root.ends.right;
        root.elements.empty();
        delete cursor.selection;
        cursor.insAtRightEnd(root);
        return this;
    }
    cmd(cmd) {
        const ctrlr = this.__controller.notify(), cursor = ctrlr.cursor;
        if (/^\\[a-z]+$/i.test(cmd) && !cursor.isTooDeep()) {
            cmd = cmd.slice(1);
            const klass = LatexCmds[cmd];
            if (klass) {
                const newCmd = new klass(cmd);
                if (cursor.selection)
                    newCmd.replaces(cursor.replaceSelection());
                newCmd.createLeftOf(cursor.show());
                this.__controller.scrollHoriz();
            }
        }
        else
            cursor.parent?.write(cursor, cmd);
        if (ctrlr.blurred)
            cursor.hide().parent?.blur();
        return this;
    }
    select() {
        this.__controller.selectAll();
        return this;
    }
    clearSelection() {
        this.__controller.cursor.clearSelection();
        return this;
    }
    moveToDirEnd(dir) {
        this.__controller.notify('move').cursor.insAtDirEnd(dir, this.__controller.root);
        return this;
    }
    moveToLeftEnd() {
        return this.moveToDirEnd('left');
    }
    moveToRightEnd() {
        return this.moveToDirEnd('right');
    }
    keystroke(keys) {
        const keyList = keys.replace(/^\s+|\s+$/g, '').split(/\s+/);
        for (const key of keyList) {
            const noPreventDefaultEvent = new KeyboardEvent('noop');
            noPreventDefaultEvent.preventDefault = noop;
            this.__controller.keystroke(key, noPreventDefaultEvent);
        }
        return this;
    }
    typedText(text) {
        for (const char of text) {
            this.__controller.typedText(char);
        }
        return this;
    }
    dropEmbedded(pageX, pageY, options) {
        const el = document.elementFromPoint(pageX - window.scrollX, pageY - window.scrollY);
        this.__controller.seek(el, pageX);
        const cmd = new LatexCmds.embed().setOptions(options);
        cmd.createLeftOf(this.__controller.cursor);
    }
    clickAt(clientX, clientY, target) {
        target = target || document.elementFromPoint(clientX, clientY);
        const ctrlr = this.__controller, root = ctrlr.root;
        if (!root.elements.firstElement.contains(target))
            target = root.elements.firstElement;
        ctrlr.seek(target, clientX + window.scrollX);
        // Force blurred window behavior to prevention selection of all content.
        this.__controller.windowBlurred = true;
        if (ctrlr.blurred)
            this.focus();
        this.__controller.windowBlurred = false;
        return this;
    }
    ignoreNextMousedown(fn) {
        this.__controller.cursor.options.ignoreNextMousedown = fn;
        return this;
    }
    setAriaPostLabel(ariaPostLabel, timeout) {
        this.__controller.setAriaPostLabel(ariaPostLabel, timeout);
        return this;
    }
    getAriaPostLabel() {
        return this.__controller.getAriaPostLabel();
    }
}

class Store {
    push(value) {
        this[Object.keys(this).length] = value;
    }
    get length() {
        return Object.keys(this).length;
    }
    get(indexName) {
        for (const strIndex of Object.keys(this)) {
            const index = parseInt(strIndex);
            if (this[index].name === indexName)
                return this[index];
        }
    }
}
class StaticMath extends AbstractMathQuill {
    static RootBlock = MathBlock;
    innerFields;
    constructor(ctrlr) {
        super(ctrlr);
        this.__controller.root.postOrder('registerInnerField', (this.innerFields = new Store()), InnerMathField);
    }
    __mathquillify() {
        super.__mathquillify('mq-math-mode');
        this.__controller.setupStaticField();
        if (this.__options.mouseEvents)
            this.__controller.delegateMouseEvents();
        this.__controller.staticMathTextareaEvents();
        return this;
    }
    latex(latex) {
        const returned = typeof latex === 'string' ? super.latex(latex) : super.latex();
        if (typeof latex !== 'undefined') {
            this.__controller.root.postOrder('registerInnerField', (this.innerFields = new Store()), InnerMathField);
            // Force an ARIA label update to remain in sync with the new LaTeX value.
            this.__controller.updateMathspeak();
        }
        return returned;
    }
}
class MathField extends EditableField {
    static RootBlock = RootMathBlock;
    __mathquillify() {
        // Disable reflow during initialization.
        const reflowSave = this.__controller.root.reflow;
        this.__controller.root.reflow = noop;
        super.__mathquillify('mq-editable-field', 'mq-math-mode');
        if (reflowSave)
            this.__controller.root.reflow = reflowSave;
        else
            delete this.__controller.root.reflow;
        return this;
    }
}
class InnerMathField extends MathField {
    name = '';
    makeStatic() {
        this.__controller.editable = false;
        this.__controller.root.blur();
        this.__controller.unbindEditablesEvents();
        this.__controller.container.classList.remove('mq-editable-field');
    }
    makeEditable() {
        this.__controller.editable = true;
        this.__controller.editablesTextareaEvents();
        this.__controller.cursor.insAtRightEnd(this.__controller.root);
        this.__controller.container.classList.add('mq-editable-field');
    }
}

// The block for abstract classes of text blocks
class RootTextBlock extends RootMathBlock {
    keystroke(key, e, ctrlr) {
        if (key === 'Spacebar' || key === 'Shift-Spacebar')
            return;
        super.keystroke(key, e, ctrlr);
        return;
    }
    write(cursor, ch) {
        cursor.show().deleteSelection();
        if (ch === '$')
            new RootMathCommand(cursor).createLeftOf(cursor);
        else {
            new VanillaSymbol(ch, ch === '<' ? '&lt;' : ch === '>' ? '&gt;' : undefined).createLeftOf(cursor);
        }
    }
}

// Abstract classes of text blocks
class TextField extends EditableField {
    static RootBlock = RootTextBlock;
    __mathquillify() {
        return super.__mathquillify('mq-editable-field', 'mq-text-mode');
    }
    latex(latex) {
        if (typeof latex !== 'undefined') {
            this.__controller.renderLatexText(latex);
            if (this.__controller.blurred)
                this.__controller.cursor.hide().parent?.blur();
            return this;
        }
        return this.__controller.exportLatex();
    }
}

// Commands and Operators.
class Style extends MathCommand {
    shouldNotSpeakDelimiters;
    constructor(ctrlSeq, tagName, attrs, ariaLabel, shouldNotSpeakDelimiters) {
        super(ctrlSeq, `<${tagName} ${attrs}>&0</${tagName}>`);
        this.ariaLabel = ariaLabel || ctrlSeq.replace(/^\\/, '');
        this.mathspeakTemplate = [`Start${this.ariaLabel},`, `End${this.ariaLabel}`];
        // In most cases, mathspeak should announce the start and end of style blocks.
        // There is one exception currently (mathrm).
        this.shouldNotSpeakDelimiters = shouldNotSpeakDelimiters;
    }
    mathspeak(opts) {
        if (!this.shouldNotSpeakDelimiters || opts?.ignoreShorthand)
            return super.mathspeak();
        return this.foldChildren('', (speech, block) => `${speech} ${block.mathspeak(opts)}`).trim();
    }
}
// fonts
LatexCmds.mathrm = bindMixin(Style, '\\mathrm', 'span', 'class="mq-roman mq-font"', 'Roman Font', true);
LatexCmds.mathit = bindMixin(Style, '\\mathit', 'i', 'class="mq-font"', 'Italic Font');
LatexCmds.mathbf = bindMixin(Style, '\\mathbf', 'b', 'class="mq-font"', 'Bold Font');
LatexCmds.mathsf = bindMixin(Style, '\\mathsf', 'span', 'class="mq-sans-serif mq-font"', 'Serif Font');
LatexCmds.mathtt = bindMixin(Style, '\\mathtt', 'span', 'class="mq-monospace mq-font"', 'Math Text');
// text-decoration
LatexCmds.underline = bindMixin(Style, '\\underline', 'span', 'class="mq-non-leaf mq-underline"', 'Underline');
LatexCmds.overline = LatexCmds.bar = bindMixin(Style, '\\overline', 'span', 'class="mq-non-leaf mq-overline"', 'Overline');
LatexCmds.overrightarrow = bindMixin(Style, '\\overrightarrow', 'span', 'class="mq-non-leaf mq-overarrow mq-arrow-right"', 'Over Right Arrow');
LatexCmds.overleftarrow = bindMixin(Style, '\\overleftarrow', 'span', 'class="mq-non-leaf mq-overarrow mq-arrow-left"', 'Over Left Arrow');
LatexCmds.overleftrightarrow = bindMixin(Style, '\\overleftrightarrow', 'span', 'class="mq-non-leaf mq-overarrow mq-arrow-both"', 'Over Left and Right Arrow');
LatexCmds.overarc = bindMixin(Style, '\\overarc', 'span', 'class="mq-non-leaf mq-overarc"', 'Over Arc');
LatexCmds.dot = class extends MathCommand {
    constructor() {
        super('\\dot', '<span class="mq-non-leaf"><span class="mq-dot-recurring-inner">' +
            '<span class="mq-dot-recurring">&#x2d9;</span>' +
            '<span class="mq-empty-box">&0</span>' +
            '</span></span>');
    }
};
// `\textcolor{color}{math}` will apply a color to the given math content, where `color` is any valid CSS Color Value.
LatexCmds.textcolor = class extends MathCommand {
    color;
    setColor(color) {
        this.color = color;
        this.htmlTemplate = `<span class="mq-textcolor" style="color:${color}">&0</span>`;
        this.ariaLabel = color.replace(/^\\/, '');
        this.mathspeakTemplate = [`Start ${this.ariaLabel},`, `End ${this.ariaLabel}`];
    }
    latex() {
        return `\\textcolor{${this.color ?? ''}}{${this.blocks[0].latex()}}`;
    }
    parser() {
        return Parser.optWhitespace
            .then(Parser.string('{'))
            .then(Parser.regex(/^[#\w\s.,()%-]*/))
            .skip(Parser.string('}'))
            .then((color) => {
            this.setColor(color);
            return super.parser();
        });
    }
    isStyleBlock() {
        return true;
    }
};
// Very similar to the \textcolor command, but will add the given CSS class.
// Usage: \class{classname}{math}
// Note regex that whitelists valid CSS classname characters:
// https://github.com/mathquill/mathquill/pull/191#discussion_r4327442
LatexCmds.class = class extends MathCommand {
    cls;
    parser() {
        return Parser.optWhitespace
            .then(Parser.string('{'))
            .then(Parser.regex(/^[-\w\s\\\xA0-\xFF]*/))
            .skip(Parser.string('}'))
            .then((cls) => {
            this.cls = cls || '';
            this.htmlTemplate = `<span class="mq-class ${cls}">&0</span>`;
            this.ariaLabel = cls + ' class';
            this.mathspeakTemplate = [`Start ${this.ariaLabel},`, `End ${this.ariaLabel}`];
            return super.parser();
        });
    }
    latex() {
        return `\\class{${this.cls ?? ''}}{${this.blocks[0].latex()}}`;
    }
    isStyleBlock() {
        return true;
    }
};
LatexCmds.subscript = LatexCmds._ = class extends SupSub {
    constructor() {
        super();
        this.supsub = 'sub';
        this.htmlTemplate =
            '<span class="mq-supsub mq-non-leaf">' +
                '<span class="mq-sub">&0</span>' +
                '<span style="display:inline-block;width:0">&#8203;</span>' +
                '</span>';
        this.textTemplate = ['_'];
        this.mathspeakTemplate = ['Subscript,', ', Baseline'];
        this.ariaLabel = 'subscript';
    }
    finalizeTree() {
        this.downInto = this.sub = this.ends.left;
        if (this.sub)
            this.sub.upOutOf = insLeftOfMeUnlessAtEnd;
        super.finalizeTree();
    }
};
LatexCmds.superscript =
    LatexCmds.supscript =
        LatexCmds['^'] =
            class extends SupSub {
                constructor() {
                    super();
                    this.supsub = 'sup';
                    this.htmlTemplate =
                        '<span class="mq-supsub mq-non-leaf mq-sup-only">' + '<span class="mq-sup">&0</span>' + '</span>';
                    this.textTemplate = ['^(', ')'];
                    this.ariaLabel = 'superscript';
                    this.mathspeakTemplate = ['Superscript,', ', Baseline'];
                }
                mathspeak(opts) {
                    // Simplify basic exponent speech for common whole numbers.
                    const child = this.upInto;
                    if (child) {
                        // Calculate this item's inner text to determine whether to shorten the returned speech.  Do not
                        // calculate its inner mathspeak until it is known that the speech is to be truncated.  Since the
                        // mathspeak computation is recursive, it should be called only once in this function to avoid
                        // performance bottlenecks.
                        const innerText = getCtrlSeqsFromBlock(child);
                        // If the superscript is a whole number, shorten the speech that is returned.
                        if (!opts?.ignoreShorthand && intRgx.test(innerText)) {
                            // Simple cases
                            if (innerText === '0')
                                return 'to the power of 0';
                            if (innerText === '1')
                                return 'to the first power';
                            else if (innerText === '2')
                                return 'squared';
                            else if (innerText === '3')
                                return 'cubed';
                            // More complex cases.
                            let suffix = '';
                            // Limit suffix addition to exponents < 1000.
                            if (/^[+-]?\d{1,3}$/.test(innerText)) {
                                if (/(11|12|13|4|5|6|7|8|9|0)$/.test(innerText))
                                    suffix = 'th';
                                else if (innerText.endsWith('1'))
                                    suffix = 'st';
                                else if (innerText.endsWith('2'))
                                    suffix = 'nd';
                                else if (innerText.endsWith('3'))
                                    suffix = 'rd';
                            }
                            return `to the ${typeof child === 'object' ? child.mathspeak() : innerText}${suffix} power`;
                        }
                    }
                    return super.mathspeak();
                }
                finalizeTree() {
                    this.upInto = this.sup = this.ends.right;
                    if (this.sup)
                        this.sup.downOutOf = insLeftOfMeUnlessAtEnd;
                    super.finalizeTree();
                }
            };
class SummationNotation extends UpperLowerLimitCommand {
    constructor(ch, html, ariaLabel) {
        super(ch, '<span class="mq-large-operator mq-non-leaf">' +
            '<span class="mq-to"><span>&1</span></span>' +
            `<big>${html}</big>` +
            '<span class="mq-from"><span>&0</span></span>' +
            '</span>', [ch.length > 1 ? ch.slice(1) : ch]);
        this.ariaLabel = ariaLabel || ch.replace(/^\\/, '');
    }
    createLeftOf(cursor) {
        super.createLeftOf(cursor);
        if (cursor.options.sumStartsWithNEquals && !this.replacedFragment) {
            new Letter('n').createLeftOf(cursor);
            new Equality().createLeftOf(cursor);
        }
    }
}
LatexCmds['\u2211'] = LatexCmds.sum = LatexCmds.summation = bindMixin(SummationNotation, '\\sum ', '&sum;');
LatexCmds['\u220f'] = LatexCmds.prod = LatexCmds.product = bindMixin(SummationNotation, '\\prod ', '&prod;', 'product');
LatexCmds.coprod = LatexCmds.coproduct = bindMixin(SummationNotation, '\\coprod ', '&#8720;', 'coproduct');
LatexCmds['\u222b'] =
    LatexCmds.int =
        LatexCmds.integral =
            class extends UpperLowerLimitCommand {
                constructor() {
                    super('\\int ', '<span class="mq-int mq-non-leaf">' +
                        '<big>&int;</big>' +
                        '<span class="mq-supsub mq-non-leaf">' +
                        '<span class="mq-sup"><span class="mq-sup-inner">&1</span></span>' +
                        '<span class="mq-sub">&0</span>' +
                        '<span style="display:inline-block;width:0">&#8203</span>' +
                        '</span>' +
                        '</span>');
                    this.ariaLabel = 'integral';
                }
            };
LatexCmds.frac = LatexCmds.dfrac = LatexCmds.cfrac = LatexCmds.fraction = Fraction;
const FractionChooseCreateLeftOfMixin = (Base) => class extends Base {
    createLeftOf(cursor) {
        if (!this.replacedFragment) {
            let leftward = cursor.left;
            while (leftward &&
                !(leftward instanceof BinaryOperator ||
                    ('text' in LatexCmds && leftward instanceof LatexCmds.text) ||
                    leftward instanceof UpperLowerLimitCommand ||
                    leftward.ctrlSeq === '\\ ' ||
                    /^[,;:]$/.test(leftward.ctrlSeq)) // lookbehind for operator
            )
                leftward = leftward.left;
            if (leftward instanceof UpperLowerLimitCommand && leftward.right instanceof SupSub) {
                leftward = leftward.right;
                if (leftward.right instanceof SupSub && leftward.right.ctrlSeq != leftward.ctrlSeq)
                    leftward = leftward.right;
            }
            if (leftward !== cursor.left &&
                !cursor.isTooDeep(1) &&
                !cursor.left?.elements.hasClass('mq-operator-name')) {
                this.replaces(new Fragment(leftward?.right || cursor.parent?.ends.left, cursor.left));
                cursor.left = leftward;
            }
        }
        super.createLeftOf(cursor);
    }
};
// LiveFraction
LatexCmds.over = CharCmds['/'] = class extends FractionChooseCreateLeftOfMixin(Fraction) {
};
// Note that the non-standard variants of these are no longer supported.  Those are cosec, cotan, and ctg, and all of
// the derived variants of those, as well as any of the ar${trigFunction}h derived variants.
for (const trigFunction of ['sin', 'cos', 'tan', 'sec', 'csc', 'cot']) {
    for (const trigVariant of [
        trigFunction,
        `arc${trigFunction}`,
        `a${trigFunction}`,
        `${trigFunction}h`,
        `arc${trigFunction}h`,
        `a${trigFunction}h`
    ]) {
        LatexCmds[trigVariant] = bindMixin(MathFunction, `\\${trigVariant}`);
    }
}
LatexCmds.exp = bindMixin(MathFunction, '\\exp');
LatexCmds.ln = bindMixin(MathFunction, '\\ln');
LatexCmds.log = class extends MathFunction {
    constructor() {
        super('\\log');
    }
    text() {
        const base = this.blocks[0].ends.left?.sub?.text() || '';
        const param = this.blocks[1].text() || '';
        const exponent = supSubText('^', this.blocks[0].ends.left?.sup);
        if (!base)
            return super.text();
        else if (base === '10') {
            return exponent ? `(log10(${param}))${exponent}` : `log10(${param})`;
        }
        else if (this.getController()?.options.logsChangeBase) {
            let leftward = this.left;
            for (; leftward && leftward.ctrlSeq === '\\ '; leftward = leftward.left)
                ;
            return exponent ||
                (leftward && !(leftward instanceof BinaryOperator)) ||
                (leftward instanceof BinaryOperator && leftward.isUnary)
                ? `(log(${param})/log(${base}))${exponent}`
                : `log(${param})/log(${base})`;
        }
        else {
            return exponent ? `(logb(${base},${param}))${exponent}` : `logb(${base},${param})`;
        }
    }
};
class SquareRoot extends MathCommand {
    constructor() {
        super();
        this.ctrlSeq = '\\sqrt';
        this.htmlTemplate =
            '<span class="mq-non-leaf mq-sqrt-container">' +
                '<span class="mq-scaled mq-sqrt-prefix">' +
                SVG_SYMBOLS.sqrt.html +
                '</span>' +
                '<span class="mq-non-leaf mq-sqrt-stem">&0</span>' +
                '</span>';
        this.textTemplate = ['sqrt(', ')'];
        this.mathspeakTemplate = ['StartSquareRoot,', ', EndSquareRoot'];
        this.ariaLabel = 'square root';
    }
    parser() {
        return latexMathParser.optBlock
            .then((optBlock) => {
            return latexMathParser.block.map((block) => {
                const nthroot = new NthRoot();
                nthroot.blocks = [optBlock, block];
                optBlock.adopt(nthroot);
                block.adopt(nthroot, optBlock);
                return nthroot;
            });
        })
            .or(super.parser());
    }
}
LatexCmds.sqrt = LatexCmds['\u221a'] = SquareRoot;
LatexCmds.hat = class extends MathCommand {
    constructor() {
        super('\\hat', '<span class="mq-non-leaf">' +
            '<span class="mq-hat-prefix">^</span>' +
            '<span class="mq-hat-stem">&0</span>' +
            '</span>', ['hat(', ')']);
    }
};
class NthRoot extends SquareRoot {
    constructor() {
        super();
        this.htmlTemplate =
            '<span class="mq-nthroot-container mq-non-leaf">' +
                '<span class="mq-nthroot mq-non-leaf">&0</span>' +
                '<span class="mq-scaled mq-sqrt-container">' +
                '<span class="mq-scaled mq-sqrt-prefix">' +
                SVG_SYMBOLS.sqrt.html +
                '</span>' +
                '<span class="mq-sqrt-stem mq-non-leaf">&1</span>' +
                '</span>' +
                '</span>';
        this.textTemplate = ['root(', ',', ')'];
        this.ariaLabel = 'root';
    }
    latex() {
        return `\\sqrt[${this.ends.left?.latex() ?? ''}]{${this.ends.right?.latex() ?? ''}}`;
    }
    text() {
        const index = this.ends.left?.text() ?? '';
        if (index === '' || index === '2')
            return `sqrt(${this.ends.right?.text() ?? ''})`;
        if (this.getController()?.options.rootsAreExponents) {
            const isSupR = this.right instanceof SupSub && this.right.supsub === 'sup';
            return `${isSupR ? '(' : ''}(${this.ends.right?.text() ?? ''})^(1/${index})${isSupR ? ')' : ''}`;
        }
        return `root(${index},${this.ends.right?.text() ?? ''})`;
    }
    mathspeak() {
        const indexMathspeak = this.ends.left?.mathspeak() ?? '';
        const radicandMathspeak = this.ends.right?.mathspeak() ?? '';
        if (this.ends.left)
            this.ends.left.ariaLabel = 'Index';
        if (this.ends.right)
            this.ends.right.ariaLabel = 'Radicand';
        if (indexMathspeak === '3')
            return `Start Cube Root, ${radicandMathspeak}, End Cube Root`;
        else
            return `Root Index ${indexMathspeak}, Start Root, ${radicandMathspeak}, End Root`;
    }
}
LatexCmds.root = LatexCmds.nthroot = NthRoot;
class DiacriticAbove extends MathCommand {
    constructor(ctrlSeq, symbol, textTemplate) {
        super(ctrlSeq, '<span class="mq-non-leaf">' +
            `<span class="mq-diacritic-above">${symbol}</span>` +
            '<span class="mq-diacritic-stem">&0</span>' +
            '</span>', textTemplate);
    }
}
LatexCmds.vec = bindMixin(DiacriticAbove, '\\vec', '&rarr;', ['vec(', ')']);
LatexCmds.tilde = bindMixin(DiacriticAbove, '\\tilde', '~', ['tilde(', ')']);
const bindCharBracketPair = (open, ctrlSeq) => {
    const curCtrlSeq = ctrlSeq || open, close = OPP_BRACKS[open], end = OPP_BRACKS[curCtrlSeq];
    CharCmds[open] = bindMixin(Bracket, 'left', open, close, curCtrlSeq, end);
    CharCmds[close] = bindMixin(Bracket, 'right', open, close, curCtrlSeq, end);
};
bindCharBracketPair('(');
bindCharBracketPair('[');
bindCharBracketPair('{', '\\{');
LatexCmds.langle = bindMixin(Bracket, 'left', '&lang;', '&rang;', '\\langle ', '\\rangle ');
LatexCmds.rangle = bindMixin(Bracket, 'right', '&lang;', '&rang;', '\\langle ', '\\rangle ');
LatexCmds.abs = CharCmds['|'] = bindMixin(Bracket, 'left', '|', '|', '|', '|');
LatexCmds.lVert = bindMixin(Bracket, 'left', '&#8741;', '&#8741;', '\\lVert ', '\\rVert ');
LatexCmds.rVert = bindMixin(Bracket, 'right', '&#8741;', '&#8741;', '\\lVert ', '\\rVert ');
LatexCmds.left = class extends MathCommand {
    parser() {
        return Parser.optWhitespace
            .then(Parser.regex(/^(?:[([|]|\\\{|\\langle(?![a-zA-Z])|\\lVert(?![a-zA-Z]))/))
            .then((ctrlSeq) => {
            let open = ctrlSeq.startsWith('\\') ? ctrlSeq.slice(1) : ctrlSeq;
            if (ctrlSeq == '\\langle') {
                open = '&lang;';
                ctrlSeq = ctrlSeq + ' ';
            }
            if (ctrlSeq == '\\lVert') {
                open = '&#8741;';
                ctrlSeq = ctrlSeq + ' ';
            }
            return latexMathParser.then((block) => {
                return Parser.string('\\right')
                    .skip(Parser.optWhitespace)
                    .then(Parser.regex(/^(?:[\])|]|\\\}|\\rangle(?![a-zA-Z])|\\rVert(?![a-zA-Z]))/))
                    .map((end) => {
                    let close = end.startsWith('\\') ? end.slice(1) : end;
                    if (end == '\\rangle') {
                        close = '&rang;';
                        end = `${end} `;
                    }
                    if (end == '\\rVert') {
                        close = '&#8741;';
                        end = `${end} `;
                    }
                    const cmd = new Bracket(undefined, open, close, ctrlSeq, end);
                    cmd.blocks = [block];
                    block.adopt(cmd);
                    return cmd;
                });
            });
        });
    }
};
LatexCmds.right = class extends MathCommand {
    parser() {
        return Parser.fail('unmatched \\right');
    }
};
class Binomial extends DelimsMixin(MathCommand) {
    constructor() {
        super('\\binom', '<span class="mq-bracket-container mq-non-leaf">' +
            `<span class="mq-paren mq-bracket-l mq-scaled" style="width:${SVG_SYMBOLS['('].width}">` +
            SVG_SYMBOLS['('].html +
            '</span>' +
            `<span class="mq-bracket-middle mq-non-leaf" style="margin-left:${SVG_SYMBOLS['('].width};margin-right:${SVG_SYMBOLS[')'].width}">` +
            '<span class="mq-array mq-non-leaf">' +
            '<span>&0</span>' +
            '<span>&1</span>' +
            '</span>' +
            '</span>' +
            `<span class="mq-paren mq-bracket-r mq-scaled" style="width:${SVG_SYMBOLS[')'].width}">` +
            SVG_SYMBOLS[')'].html +
            '</span>' +
            '</span>', ['choose(', ',', ')']);
        this.mathspeakTemplate = ['StartBinomial,', 'Choose', ', EndBinomial'];
        this.ariaLabel = 'binomial';
    }
    finalizeTree() {
        this.upInto = this.ends.left;
        this.downInto = this.ends.right;
        if (this.ends.right) {
            this.ends.right.upOutOf = this.ends.left;
            this.ends.right.ariaLabel = 'lower index';
        }
        if (this.ends.left) {
            this.ends.left.downOutOf = this.ends.right;
            this.ends.left.ariaLabel = 'upper index';
        }
    }
}
LatexCmds.binom = LatexCmds.binomial = Binomial;
LatexCmds.choose = class extends FractionChooseCreateLeftOfMixin(Binomial) {
};
// backcompat with before cfd3620 on #233
LatexCmds.editable = LatexCmds.MathQuillMathField = class extends MathCommand {
    name = '';
    field;
    constructor() {
        super('\\MathQuillMathField', '<span class="mq-editable-field">' + '<span class="mq-root-block" aria-hidden="true">&0</span>' + '</span>');
    }
    parser() {
        return Parser.string('[')
            .then(Parser.regex(/^[a-z][a-z0-9]*/i))
            .skip(Parser.string(']'))
            .map((name) => (this.name = name))
            .or(Parser.succeed())
            .then(super.parser());
    }
    finalizeTree(options) {
        if (!this.ends.left)
            throw new Error('Missing left end finalizing editable tree');
        const ctrlr = new Controller(this.ends.left, this.elements.firstElement, options);
        ctrlr.KIND_OF_MQ = 'MathField';
        this.field = new InnerMathField(ctrlr);
        this.field.name = this.name;
        ctrlr.editable = true;
        ctrlr.createTextarea();
        ctrlr.editablesTextareaEvents();
        ctrlr.cursor.insAtRightEnd(ctrlr.root);
        RootBlockMixin(ctrlr.root);
        // MathQuill applies aria-hidden to .mq-root-block containers because these contain math notation that screen
        // readers can't interpret directly. MathQuill uses an aria-live region as a sibling of these block containers
        // to provide an alternative representation for screen readers.
        //
        // MathQuillMathFields have their own focusable text aria and aria live region, so it is incorrect for any
        // parent of the editable field to have an aria-hidden property.
        //
        // https://developer.mozilla.org/en-US/docs/Web/Accessibility/ARIA/Attributes/aria-hidden
        //
        // Handle this by recursively walking the parents of this element until we hit a root block, and if we hit any
        // parent with aria-hidden="true", removing the property from the parent and pushing it down to each of the
        // parents children. This should result in no parent of this node having aria-hidden="true", but should keep as
        // much of what was previously hidden hidden as possible while obeying this constraint.
        const pushDownAriaHidden = (node) => {
            if (!(node instanceof Element))
                return;
            if (node.getAttribute('aria-hidden') === 'true') {
                node.removeAttribute('aria-hidden');
                node.childNodes.forEach((child) => {
                    if (child instanceof Element)
                        child.setAttribute('aria-hidden', 'true');
                });
            }
            if (node.parentNode && !node.classList.contains('mq-root-block'))
                pushDownAriaHidden(node.parentNode);
        };
        pushDownAriaHidden(this.elements.first.parentNode);
        this.elements.removeClass('aria-hidden');
        this.field.blur();
    }
    registerInnerField(innerFields) {
        if (!this.field)
            throw new Error('Unable to register editable without field');
        innerFields.push(this.field);
    }
    latex() {
        return this.ends.left?.latex() ?? '';
    }
    text() {
        return this.ends.left?.text() ?? '';
    }
};
// Embed arbitrary things
// Probably the closest DOM analogue would be an iframe?
// From MathQuill's perspective, it's a Symbol, it can be
// anywhere and the cursor can go around it but never in it.
// Create by calling public API method .dropEmbedded(),
// or by calling the global public API method .registerEmbed()
// and rendering LaTeX like \embed{registeredName} (see test).
LatexCmds.embed = class extends Symbol {
    setOptions(options) {
        const noop = () => '';
        this.text = options.text || noop;
        this.htmlTemplate = options.htmlString || '';
        this.latex = options.latex || noop;
        return this;
    }
    parser() {
        return Parser.string('{')
            .then(Parser.regex(/^[a-z][a-z0-9]*/i))
            .skip(Parser.string('}'))
            .then((name) => 
        // the chars allowed in the optional data block are arbitrary other than
        // excluding curly braces and square brackets (which'd be too confusing)
        Parser.string('[')
            .then(Parser.regex(/^[-\w\s]*/))
            .skip(Parser.string(']'))
            .or(Parser.succeed())
            .map((data) => this.setOptions(EMBEDS[name](data))));
    }
};

// Input box to type backslash commands
CharCmds['\\'] = class LatexCommandInput extends MathCommand {
    _replacedFragment;
    constructor() {
        super();
        this.ctrlSeq = '\\';
        this.htmlTemplate = '<span class="mq-latex-command-input mq-non-leaf">\\<span>&0</span></span>';
        this.textTemplate = ['\\'];
    }
    replaces(replacedFragment) {
        this._replacedFragment = replacedFragment?.disown();
        this.isEmpty = () => false;
    }
    createBlocks() {
        super.createBlocks();
        const leftEnd = this.ends.left;
        if (!leftEnd)
            return;
        leftEnd.focus = () => {
            leftEnd.parent?.elements.addClass('mq-has-cursor');
            if (leftEnd.isEmpty())
                leftEnd.parent?.elements.removeClass('mq-empty');
            return leftEnd;
        };
        leftEnd.blur = () => {
            leftEnd.parent?.elements.removeClass('mq-has-cursor');
            if (leftEnd.isEmpty())
                leftEnd.parent?.elements.addClass('mq-empty');
            return leftEnd;
        };
        leftEnd.write = (cursor, ch) => {
            cursor.show().deleteSelection();
            if (/[a-z]/i.exec(ch)) {
                new VanillaSymbol(ch).createLeftOf(cursor);
                cursor.controller.aria.alert(ch);
            }
            else {
                const cmd = leftEnd.parent.renderCommand(cursor);
                cursor.controller.aria.queue(cmd.mathspeak({ createdLeftOf: cursor }));
                if (ch !== '\\' || !leftEnd.isEmpty())
                    cursor.parent?.write(cursor, ch);
                else
                    cursor.controller.aria.alert();
            }
        };
        leftEnd.keystroke = (key, e, ctrlr) => {
            if (key === 'Tab' || key === 'Shift-Tab' || key === 'Enter' || key === 'Spacebar') {
                const cmd = leftEnd.parent.renderCommand(ctrlr.cursor);
                ctrlr.aria.alert(cmd.mathspeak({ createdLeftOf: ctrlr.cursor }));
                e.preventDefault();
                return;
            }
            super.keystroke(key, e, ctrlr);
            return;
        };
    }
    createLeftOf(cursor) {
        super.createLeftOf(cursor);
        if (this._replacedFragment) {
            const el = this.elements.first;
            this._replacedFragment.elements.addClass('mq-blur');
            this.elements.first.before(...this._replacedFragment.elements.contents);
            this.elements.add(this._replacedFragment.elements);
            // FIXME: Is monkey-patching the mousedown and mousemove handlers the right way to do this?
            const handler = (e) => {
                e.stopPropagation();
                e.preventDefault();
                el.dispatchEvent(new MouseEvent(e.type, e));
            };
            this.elements.firstElement.addEventListener('mousedown', handler);
            this.elements.firstElement.addEventListener('mousemove', handler);
        }
    }
    latex() {
        return '\\' + (this.ends.left?.latex() ?? '') + ' ';
    }
    renderCommand(cursor) {
        this.elements = new VNode(this.elements.last);
        this.remove();
        if (this.right)
            cursor.insLeftOf(this.right);
        else if (this.parent)
            cursor.insAtRightEnd(this.parent);
        const latex = this.ends.left?.latex() || ' ';
        if (latex in LatexCmds) {
            const cmd = new LatexCmds[latex](latex);
            if (this._replacedFragment)
                cmd.replaces(this._replacedFragment);
            cmd.createLeftOf(cursor);
            return cmd;
        }
        else {
            const cmd = new TextBlock();
            cmd.replaces(latex);
            cmd.createLeftOf(cursor);
            cursor.insRightOf(cmd);
            if (this._replacedFragment)
                this._replacedFragment.remove();
            return cmd;
        }
    }
};

// Symbols for Basic Mathematics
class OperatorName extends Symbol {
    constructor(fn) {
        super();
        this.ctrlSeq = fn;
    }
    createLeftOf(cursor) {
        for (const char of this.ctrlSeq) {
            new Letter(char).createLeftOf(cursor);
        }
    }
    parser() {
        const block = new MathBlock();
        for (const char of this.ctrlSeq) {
            new Letter(char).adopt(block, block.ends.right);
        }
        return Parser.succeed(block.children());
    }
}
for (const fn in BuiltInOpNames) {
    LatexCmds[fn] = OperatorName;
}
LatexCmds.operatorname = class extends MathCommand {
    constructor(ctrlSeq, htmlTemplate, textTemplate) {
        super(ctrlSeq, htmlTemplate, textTemplate);
        this.createLeftOf = noop;
    }
    numBlocks() {
        return 1;
    }
    parser() {
        return latexMathParser.block.map((b) => b.children());
    }
};
LatexCmds.f = class extends Letter {
    constructor() {
        super('f', '<var class="mq-f">f</var>');
        this.letter = 'f';
    }
    italicize(bool) {
        this.elements.html('f');
        this.elements.toggleClass('mq-f', bool);
        return super.italicize(bool);
    }
};
// VanillaSymbol's
LatexCmds[' '] = LatexCmds.space = bindMixin(VanillaSymbol, '\\ ', '&nbsp;', '', ' ');
LatexCmds["'"] = LatexCmds.prime = bindMixin(VanillaSymbol, "'", '&prime;', '', 'prime');
// LatexCmds['\u2033'] = LatexCmds.dprime = bindMixin(VanillaSymbol, '\u2033', '&Prime;');
LatexCmds.backslash = bindMixin(VanillaSymbol, '\\backslash ', '\\', '', 'backslash');
LatexCmds.$ = bindMixin(VanillaSymbol, '\\$', '$', '', 'dollar');
// does not use Symbola font
class NonSymbolaSymbol extends Symbol {
    constructor(ch, html) {
        super(ch, `<span class="mq-non-symbola">${html || ch}</span>`);
    }
}
LatexCmds['@'] = NonSymbolaSymbol;
LatexCmds['&'] = bindMixin(NonSymbolaSymbol, '\\&', '&amp;');
LatexCmds['%'] = bindMixin(NonSymbolaSymbol, '\\%', '%');
// Lowercase Greek letter variables
// See http://www.ams.org/STIX/ion/stixsig03.html
LatexCmds.alpha =
    LatexCmds.beta =
        LatexCmds.gamma =
            LatexCmds.delta =
                LatexCmds.zeta =
                    LatexCmds.eta =
                        LatexCmds.theta =
                            LatexCmds.iota =
                                LatexCmds.kappa =
                                    LatexCmds.mu =
                                        LatexCmds.nu =
                                            LatexCmds.xi =
                                                LatexCmds.rho =
                                                    LatexCmds.sigma =
                                                        LatexCmds.tau =
                                                            LatexCmds.chi =
                                                                LatexCmds.psi =
                                                                    LatexCmds.omega =
                                                                        class extends Variable {
                                                                            constructor(latex) {
                                                                                super(`\\${latex} `, `&${latex};`);
                                                                            }
                                                                        };
LatexCmds.phi = // W3C or Unicode?
    bindMixin(Variable, '\\phi ', '&#981;');
LatexCmds.phiv = // Elsevier and 9573-13
    LatexCmds.varphi = // AMS and LaTeX
        bindMixin(Variable, '\\varphi ', '&phi;');
LatexCmds.epsilon = // W3C or Unicode?
    bindMixin(Variable, '\\epsilon ', '&#1013;');
LatexCmds.epsiv = // Elsevier and 9573-13
    LatexCmds.varepsilon = // AMS and LaTeX
        bindMixin(Variable, '\\varepsilon ', '&epsilon;');
LatexCmds.piv = // W3C/Unicode and Elsevier and 9573-13
    LatexCmds.varpi = // AMS and LaTeX
        bindMixin(Variable, '\\varpi ', '&piv;');
LatexCmds.sigmaf = // W3C/Unicode
    LatexCmds.sigmav = // Elsevier
        LatexCmds.varsigma = // LaTeX
            bindMixin(Variable, '\\varsigma ', '&sigmaf;');
LatexCmds.thetav = // Elsevier and 9573-13
    LatexCmds.vartheta = // AMS and LaTeX
        LatexCmds.thetasym = // W3C/Unicode
            bindMixin(Variable, '\\vartheta ', '&thetasym;');
LatexCmds.upsilon = // AMS and LaTeX and W3C/Unicode
    LatexCmds.upsi = // Elsevier and 9573-13
        bindMixin(Variable, '\\upsilon ', '&upsilon;');
// These aren't even mentioned in the HTML character entity references
LatexCmds.gammad = // Elsevier
    LatexCmds.Gammad = // 9573-13 -- WTF, right? I dunno if this was a typo in the reference (see above)
        LatexCmds.digamma = // LaTeX
            bindMixin(Variable, '\\digamma ', '&#989;');
LatexCmds.kappav = // Elsevier
    LatexCmds.varkappa = // AMS and LaTeX
        bindMixin(Variable, '\\varkappa ', '&#1008;');
LatexCmds.rhov = // Elsevier and 9573-13
    LatexCmds.varrho = // AMS and LaTeX
        bindMixin(Variable, '\\varrho ', '&#1009;');
// Greek constants, look best in non-italicized Times New Roman
LatexCmds.pi = LatexCmds['\u03c0'] = bindMixin(NonSymbolaSymbol, '\\pi ', '&pi;');
LatexCmds.lambda = bindMixin(NonSymbolaSymbol, '\\lambda ', '&lambda;');
// uppercase greek letters
LatexCmds.Upsilon = // LaTeX
    LatexCmds.Upsi = // Elsevier and 9573-13
        LatexCmds.upsih = // W3C/Unicode "upsilon with hook"
            LatexCmds.Upsih =
                // Symbola's 'upsilon with a hook' is a capital Y without hooks :(
                bindMixin(Symbol, '\\Upsilon ', '<var style="font-family: serif">&upsih;</var>');
// Other symbols with the same LaTeX command and HTML character entity reference
LatexCmds.Gamma =
    LatexCmds.Delta =
        LatexCmds.Theta =
            LatexCmds.Lambda =
                LatexCmds.Xi =
                    LatexCmds.Pi =
                        LatexCmds.Sigma =
                            LatexCmds.Phi =
                                LatexCmds.Psi =
                                    LatexCmds.Omega =
                                        LatexCmds.forall =
                                            class extends VanillaSymbol {
                                                constructor(latex) {
                                                    super(`\\${latex} `, `&${latex};`);
                                                }
                                            };
// symbols that aren't a single MathCommand, but are instead a whole
// Fragment. Creates the Fragment from a LaTeX string
class LatexFragment extends MathCommand {
    latexStr;
    constructor(latex) {
        super();
        this.latex = () => latex;
        this.latexStr = latex;
    }
    createLeftOf(cursor) {
        const block = latexMathParser.parse(this.latex());
        if (cursor.parent)
            block.children().adopt(cursor.parent, cursor.left, cursor.right);
        cursor.left = block.ends.right;
        cursor.element.before(...block.domify().contents);
        block.finalizeInsert(cursor.options, cursor);
        block.ends.right?.right?.siblingCreated?.(cursor.options, 'left');
        block.ends.left?.left?.siblingCreated?.(cursor.options, 'right');
        cursor.parent?.bubble('reflow');
    }
    mathspeak() {
        const block = latexMathParser.parse(this.latexStr);
        return block.mathspeak();
    }
    parser() {
        const block = latexMathParser.parse(this.latex());
        return Parser.succeed(block.children());
    }
}
// For what seems to me like [stupid reasons][1], Unicode provides
// subscripted and superscripted versions of all ten Arabic numerals,
// as well as [so-called "vulgar fractions"][2].
// Nobody really cares about most of them, but some of them actually
// predate Unicode, dating back to [ISO-8859-1][3], apparently also
// known as "Latin-1", which among other things [Windows-1252][4]
// largely coincides with, so Microsoft Word sometimes inserts them
// and they get copy-pasted into MathQuill.
//
// (Irrelevant but funny story: though not a superset of Latin-1 aka
// ISO-8859-1, Windows-1252 **is** a strict superset of the "closely
// related but distinct"[3] "ISO 8859-1" -- see the lack of a dash
// after "ISO"? Completely different character set, like elephants vs
// elephant seals, or "Zombies" vs "Zombie Redneck Torture Family".
// What kind of idiot would get them confused.
// People in fact got them confused so much, it was so common to
// mislabel Windows-1252 text as ISO-8859-1, that most modern web
// browsers and email clients treat the MIME charset of ISO-8859-1
// as actually Windows-1252, behavior now standard in the HTML5 spec.)
//
// [1]: http://en.wikipedia.org/wiki/Unicode_subscripts_andsuper_scripts
// [2]: http://en.wikipedia.org/wiki/Number_Forms
// [3]: http://en.wikipedia.org/wiki/ISO/IEC_8859-1
// [4]: http://en.wikipedia.org/wiki/Windows-1252
LatexCmds['\u00b9'] = bindMixin(LatexFragment, '^1');
LatexCmds['\u00b2'] = bindMixin(LatexFragment, '^2');
LatexCmds['\u00b3'] = bindMixin(LatexFragment, '^3');
LatexCmds['\u00bc'] = bindMixin(LatexFragment, '\\frac14');
LatexCmds['\u00bd'] = bindMixin(LatexFragment, '\\frac12');
LatexCmds['\u00be'] = bindMixin(LatexFragment, '\\frac34');
class PlusMinus extends BinaryOperator {
    constructor(ctrlSeq, html, text, mathspeak) {
        super(ctrlSeq, `<span>${html || ctrlSeq}</span>`, text, mathspeak, true);
        this.siblingCreated = this.siblingDeleted = (opts, dir) => this.contactWeld(opts, dir);
    }
    contactWeld(_opts, dir) {
        const isUnary = (node) => {
            if (node.left) {
                // If the left sibling is a binary operator or a separator (comma, semicolon, colon)
                // or an open bracket (open parenthesis, open square bracket)
                // consider the operator to be unary.
                if (node.left instanceof BinaryOperator || /^[,;:([]$/.test(node.left.ctrlSeq)) {
                    return true;
                }
            }
            else if (node.parent?.parent?.isStyleBlock()) {
                // If we are in a style block at the leftmost edge, determine unary/binary based on the style block.
                // This allows style blocks to be transparent for unary/binary purposes.
                return isUnary(node.parent.parent);
            }
            else {
                return true;
            }
            return false;
        };
        if (dir === 'right')
            return; // ignore if sibling only changed on the right
        this.isUnary = isUnary(this);
        this.elements.firstElement.className = this.isUnary ? '' : 'mq-binary-operator';
        return this;
    }
}
LatexCmds['+'] = class extends PlusMinus {
    constructor() {
        super('+', '+');
    }
    mathspeak() {
        return this.isUnary ? 'positive' : 'plus';
    }
};
LatexCmds['\u2013'] = LatexCmds['-'] = class extends PlusMinus {
    constructor() {
        super('-', '&minus;');
    }
    mathspeak() {
        return this.isUnary ? 'negative' : 'minus';
    }
};
LatexCmds['\u00b1'] =
    LatexCmds.pm =
        LatexCmds.plusmn =
            LatexCmds.plusminus =
                bindMixin(PlusMinus, '\\pm ', '&plusmn;', '', 'plus-or-minus');
LatexCmds.mp = LatexCmds.mnplus = LatexCmds.minusplus = bindMixin(PlusMinus, '\\mp ', '&#8723;', '', 'minus-or-plus');
// Semantically should be &sdot;, but &middot; looks better
CharCmds['*'] = LatexCmds.sdot = LatexCmds.cdot = bindMixin(BinaryOperator, '\\cdot ', '&middot;', '*', 'times');
const less = {
    ctrlSeq: '\\le ',
    html: '&le;',
    text: '<=',
    mathspeak: 'less than or equal to',
    ctrlSeqStrict: '<',
    htmlStrict: '&lt;',
    textStrict: '<',
    mathspeakStrict: 'less than'
};
const greater = {
    ctrlSeq: '\\ge ',
    html: '&ge;',
    text: '>=',
    mathspeak: 'greater than or equal to',
    ctrlSeqStrict: '>',
    htmlStrict: '&gt;',
    textStrict: '>',
    mathspeakStrict: 'greater than'
};
const neq = {
    ctrlSeq: '\\ne ',
    html: '&ne;',
    text: '!=',
    mathspeak: 'not equal to',
    ctrlSeqStrict: '!',
    htmlStrict: '!',
    textStrict: '!',
    mathspeakStrict: 'factorial'
};
LatexCmds['<'] = LatexCmds.lt = bindMixin(Inequality, less, true);
LatexCmds['>'] = LatexCmds.gt = bindMixin(Inequality, greater, true);
LatexCmds['!'] = LatexCmds.factorial = bindMixin(FactorialOrNEQ, neq, true);
LatexCmds['\u2264'] = LatexCmds.le = LatexCmds.leq = bindMixin(Inequality, less, false);
LatexCmds['\u2265'] = LatexCmds.ge = LatexCmds.geq = bindMixin(Inequality, greater, false);
LatexCmds['\u2260'] = LatexCmds.ne = LatexCmds.neq = bindMixin(FactorialOrNEQ, neq, false);
LatexCmds['='] = Equality;
LatexCmds['\u00d7'] = LatexCmds.times = bindMixin(BinaryOperator, '\\times ', '&times;', '*', 'times');
LatexCmds['\u00f7'] =
    LatexCmds.div =
        LatexCmds.divide =
            LatexCmds.divides =
                bindMixin(BinaryOperator, '\\div ', '&divide;', '/', 'divided by');
CharCmds['~'] = LatexCmds.sim = bindMixin(BinaryOperator, '\\sim ', '~', '~', 'tilde');

// Symbols for Advanced Mathematics
LatexCmds.notin =
    LatexCmds.cong =
        LatexCmds.equiv =
            LatexCmds.oplus =
                LatexCmds.otimes =
                    class extends BinaryOperator {
                        constructor(latex) {
                            super(`\\${latex} `, `&${latex};`, 'not in');
                        }
                    };
LatexCmds['\u2217'] =
    LatexCmds.ast =
        LatexCmds.star =
            LatexCmds.loast =
                LatexCmds.lowast =
                    bindMixin(BinaryOperator, '\\ast ', '&lowast;', 'asterisk');
LatexCmds.therefor = LatexCmds.therefore = bindMixin(BinaryOperator, '\\therefore ', '&there4;');
LatexCmds.cuz = // l33t
    LatexCmds.because = bindMixin(BinaryOperator, '\\because ', '&#8757;');
LatexCmds.prop = LatexCmds.propto = bindMixin(BinaryOperator, '\\propto ', '&prop;', 'proportional to');
LatexCmds['\u2248'] =
    LatexCmds.asymp =
        LatexCmds.approx =
            bindMixin(BinaryOperator, '\\approx ', '&asymp;', 'approximately equal to');
LatexCmds.isin = LatexCmds.in = bindMixin(BinaryOperator, '\\in ', '&isin;', 'is in');
LatexCmds.ni = LatexCmds.contains = bindMixin(BinaryOperator, '\\ni ', '&ni;', 'contains');
LatexCmds.notni =
    LatexCmds.niton =
        LatexCmds.notcontains =
            LatexCmds.doesnotcontain =
                bindMixin(BinaryOperator, '\\not\\ni ', '&#8716;', 'does not contain');
LatexCmds.sub = LatexCmds.subset = bindMixin(BinaryOperator, '\\subset ', '&sub;', 'subset');
LatexCmds.sup = LatexCmds.supset = LatexCmds.superset = bindMixin(BinaryOperator, '\\supset ', '&sup;', 'superset');
LatexCmds.nsub =
    LatexCmds.notsub =
        LatexCmds.nsubset =
            LatexCmds.notsubset =
                bindMixin(BinaryOperator, '\\not\\subset ', '&#8836;', 'not a subset');
LatexCmds.nsup =
    LatexCmds.notsup =
        LatexCmds.nsupset =
            LatexCmds.notsupset =
                LatexCmds.nsuperset =
                    LatexCmds.notsuperset =
                        bindMixin(BinaryOperator, '\\not\\supset ', '&#8837;', 'not a superset');
LatexCmds.sube =
    LatexCmds.subeq =
        LatexCmds.subsete =
            LatexCmds.subseteq =
                bindMixin(BinaryOperator, '\\subseteq ', '&sube;', 'subset or equal to');
LatexCmds.supe =
    LatexCmds.supeq =
        LatexCmds.supsete =
            LatexCmds.supseteq =
                LatexCmds.supersete =
                    LatexCmds.superseteq =
                        bindMixin(BinaryOperator, '\\supseteq ', '&supe;', 'superset or equal to');
LatexCmds.nsube =
    LatexCmds.nsubeq =
        LatexCmds.notsube =
            LatexCmds.notsubeq =
                LatexCmds.nsubsete =
                    LatexCmds.nsubseteq =
                        LatexCmds.notsubsete =
                            LatexCmds.notsubseteq =
                                bindMixin(BinaryOperator, '\\not\\subseteq ', '&#8840;', 'not subset or equal to');
LatexCmds.nsupe =
    LatexCmds.nsupeq =
        LatexCmds.notsupe =
            LatexCmds.notsupeq =
                LatexCmds.nsupsete =
                    LatexCmds.nsupseteq =
                        LatexCmds.notsupsete =
                            LatexCmds.notsupseteq =
                                LatexCmds.nsupersete =
                                    LatexCmds.nsuperseteq =
                                        LatexCmds.notsupersete =
                                            LatexCmds.notsuperseteq =
                                                bindMixin(BinaryOperator, '\\not\\supseteq ', '&#8841;', 'not superset or equal to');
// The canonical sets of numbers
LatexCmds.mathbb = class extends MathCommand {
    constructor(ctrlSeq, htmlTemplate, textTemplate) {
        super(ctrlSeq, htmlTemplate, textTemplate);
        this.createLeftOf = noop;
    }
    numBlocks() {
        return 1;
    }
    parser() {
        return (Parser.optWhitespace
            .then(Parser.string('{'))
            .then(Parser.optWhitespace)
            .then(Parser.regex(/^[NPZQRCH]/))
            .skip(Parser.optWhitespace)
            .skip(Parser.string('}'))
            // Instantiate the class for the matching char
            .map((c) => new LatexCmds[c]()));
    }
};
LatexCmds.N = LatexCmds.naturals = LatexCmds.Naturals = bindMixin(VanillaSymbol, '\\mathbb{N}', '&#8469;', 'naturals');
LatexCmds.P =
    LatexCmds.primes =
        LatexCmds.Primes =
            LatexCmds.projective =
                LatexCmds.Projective =
                    LatexCmds.probability =
                        LatexCmds.Probability =
                            bindMixin(VanillaSymbol, '\\mathbb{P}', '&#8473;', 'P');
LatexCmds.Z = LatexCmds.integers = LatexCmds.Integers = bindMixin(VanillaSymbol, '\\mathbb{Z}', '&#8484;', 'integers');
LatexCmds.Q =
    LatexCmds.rationals =
        LatexCmds.Rationals =
            bindMixin(VanillaSymbol, '\\mathbb{Q}', '&#8474;', 'rationals');
LatexCmds.R = LatexCmds.reals = LatexCmds.Reals = bindMixin(VanillaSymbol, '\\mathbb{R}', '&#8477;', 'reals');
LatexCmds.C =
    LatexCmds.complex =
        LatexCmds.Complex =
            LatexCmds.complexes =
                LatexCmds.Complexes =
                    LatexCmds.complexplane =
                        LatexCmds.Complexplane =
                            LatexCmds.ComplexPlane =
                                bindMixin(VanillaSymbol, '\\mathbb{C}', '&#8450;', 'comlexes');
LatexCmds.H =
    LatexCmds.Hamiltonian =
        LatexCmds.quaternions =
            LatexCmds.Quaternions =
                bindMixin(VanillaSymbol, '\\mathbb{H}', '&#8461;', 'quaternions');
// spacing
LatexCmds.quad = LatexCmds.emsp = bindMixin(VanillaSymbol, '\\quad ', '    ', '4 spaces');
LatexCmds.qquad = bindMixin(VanillaSymbol, '\\qquad ', '        ', '8 spaces');
// spacing special characters, gonna have to implement this in LatexCommandInput::onText somehow
// case ',':
//  return VanillaSymbol('\\, ',' ');
// case ':':
//  return VanillaSymbol('\\: ','  ');
// case ';':
//  return VanillaSymbol('\\; ','   ');
// case '!':
//  return Symbol('\\! ', '<span style="margin-right:-.2em"></span>');
// binary operators
LatexCmds.diamond = bindMixin(VanillaSymbol, '\\diamond ', '&#9671;');
LatexCmds.bigtriangleup = bindMixin(VanillaSymbol, '\\bigtriangleup ', '&#9651;', 'big triangle up');
LatexCmds.ominus = bindMixin(VanillaSymbol, '\\ominus ', '&#8854;', 'o minus');
LatexCmds.uplus = bindMixin(VanillaSymbol, '\\uplus ', '&#8846;', 'disjoint union');
LatexCmds.bigtriangledown = bindMixin(VanillaSymbol, '\\bigtriangledown ', '&#9661;', 'big triangle down');
LatexCmds.sqcap = bindMixin(VanillaSymbol, '\\sqcap ', '&#8851;', 'square cap');
LatexCmds.triangleleft = bindMixin(VanillaSymbol, '\\triangleleft ', '&#8882;', 'triangle left');
LatexCmds.sqcup = bindMixin(VanillaSymbol, '\\sqcup ', '&#8852;', 'square cup');
LatexCmds.triangleright = bindMixin(VanillaSymbol, '\\triangleright ', '&#8883;', 'triangle right');
// circledot is not a not real LaTex command see https://github.com/mathquill/mathquill/pull/552 for more details
LatexCmds.odot = LatexCmds.circledot = bindMixin(VanillaSymbol, '\\odot ', '&#8857;', 'circle dot');
LatexCmds.bigcirc = bindMixin(VanillaSymbol, '\\bigcirc ', '&#9711;', 'big circle');
LatexCmds.dagger = bindMixin(VanillaSymbol, '\\dagger ', '&#0134;', 'dagger');
LatexCmds.ddagger = bindMixin(VanillaSymbol, '\\ddagger ', '&#135;', 'big dagger');
LatexCmds.wr = bindMixin(VanillaSymbol, '\\wr ', '&#8768;', 'wreath');
LatexCmds.amalg = bindMixin(VanillaSymbol, '\\amalg ', '&#8720;', 'amalgam');
// relationship symbols
LatexCmds.models = bindMixin(VanillaSymbol, '\\models ', '&#8872;');
LatexCmds.prec = bindMixin(VanillaSymbol, '\\prec ', '&#8826;', 'precedes');
LatexCmds.succ = bindMixin(VanillaSymbol, '\\succ ', '&#8827;', 'succeeds');
LatexCmds.preceq = bindMixin(VanillaSymbol, '\\preceq ', '&#8828;', 'precedes or equals');
LatexCmds.succeq = bindMixin(VanillaSymbol, '\\succeq ', '&#8829;', 'succeeds or equals');
LatexCmds.simeq = bindMixin(VanillaSymbol, '\\simeq ', '&#8771;', 'similar or equal to');
LatexCmds.mid = bindMixin(VanillaSymbol, '\\mid ', '&#8739;', 'divides');
LatexCmds.ll = bindMixin(VanillaSymbol, '\\ll ', '&#8810;', 'sufficiently less than');
LatexCmds.gg = bindMixin(VanillaSymbol, '\\gg ', '&#8811;', 'sufficiently greater than');
LatexCmds.parallel = bindMixin(VanillaSymbol, '\\parallel ', '&#8741;', 'parallel to');
LatexCmds.nparallel = bindMixin(VanillaSymbol, '\\nparallel ', '&#8742;', 'not parallel to');
LatexCmds.bowtie = bindMixin(VanillaSymbol, '\\bowtie ', '&#8904;');
LatexCmds.sqsubset = bindMixin(VanillaSymbol, '\\sqsubset ', '&#8847;', 'square subset');
LatexCmds.sqsupset = bindMixin(VanillaSymbol, '\\sqsupset ', '&#8848;', 'square superset');
LatexCmds.smile = bindMixin(VanillaSymbol, '\\smile ', '&#8995;');
LatexCmds.sqsubseteq = bindMixin(VanillaSymbol, '\\sqsubseteq ', '&#8849;', 'square subset or equal to');
LatexCmds.sqsupseteq = bindMixin(VanillaSymbol, '\\sqsupseteq ', '&#8850;', 'square superset or equal to');
LatexCmds.doteq = bindMixin(VanillaSymbol, '\\doteq ', '&#8784;', 'dotted equals');
LatexCmds.frown = bindMixin(VanillaSymbol, '\\frown ', '&#8994;');
LatexCmds.vdash = bindMixin(VanillaSymbol, '\\vdash ', '&#8870;', 'v dash');
LatexCmds.dashv = bindMixin(VanillaSymbol, '\\dashv ', '&#8867;', 'dash v');
LatexCmds.nless = bindMixin(VanillaSymbol, '\\nless ', '&#8814;', 'not less than');
LatexCmds.ngtr = bindMixin(VanillaSymbol, '\\ngtr ', '&#8815;', 'not greater than');
// arrows
LatexCmds.longleftarrow = bindMixin(VanillaSymbol, '\\longleftarrow ', '&#8592;', 'left arrow');
LatexCmds.longrightarrow = bindMixin(VanillaSymbol, '\\longrightarrow ', '&#8594;', 'right arrow');
LatexCmds.Longleftarrow = bindMixin(VanillaSymbol, '\\Longleftarrow ', '&#8656;', 'left arrow');
LatexCmds.Longrightarrow = bindMixin(VanillaSymbol, '\\Longrightarrow ', '&#8658;', 'right arrow');
LatexCmds.longleftrightarrow = bindMixin(VanillaSymbol, '\\longleftrightarrow ', '&#8596;', 'left and right arrow');
LatexCmds.updownarrow = bindMixin(VanillaSymbol, '\\updownarrow ', '&#8597;', 'up and down arrow');
LatexCmds.Longleftrightarrow = bindMixin(VanillaSymbol, '\\Longleftrightarrow ', '&#8660;', 'left and right arrow');
LatexCmds.Updownarrow = bindMixin(VanillaSymbol, '\\Updownarrow ', '&#8661;', 'up and down arrow');
LatexCmds.mapsto = bindMixin(VanillaSymbol, '\\mapsto ', '&#8614;', 'maps to');
LatexCmds.nearrow = bindMixin(VanillaSymbol, '\\nearrow ', '&#8599;', 'northeast arrow');
LatexCmds.hookleftarrow = bindMixin(VanillaSymbol, '\\hookleftarrow ', '&#8617;', 'hook left arrow');
LatexCmds.hookrightarrow = bindMixin(VanillaSymbol, '\\hookrightarrow ', '&#8618;', 'hook right arrow');
LatexCmds.searrow = bindMixin(VanillaSymbol, '\\searrow ', '&#8600;', 'southeast arrow');
LatexCmds.leftharpoonup = bindMixin(VanillaSymbol, '\\leftharpoonup ', '&#8636;', 'left harpoon up');
LatexCmds.rightharpoonup = bindMixin(VanillaSymbol, '\\rightharpoonup ', '&#8640;', 'right harpoon up');
LatexCmds.swarrow = bindMixin(VanillaSymbol, '\\swarrow ', '&#8601;', 'southwest arrow');
LatexCmds.leftharpoondown = bindMixin(VanillaSymbol, '\\leftharpoondown ', '&#8637;', 'left harpoon down');
LatexCmds.rightharpoondown = bindMixin(VanillaSymbol, '\\rightharpoondown ', '&#8641;', 'right harpoon down');
LatexCmds.nwarrow = bindMixin(VanillaSymbol, '\\nwarrow ', '&#8598;', 'northwest arrow');
// Misc
LatexCmds.ldots = bindMixin(VanillaSymbol, '\\ldots ', '&#8230;', 'ellipsis');
LatexCmds.cdots = bindMixin(VanillaSymbol, '\\cdots ', '&#8943;', 'multiplication ellipsis');
LatexCmds.vdots = bindMixin(VanillaSymbol, '\\vdots ', '&#8942;', 'vertical ellipsis');
LatexCmds.ddots = bindMixin(VanillaSymbol, '\\ddots ', '&#8945;', 'diagonal ellipsis');
LatexCmds.surd = bindMixin(VanillaSymbol, '\\surd ', '&#8730;', 'unresolved root');
LatexCmds.triangle = bindMixin(VanillaSymbol, '\\triangle ', '&#9651;');
LatexCmds.ell = bindMixin(VanillaSymbol, '\\ell ', '&#8467;');
LatexCmds.top = bindMixin(VanillaSymbol, '\\top ', '&#8868;');
LatexCmds.flat = bindMixin(VanillaSymbol, '\\flat ', '&#9837;');
LatexCmds.natural = bindMixin(VanillaSymbol, '\\natural ', '&#9838;');
LatexCmds.sharp = bindMixin(VanillaSymbol, '\\sharp ', '&#9839;');
LatexCmds.wp = bindMixin(VanillaSymbol, '\\wp ', '&#8472;');
LatexCmds.bot = bindMixin(VanillaSymbol, '\\bot ', '&#8869;');
LatexCmds.clubsuit = bindMixin(VanillaSymbol, '\\clubsuit ', '&#9827;', 'club suit');
LatexCmds.diamondsuit = bindMixin(VanillaSymbol, '\\diamondsuit ', '&#9826;', 'diamond suit');
LatexCmds.heartsuit = bindMixin(VanillaSymbol, '\\heartsuit ', '&#9825;', 'heart suit');
LatexCmds.spadesuit = bindMixin(VanillaSymbol, '\\spadesuit ', '&#9824;', 'spade suit');
// not real LaTex command see https://github.com/mathquill/mathquill/pull/552 for more details
LatexCmds.parallelogram = bindMixin(VanillaSymbol, '\\parallelogram ', '&#9649;');
LatexCmds.square = bindMixin(VanillaSymbol, '\\square ', '&#11036;');
// variable-sized
LatexCmds.oint = bindMixin(VanillaSymbol, '\\oint ', '&#8750;', 'o int');
LatexCmds.bigcap = bindMixin(VanillaSymbol, '\\bigcap ', '&#8745;', 'big cap');
LatexCmds.bigcup = bindMixin(VanillaSymbol, '\\bigcup ', '&#8746;', 'big cup');
LatexCmds.bigsqcup = bindMixin(VanillaSymbol, '\\bigsqcup ', '&#8852;', 'big square cup');
LatexCmds.bigvee = bindMixin(VanillaSymbol, '\\bigvee ', '&#8744;', 'big vee');
LatexCmds.bigwedge = bindMixin(VanillaSymbol, '\\bigwedge ', '&#8743;', 'big wedge');
LatexCmds.bigodot = bindMixin(VanillaSymbol, '\\bigodot ', '&#8857;', 'big o dot');
LatexCmds.bigotimes = bindMixin(VanillaSymbol, '\\bigotimes ', '&#8855;', 'big o times');
LatexCmds.bigoplus = bindMixin(VanillaSymbol, '\\bigoplus ', '&#8853;', 'big o plus');
LatexCmds.biguplus = bindMixin(VanillaSymbol, '\\biguplus ', '&#8846;', 'big u plus');
// delimiters
LatexCmds.lfloor = bindMixin(VanillaSymbol, '\\lfloor ', '&#8970;', 'left floor');
LatexCmds.rfloor = bindMixin(VanillaSymbol, '\\rfloor ', '&#8971;', 'right floor');
LatexCmds.lceil = bindMixin(VanillaSymbol, '\\lceil ', '&#8968;', 'left ceiling');
LatexCmds.rceil = bindMixin(VanillaSymbol, '\\rceil ', '&#8969;', 'right ceiling');
LatexCmds.opencurlybrace = LatexCmds.lbrace = bindMixin(VanillaSymbol, '\\lbrace ', '{', 'left brace');
LatexCmds.closecurlybrace = LatexCmds.rbrace = bindMixin(VanillaSymbol, '\\rbrace ', '}', 'right brace');
LatexCmds.lbrack = bindMixin(VanillaSymbol, '[', 'left bracket');
LatexCmds.rbrack = bindMixin(VanillaSymbol, ']', 'right bracket');
// various symbols
LatexCmds.slash = bindMixin(VanillaSymbol, '/', '', '', 'slash');
LatexCmds.vert = bindMixin(VanillaSymbol, '|', '', '', 'vertical bar');
LatexCmds.perp = LatexCmds.perpendicular = bindMixin(VanillaSymbol, '\\perp ', '&perp;', 'perpendicular');
LatexCmds.nabla = LatexCmds.del = bindMixin(VanillaSymbol, '\\nabla ', '&nabla;');
LatexCmds.hbar = bindMixin(VanillaSymbol, '\\hbar ', '&#8463;', 'horizontal bar');
// FIXME: \AA is not valid LaTeX in math mode.  Neither is \text\AA (which is what this was before).  Furthermore,
// \text\AA does not parse correctly.  Valid LaTeX in math mode without any packages would be \textup{~\AA}, but that
// also does not parse correctly.
LatexCmds.AA =
    LatexCmds.Angstrom =
        LatexCmds.angstrom =
            bindMixin(VanillaSymbol, '\\AA ', '&#8491;', '\u00C5', 'angstrom');
LatexCmds.ring = LatexCmds.circ = LatexCmds.circle = bindMixin(VanillaSymbol, '\\circ ', '&#8728;', 'circle');
LatexCmds.bull = LatexCmds.bullet = bindMixin(VanillaSymbol, '\\bullet ', '&bull;');
LatexCmds.setminus = LatexCmds.smallsetminus = bindMixin(VanillaSymbol, '\\setminus ', '&#8726;', 'set minus');
LatexCmds.not = LatexCmds['\u00ac'] = LatexCmds.neg = bindMixin(VanillaSymbol, '\\neg ', '&not;', '', 'not');
LatexCmds['\u2026'] =
    LatexCmds.dots =
        LatexCmds.ellip =
            LatexCmds.hellip =
                LatexCmds.ellipsis =
                    LatexCmds.hellipsis =
                        bindMixin(VanillaSymbol, '\\dots ', '&hellip;', 'ellipsis');
LatexCmds.converges =
    LatexCmds.darr =
        LatexCmds.dnarr =
            LatexCmds.dnarrow =
                LatexCmds.downarrow =
                    bindMixin(VanillaSymbol, '\\downarrow ', '&darr;', 'converges with');
LatexCmds.dArr =
    LatexCmds.dnArr =
        LatexCmds.dnArrow =
            LatexCmds.Downarrow =
                bindMixin(VanillaSymbol, '\\Downarrow ', '&dArr;', 'down arrow');
LatexCmds.diverges =
    LatexCmds.uarr =
        LatexCmds.uparrow =
            bindMixin(VanillaSymbol, '\\uparrow ', '&uarr;', 'diverges from');
LatexCmds.uArr = LatexCmds.Uparrow = bindMixin(VanillaSymbol, '\\Uparrow ', '&uArr;', 'up arrow');
LatexCmds.to = bindMixin(BinaryOperator, '\\to ', '&rarr;');
LatexCmds.rarr = LatexCmds.rightarrow = bindMixin(VanillaSymbol, '\\rightarrow ', '&rarr;', 'right arrow');
LatexCmds.implies = bindMixin(BinaryOperator, '\\Rightarrow ', '&rArr;', 'implies');
LatexCmds.rArr = LatexCmds.Rightarrow = bindMixin(VanillaSymbol, '\\Rightarrow ', '&rArr;', 'right arrow');
LatexCmds.gets = bindMixin(BinaryOperator, '\\gets ', '&larr;');
LatexCmds.larr = LatexCmds.leftarrow = bindMixin(VanillaSymbol, '\\leftarrow ', '&larr;', 'left arrow');
LatexCmds.impliedby = bindMixin(BinaryOperator, '\\Leftarrow ', '&lArr;', 'implied by');
LatexCmds.lArr = LatexCmds.Leftarrow = bindMixin(VanillaSymbol, '\\Leftarrow ', '&lArr;', 'left arrow');
LatexCmds.harr =
    LatexCmds.lrarr =
        LatexCmds.leftrightarrow =
            bindMixin(VanillaSymbol, '\\leftrightarrow ', '&harr;', 'left and right arrow');
LatexCmds.iff = bindMixin(BinaryOperator, '\\Leftrightarrow ', '&hArr;', 'if and only if');
LatexCmds.hArr =
    LatexCmds.lrArr =
        LatexCmds.Leftrightarrow =
            bindMixin(VanillaSymbol, '\\Leftrightarrow ', '&hArr;', 'left and right arrow');
LatexCmds.Re = LatexCmds.Real = LatexCmds.real = bindMixin(VanillaSymbol, '\\Re ', '&real;', 'real');
LatexCmds.Im =
    LatexCmds.imag =
        LatexCmds.image =
            LatexCmds.imagin =
                LatexCmds.imaginary =
                    LatexCmds.Imaginary =
                        bindMixin(VanillaSymbol, '\\Im ', '&image;', 'imaginary');
LatexCmds.part = LatexCmds.partial = bindMixin(VanillaSymbol, '\\partial ', '&part;');
LatexCmds.inf =
    LatexCmds.infty =
        LatexCmds.infin =
            LatexCmds.infinity =
                bindMixin(VanillaSymbol, '\\infty ', '&infin;', 'inf', 'infinity');
LatexCmds.pounds = bindMixin(VanillaSymbol, '\\pounds ', '&pound;');
LatexCmds.alef =
    LatexCmds.alefsym =
        LatexCmds.aleph =
            LatexCmds.alephsym =
                bindMixin(VanillaSymbol, '\\aleph ', '&alefsym;');
LatexCmds.xist =
    LatexCmds.xists =
        LatexCmds.exist =
            LatexCmds.exists =
                bindMixin(VanillaSymbol, '\\exists ', '&exist;', 'there exists');
LatexCmds.nexists = LatexCmds.nexist = bindMixin(VanillaSymbol, '\\nexists ', '&#8708;', 'there is no');
LatexCmds.and = LatexCmds.land = LatexCmds.wedge = bindMixin(BinaryOperator, '\\wedge ', '&and;', 'and');
LatexCmds.or = LatexCmds.lor = LatexCmds.vee = bindMixin(BinaryOperator, '\\vee ', '&or;', 'or');
LatexCmds.o =
    LatexCmds.O =
        LatexCmds.empty =
            LatexCmds.emptyset =
                LatexCmds.oslash =
                    LatexCmds.Oslash =
                        LatexCmds.nothing =
                            LatexCmds.varnothing =
                                bindMixin(BinaryOperator, '\\varnothing ', '&empty;', 'empty set');
LatexCmds.U = LatexCmds.cup = LatexCmds.union = bindMixin(BinaryOperator, '\\cup ', '&cup;', 'U', 'union');
LatexCmds.cap =
    LatexCmds.intersect =
        LatexCmds.intersection =
            bindMixin(BinaryOperator, '\\cap ', '&cap;', 'intersection');
// FIXME: the correct LaTeX would be ^\circ but we can't parse that
LatexCmds.deg = LatexCmds.degree = class degree extends VanillaSymbol {
    constructor() {
        super('\\degree ', '&deg;', '', 'degrees');
    }
    text() {
        const rightText = this.right?.text();
        return `\u00B0${rightText && /^[^FCK]$/.test(rightText) ? ' ' : ''}`;
    }
};
LatexCmds.ang = LatexCmds.angle = bindMixin(VanillaSymbol, '\\angle ', '&ang;');
LatexCmds.measuredangle = bindMixin(VanillaSymbol, '\\measuredangle ', '&#8737;', 'measured angle');

// The publicly exposed MathQuill API.
const isBrowser = Object.getPrototypeOf(Object.getPrototypeOf(globalThis)) !== Object.prototype;
// globally exported API object
const mathQuill = {
    origMathQuill: isBrowser ? window.MathQuill : undefined,
    VERSION: '0.0.1',
    getInterface() {
        // Function that takes an HTML element and, if it's the root HTML element of a
        // static math or math or text field, returns an API object for it (else, undefined).
        //   const mathfield = MQ.MathField(mathFieldSpan);
        //   assert(MQ(mathFieldSpan).id === mathfield.id);
        //   assert(MQ(mathFieldSpan).id === MQ(mathFieldSpan).id);
        const MQ = (el) => {
            if (!(el instanceof HTMLElement))
                return;
            const blockId = el.querySelector('.mq-root-block')?.getAttribute(mqBlockId) ?? false;
            const ctrlr = blockId ? TNode.byId.get(parseInt(blockId))?.controller : undefined;
            return ctrlr?.apiClass;
        };
        MQ.saneKeyboardEvents = saneKeyboardEvents;
        MQ.config = (opts) => {
            Options.config(Options.prototype, opts);
            return MQ;
        };
        MQ.registerEmbed = (name, options) => {
            if (!/^[a-z][a-z0-9]*$/i.test(name)) {
                throw new Error('Embed name must start with letter and be only letters and digits');
            }
            EMBEDS[name] = options;
        };
        // Export the API functions that MathQuill-ify an HTML element into API objects
        // of each class. If the element had already been MathQuill-ified but into a
        // different kind (or it's not an HTML element), return undefined.
        const createEntrypoint = (kind, APIClass) => {
            function mqEntrypoint(el, opts) {
                if (!(el instanceof HTMLElement))
                    return;
                const mq = MQ(el);
                if (!(el instanceof HTMLElement) || mq instanceof APIClass)
                    return mq;
                const ctrlr = new Controller(new APIClass.RootBlock(), el, new Options());
                ctrlr.KIND_OF_MQ = kind;
                return new APIClass(ctrlr).config(opts ?? {}).__mathquillify();
            }
            return mqEntrypoint;
        };
        MQ.StaticMath = createEntrypoint('StaticMath', StaticMath);
        MQ.StaticMath.prototype = StaticMath.prototype;
        MQ.MathField = createEntrypoint('MathField', MathField);
        MQ.MathField.prototype = MathField.prototype;
        MQ.InnerMathField = createEntrypoint('InnerMathField', InnerMathField);
        MQ.InnerMathField.prototype = InnerMathField.prototype;
        MQ.TextField = createEntrypoint('TextField', TextField);
        MQ.TextField.prototype = TextField.prototype;
        return MQ;
    },
    noConflict() {
        if (!isBrowser)
            return mathQuill;
        window.MathQuill = this.origMathQuill;
        return mathQuill;
    }
};

export { mathQuill as default };
